/******************************************************************************
AssinaturaEspectral.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

	Classe para definir a assinatura espectral do conjunto de reas testes
	definidas pelo usurio.

	Ela tem duas funes:

	1. Gerar um arquivo .sig, com as estatsticas da assinatura digital.
		 Esse arquivo  usado pelos mtodos de classificao supervisionada
		 tradicionais.
	2. Gerar um arquivo .rna, com os valores dos pixels na forma normalizada
		 no intervalo [0, 1], para ser usado pelo algoritmo de redes de 
		 neurnios artificiais, o retropropagao do erro. Essa  uma
		 exigncia do algoritmo implementado na API ANNeF. As categorias
		 tambm devem ter valores no mesmo intervalo.

			O formato do arquivo .rna:

			Banda_1		Banda_2		...		Banda_n			Categoria
			0.2				0.6				...		0.8					0.5
			0.3				0.5				...		0.9					0.5
			etc.
	3. Gerar arquivo .hdr, com informaes sobre os temas e bandas.
	
Entrada:

  Arquivo de treinamento (.dat) com os temas, distribudos em amostras 
	por polgonos (ROIs).

Sada:

  Trs arquivos:
	
	1 - .sig - contendo estatsticas de cada tema.
	2 - .rna - contendo valores normalizados dos pixels de treinamento e 
						 das classes.
	3 - .hdr - contendo informaes sobre nmero de temas, nmero de bandas e
						 nome dos temas.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/
//
package pID;

// Pacotes genricos
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.BufferedWriter;
import java.io.BufferedReader;

import java.text.DecimalFormat;

import java.util.Vector;
import java.util.StringTokenizer;

// Pacotes do AWT
import java.awt.Point;
import java.awt.image.Raster;
import java.awt.geom.GeneralPath;
import java.awt.image.BufferedImage;

// Pacotes do Swing
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.swing.JDesktopPane;

// Leitura em modo imediato - para J2SE 1.4+
import javax.imageio.ImageIO;

public class AssinaturaEspectral {

	// ATRIBUTOS

	// sTemas = guarda os temas definidos pelo usurio.
	private String[] sTemas;

	// vetorNomeArquivos - vetor de File(s) com nomes dos arquivos de imagens.
	// vetorImagens - vetor de BufferedImage.
	//
	public Vector vetorNomeArquivos;
	public Vector vetorImagens;
	private	Vector vetorMinimo, vetorMaximo, vetorMedia, vetorCov,
								 vetorVar, vetorMin, vetorMax, vetorSTD, vetorNorm;
	private JDesktopPane desk;

	// CONSTRUTOR DA CLASSE
	//
	public AssinaturaEspectral(JDesktopPane dsk) {
		desk = dsk;
	}

	// MTODOS PBLICOS DA CLASSE
	// Este mtodo apenas abre o arquivo (.dat) que contm os temas, 
	// as amostras e os vrtices dos polgonos (ROIs).
	//
	public File abreArquivoAmostras() {
		int tam;
		String aFile, sufixo;
		File arquivoAmostras = null;

		JFileChooser fileChooser = new JFileChooser(".");
		fileChooser.setDialogTitle("Abre arquivo com amostras (*.dat)");
		fileChooser.addChoosableFileFilter(new TipoArquivo(".dat"));
    int status = fileChooser.showOpenDialog(null);
    if (status == JFileChooser.APPROVE_OPTION) {
			arquivoAmostras = fileChooser.getSelectedFile();
			aFile = arquivoAmostras.getName();
			tam = aFile.length();
			sufixo = aFile.substring((tam-3), tam);
			if (!sufixo.equals("dat")) {
				String str = "Arquivo tem que ser do tipo .dat!";
				JOptionPane.showMessageDialog(desk, str);
				arquivoAmostras = null;
			}
		} else 
			if (status == JFileChooser.CANCEL_OPTION)
				arquivoAmostras = null;

		return arquivoAmostras;
	}

	// Este mtodo abre os arquivos imagens com as bandas selecionadas
	// pelo usurio, armazenando os nomes dos arquivos em vetorNomeArquivos e
	// as imagens, no formato BufferedImage, no vetorImagens.
	//
	public void abreArquivoImagens() {
		File[] arquivos;
		String msg;

		vetorNomeArquivos = new Vector();
		vetorImagens = new Vector();
		JFileChooser fileChooser = new JFileChooser(".");
		fileChooser.setDialogTitle("Abre arquivos com imagens");
    fileChooser.addChoosableFileFilter(new TipoImagem());
		fileChooser.setMultiSelectionEnabled(true);
    int status = fileChooser.showOpenDialog(null);
    if (status == JFileChooser.APPROVE_OPTION) {
			arquivos = fileChooser.getSelectedFiles();
			msg = "AssinaturaEspectral::abreArquivoImagens: arquivo ";
			for (int i = 0; i < arquivos.length; i++) {
				System.out.println(msg + arquivos[i].getParent() + 
													 " ---> " + arquivos[i].getName());
				vetorNomeArquivos.addElement(arquivos[i]);
				abreImagens(arquivos[i]);
			}
		} else 
			if (status == JFileChooser.CANCEL_OPTION)
				arquivos = null;
	}

	// Este mtodo calcula as assinaturas espectrais das amostras
	// definidas pelo usurio que esto armazenadas no arquivo de
	// amostras (.dat).
	//
	public void calculaAssinatura(File arquivoAmostras) {
		int numTemas, numAmostras, numVertices, x = 0, y = 0, xmin, 
				xmax, ymin, ymax;
		double dTemaNorma;
		String line, virgula, aFile;

		FileReader coordFile = null;
		GeneralPath gp;
		StringTokenizer st = null;
		Point pt, ptMin, ptMax;

		// vetorAmostras - vetor com GeneralPath
		//
		Vector vetorAmostras;

		// Estes vetores so usados para armazenar os valores de
		// mximo, mnimo, mdia, varincia/covarincia e valores
		// normalizados dos pixels de cada tema. Esses valores sero usados 
		// para gravar os arquivos de assinatura espectral (.sig) e de valores
		// normalizados (.rna).
		//
		vetorMinimo = new Vector();
		vetorMaximo = new Vector();
		vetorMedia = new Vector();
		vetorNorm = new Vector();
		vetorCov = new Vector();
		vetorVar = new Vector();
		vetorSTD = new Vector();

		try {
			vetorAmostras = new Vector();
			vetorMin = new Vector();
			vetorMax = new Vector();

			// aFile contm somente o nome do arquivo, sem o path.
			// Deu problema rodando de outro diretrio, por isso,
			// mudei abaixo o coordFile! 21/07/2006
			//
			aFile = arquivoAmostras.getName();
			System.out.println("\nAssinaturaEspectral ==> aFile: " + aFile);
			coordFile = new FileReader(arquivoAmostras);
			BufferedReader br = new BufferedReader(coordFile);

			// L a primeira linha do arquivo e pega nmero de temas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numTemas = Integer.parseInt(st.nextToken());
			sTemas = new String[numTemas];

			// Percorre todos os temas
			//
			for (int i = 0; i < numTemas; i++) {

				// Normaliza o valor do tema, para ser usado no algoritmo
				// da rede de neurnios artificiais
				//
				dTemaNorma = (double)((i+1) - 1)/(double)(numTemas - 1);

				// L linha com nome do tema
				//
				line = br.readLine();
				st = new StringTokenizer(line);
				sTemas[i] = st.nextToken();
				System.out.println("\nAssinaturaEspectral::Tema: " + sTemas[i]);

				// L linha com quantidade de amostras
				//
				line = br.readLine();
				st = new StringTokenizer(line);
				numAmostras = Integer.parseInt(st.nextToken());

				// Percorre todas as amostras
				//
				for (int j = 0; j < numAmostras; j++)	{
					gp = new GeneralPath(GeneralPath.WIND_EVEN_ODD);		

					// L linha com quantidade de vrtices da ROI
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					numVertices = Integer.parseInt(st.nextToken());

					// Percorre todos os vrtices
					// L primeiras coordenadas do vrtice
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					x = Integer.parseInt(st.nextToken());
					virgula = st.nextToken();
					y = Integer.parseInt(st.nextToken());
					pt = new Point(x, y);
					xmin = pt.x;
					xmax = pt.x;
					ymin = pt.y;
					ymax = pt.y;

					// Cria general path
					// 1. move para ponto inicial
					//
					gp.moveTo((float)x, (float)y);

					// L demais coordenadas dos vrtices.
					//
					for (int k = 1; k < numVertices; k++)	{

						// L linha com coordenadas do vrtice
						//
						line = br.readLine();
						st = new StringTokenizer(line);
						x = Integer.parseInt(st.nextToken());
						virgula = st.nextToken();
						y = Integer.parseInt(st.nextToken());
						pt = new Point(x, y);

						// 2. Desenha o segmento de reta.
						//
						gp.lineTo((float)x, (float)y);
						xmin = Math.min(xmin, x);
						xmax = Math.max(xmax, x);
						ymin = Math.min(ymin, y);
						ymax = Math.max(ymax, y);
					}	// Fim vrtices

					// 3. Fecha o polgono (GeneralPath).
					//
					gp.closePath();

					// Adiciona a GeneralPath no vetor apropriado.
					//
					vetorAmostras.addElement(gp);

					// Adiciona os valores extremos
					//
					ptMin = new Point(xmin, ymin);
					ptMax = new Point(xmax, ymax);
					vetorMin.addElement(ptMin);
					vetorMax.addElement(ptMax);
				}	// Fim amostras

				// Agora que temos todas as amostras do mesmo tema, vamos
				// calcular os parmetros estatsticos do tema e normalizar seus
				// valores.
				//
				calculaEstatisticas(vetorAmostras, vetorMin, vetorMax, dTemaNorma);
				vetorAmostras = new Vector();
				vetorMin = new Vector();
				vetorMax = new Vector();
			}	//	Fim temas
			coordFile.close();
    } catch (IOException e) {
				String msg1 = "AssinaturaEspectral::calculaAssinatura: ";
				String msg2 = "problema com imagem: ";
        System.out.println(msg1 + msg2 + arquivoAmostras);
        System.exit(0);
    }

		// Agora que temos todos os valores uni, multivariados e normalizados 
		// dos temas para cada banda, vamos gravar esses valores nos arquivos de
		// assinatura espectral (.sig) e valores normalizados (.rna).
		//
		String arquivo = gravaArquivoAssinatura();
		gravaArquivoRNA(arquivo);

	}	// Fim calculaAssinatura
	
	// MTODOS PRIVADOS DA CLASSE

	// Abre arquivo de imagens
	//
	private void abreImagens(File aFile) {
		BufferedImage imagem = null;
		String msg = null;

		try {
			imagem = ImageIO.read(aFile);
	  } catch(Exception e) {
			msg = "AssinaturaEspectral::abreImagens: Problema lendo '";
	    System.out.println(msg + aFile + "'.");
	  }
		vetorImagens.addElement(imagem);
	}

	// Este mtodo calcula as estatsticas e gera valores normalizados de 
	// *UM* tema. Os valores normalizados so para uso do algoritmo da rede
	// de neurnios artificiais (RNA).
	//
	// Entrada:
	//
	//	1. Vetor com as GeneralPath das amostras.
	//	2. Vetor com as BufferedImage, um atributo da classe.
	//	3. Vetores com pontos extremos de cada amostra.
	//
	// Sadas:
	//
	//	1. Vector com as estatsticas do tema por banda: 
	//		 mnimo, mximo, mdia, varincia, desvio padro e covarincia 
	//		 com as outras bandas.
	//	2. Vector com valores normalizados (RNA).
	//
	//	Esses Vectors sero usados posteriormente para gravao dos arquivos:
	//		.sig - para usar nos mtodos tradicionais de classificao.
	//		.rna - para a RNA.
	//
	private void calculaEstatisticas(Vector vecGP, Vector vecMin, 
																	 Vector vecMax, double temaNorma) {
		GeneralPath gp;
		BufferedImage bi;
		Point ptMin, ptMax;
		Vector vecPixels, vecBandas, vecTemp;

    // Variveis para as bandas
		//
		int minimo, maximo, xmin, xmax, ymin, ymax, nc = 0, nc_m = 0, nc_k = 0,
				numPixels, numerador, denominador, nBandas;
    int[]  min_bandaN, max_bandaN;
    long[] sum_bandaN, sum_sqr_bandaN;
    double[] dVAR_bandaN, dMEAN_bandaN, dDESVP_bandaN, dSECOND_bandaN;

		// Array para armazear os valores normalizados. Na primeira dimenso,
		// ele armazena o nmero de pixels das amostras do tema. Na segunda
		// dimenso, ele armazena o nmero de bandas selecionadas pelo usurio
		// mais um, para armazenar o valor normalizado da categoria. Exemplo, se
		// o usurio escolher trs bandas, ento a segunda dimenso ter quatro
		// elementos.
		//
    double[][] dNorm_bandaN;

		// Array para armazenar os valores de pixels por banda.
		//
		int[][] pixelsBanda;

		// Variveis globais
		//
    long[] sum_b1xb2;
		long total = 0;
    double[] dCOVAR_b1xb2, dCORREL_b1xb2, dSUMPROD_b1xb2;
		
		nBandas = vetorImagens.size();

		vecBandas = new Vector();
		for (int j = 0; j < nBandas; j++)
			vecBandas.addElement(new Vector());
		
		// Percorre todas as amostras para obter valores dos pixels
		// das reas (ou conjunto) de treinamento.
		//
		for (int i = 0; i < vecGP.size(); i++) {
			gp = (GeneralPath)vecGP.elementAt(i);
			ptMin = (Point)vecMin.elementAt(i);
			ptMax = (Point)vecMax.elementAt(i);
			xmin = ptMin.x;
			ymin = ptMin.y;
			xmax = ptMax.x;
			ymax = ptMax.y;

			// Percorre as imagens, armazenando os valores de pixels
			// das amostras que devem ser usados para os clculos 
			// estatsticos e normalizao.
			//
			for (int j = 0; j < nBandas; j++) {
				bi = (BufferedImage)vetorImagens.elementAt(j);
				Raster srcR = bi.getRaster();
				vecPixels = (Vector) vecBandas.elementAt(j); 
				for (int k = ymin; k <= ymax; k++) {
					for (int m = xmin; m <= xmax; m++) {
						if (gp.contains((double)m, (double)k)) {
							nc = srcR.getSample(m, k, 0);
							vecPixels.addElement(new Integer(nc));
						}
					}
				}
			}	
		}

		// Agora que temos todos os pixels das amostras de cada
		// tema por banda, calculamos as estatsticas e normalizamos.

		// Inicia variveis
		//
		min_bandaN = new int[nBandas];
		max_bandaN = new int[nBandas];
		sum_bandaN = new long[nBandas];
		sum_sqr_bandaN = new long[nBandas];
		dVAR_bandaN = new double[nBandas];
		dMEAN_bandaN = new double[nBandas];
		dDESVP_bandaN = new double[nBandas];
		dSECOND_bandaN = new double[nBandas];
		for (int i = 0; i < nBandas; i++) {
			min_bandaN[i] = 255;
			max_bandaN[i] = 0;
			sum_bandaN[i] = 0;
			sum_sqr_bandaN[i] = 0;
			dVAR_bandaN[i] = 0.0d;
			dMEAN_bandaN[i] = 0.0d;
			dDESVP_bandaN[i] = 0.0d;
			dSECOND_bandaN[i] = 0.0d;
		}

		int len = 0;
		for (int i = 1; i < nBandas; i++)	{
			len += (nBandas-i);
		}

		dCOVAR_b1xb2 = new double[len];
		dCORREL_b1xb2 = new double[len];
		dSUMPROD_b1xb2 = new double[len];
		sum_b1xb2 = new long[len];

		for (int i = 0; i < len; i++)	{
			dCOVAR_b1xb2[i] = 0.0d;
			dCORREL_b1xb2[i] = 0.0d;
			dSUMPROD_b1xb2[i] = 0.0d;
		}

		vecTemp = (Vector)vecBandas.elementAt(0);

		// O nmero de pixels por tema  o mesmo em cada banda.
		//
		numPixels = vecTemp.size();
		pixelsBanda = new int[nBandas][numPixels];
		dNorm_bandaN = new double[numPixels][nBandas + 1];

		for (int i = 0; i < nBandas; i++) {
			vecPixels = (Vector)vecBandas.elementAt(i);

			// Manipula todos os pixels das amostras do tema em questo.
			//
			for (int j = 0; j < numPixels; j++) {
				Integer integer = (Integer)vecPixels.elementAt(j);
				nc = integer.intValue();
				pixelsBanda[i][j] = nc;

				// Acha mximos e mnimos em cada banda N
				//
        min_bandaN[i] = Math.min(nc, min_bandaN[i]);
        max_bandaN[i] = Math.max(nc, max_bandaN[i]);

				// Acumula valores para clculos univariados da banda N
				//
        sum_bandaN[i] += nc;
        sum_sqr_bandaN[i] += nc*nc;
			}
		}

		int m = 0;
		int indiceSum = 0;
		while (m < nBandas - 1) {
			for (int k = m+1; k < nBandas; k++)	{
				for (int i = 0; i < numPixels; i++) {

						// Pega pixels da banda k e k+1
						//
						nc_m = pixelsBanda[m][i];
						nc_k = pixelsBanda[k][i];

						// Acumula valores para clculo da covarincia
						//
						sum_b1xb2[indiceSum] += nc_m*nc_k;
				}
				indiceSum++;
			}
			m++;
		}

    // Valores univariados da banda k
		//
		total = numPixels;
		for (int k = 0; k < nBandas; k++)	{
			dMEAN_bandaN[k]   = (double)sum_bandaN[k]/total;
      dSECOND_bandaN[k] = (double)sum_sqr_bandaN[k]/total;
		  dVAR_bandaN[k]    = dSECOND_bandaN[k] - dMEAN_bandaN[k]*dMEAN_bandaN[k];
      dDESVP_bandaN[k]  = Math.sqrt(dVAR_bandaN[k]);
		}

		// Armazena os valores univariados do tema para posterior 
		// gravao do arquivo de assinatura espectral (.sig).
		//
		for (int i = 0; i < nBandas; i++) {
			vetorMinimo.addElement(new Integer(min_bandaN[i]));
			vetorMaximo.addElement(new Integer(max_bandaN[i]));
			vetorMedia.addElement(new Double(dMEAN_bandaN[i]));
			vetorVar.addElement(new Double(dVAR_bandaN[i]));
			vetorSTD.addElement(new Double(dDESVP_bandaN[i]));
		}
		
		// Valores multivariados
		// Calcula soma corrigida dos produtos, covarincia e correlao
		//
		m = 0;
		indiceSum = 0;
		while (m < nBandas - 1) {
			for (int k = m+1; k < nBandas; k++)	{
				dSUMPROD_b1xb2[indiceSum] = (double)(sum_b1xb2[indiceSum] - 
								(sum_bandaN[m]*sum_bandaN[k])/total);
				dCOVAR_b1xb2[indiceSum] = 
								(double)(dSUMPROD_b1xb2[indiceSum]/(total - 1));
				dCORREL_b1xb2[indiceSum] = 
								dCOVAR_b1xb2[indiceSum]/(dDESVP_bandaN[m]*dDESVP_bandaN[k]);
				indiceSum++;
			}
			m++;
		}

		// Armazena o valor multivariado do tema para posterior 
		// gravao do arquivo de assinatura espectral (.sig).
		//
		m = 0;
		for (int i = 0; i < nBandas - 1; i++)
			m += (i + 1);
		for (int i = 0; i < m; i++) {
			vetorCov.addElement(new Double(dCOVAR_b1xb2[i]));
		}

		// Normaliza valores dos pixels do tema, para a RNA.
		// Os valores so normalizados pela frmula:
		// x(n) = (x - min(x))/(max(x) - min(x))
		//
		for (int j = 0; j < numPixels; j++) {
			for (int i = 0; i < nBandas; i++) {
				//numerador = pixelsBanda[i][j] - min_bandaN[i];
				//denominador = max_bandaN[i] - min_bandaN[i];
				//dNorm_bandaN[j][i] = (double)numerador/(double)denominador;
				dNorm_bandaN[j][i] = (double)pixelsBanda[i][j]/255.0d;
			}

			// Armazena o valor do tema no ltimo elemento do array
			//
			dNorm_bandaN[j][nBandas] = temaNorma;
		}

		for (int j = 0; j < 10; j++) {
			for (int i = 0; i <= nBandas; i++) {
				System.out.print(dNorm_bandaN[j][i]);
				System.out.print(", ");
			}
			System.out.println();
		}
		
		// Armazena os valores normalizados do tema para posterior 
		// gravao do arquivo de valores normalizados (.rna).
		//
		for (int j = 0; j < numPixels; j++)
			for (int i = 0; i <= nBandas; i++)
				vetorNorm.addElement(new Double(dNorm_bandaN[j][i]));

		// ****** Imprime resultados no prompt do DOS ***********
		// Monta matriz de varincia e covarincia
		//
		double[][] outCov = new double[nBandas][nBandas];

		// Diagonal
		//
		for (int k = 0; k < nBandas; k++)
			outCov[k][k] = dVAR_bandaN[k];

		// Matriz triangular inferior
		//
		m = 0;
		for (int k = 0; k < nBandas; k++)
			for (int i = k+1; i < nBandas; i++)
				outCov[k][i] = dCOVAR_b1xb2[m++];

		// Matriz triangular superior
		//
		m = 0;
		for (int k = 0; k < nBandas; k++)
			for (int i = k+1; i < nBandas; i++)
				outCov[i][k] = dCOVAR_b1xb2[m++];

		// Monta matriz de correlao
		//
		double[][] outCor = new double[nBandas][nBandas];

		// Diagonal
		//
		for (int k = 0; k < nBandas; k++)
			outCor[k][k] = 1.0d;

		// Matriz triangular inferior
		//
		m = 0;
		for (int k = 0; k < nBandas; k++)
			for (int i = k+1; i < nBandas; i++)
				outCor[k][i] = dCORREL_b1xb2[m++];

		// Matriz triangular superior
		//
		m = 0;
		for (int k = 0; k < nBandas; k++)
			for (int i = k+1; i < nBandas; i++)
				outCor[i][k] = dCORREL_b1xb2[m++];

		// Define os formatos de impresso
		//
		String padrao = "##.##";
		DecimalFormat formato = null;
		String[] media_banda = new String[nBandas];
		String[] varia_banda = new String[nBandas];
		String[] desvp_banda = new String[nBandas];
		String[] outCov_banda = new String[nBandas*nBandas];
		String[] outCor_banda = new String[nBandas*nBandas];
		for (int k = 0; k < nBandas; k++)	{
			formato = new DecimalFormat(padrao);
			media_banda[k] = formato.format(dMEAN_bandaN[k]);
			varia_banda[k] = formato.format(dVAR_bandaN[k]);
			desvp_banda[k] = formato.format(dDESVP_bandaN[k]);
		}
		for (int k = 0; k < nBandas; k++)
			for (int i = 0; i < nBandas; i++)
				outCov_banda[k*nBandas+i] = formato.format(outCov[k][i]);
		for (int k = 0; k < nBandas; k++)
			for (int i = 0; i < nBandas; i++)
				outCor_banda[k*nBandas+i] = formato.format(outCor[k][i]);
	
    System.out.println("\nEstatisticas univariadas:"); 
    System.out.print("\nBanda");
		for (int k = 1; k <= nBandas; k++)
			System.out.print("\t"+k);
    System.out.print("\nmedia\t");
		for (int k = 0; k < nBandas; k++)
			System.out.print(media_banda[k] + "\t");
    System.out.print("\nvar.\t");
		for (int k = 0; k < nBandas; k++)
			System.out.print(varia_banda[k] + "\t");
    System.out.print("\ndesvp.\t");
		for (int k = 0; k < nBandas; k++)
			System.out.print(desvp_banda[k] + "\t");
    System.out.print("\nminimo\t");
		for (int k = 0; k < nBandas; k++)
			System.out.print(min_bandaN[k] + "\t");
    System.out.print("\nmaximo\t");
		for (int k = 0; k < nBandas; k++)
			System.out.print(max_bandaN[k] + "\t");
		System.out.println("\n\nEstatisticas multivariadas:");
    System.out.println("\nMatriz de variancia-covariancia");
    System.out.print("\nBanda");
		for (int k = 1; k <= nBandas; k++)
			System.out.print("\t"+k);
		for (int k = 1; k <= nBandas; k++){
			System.out.print("\n"+k);
			for (int i = 0; i < nBandas; i++)
				System.out.print("\t" + outCov_banda[(k-1)*nBandas+i]);
		}
    System.out.println("\n\nMatriz de correlacao");
    System.out.print("\nBanda");
		for (int k = 1; k <= nBandas; k++)
			System.out.print("\t"+k);
		for (int k = 1; k <= nBandas; k++){
			System.out.print("\n"+k);
			for (int i = 0; i < nBandas; i++)
				System.out.print("\t" + outCor_banda[(k-1)*nBandas+i]);
		}
    System.out.print("\n");
	}	// Fim calculaEstatisticas

	// Grava arquivos de assinaturas espectrais (.sig)
	// Para o usurio, a gravao do arquivo .rna  transparente. Ele  feito
	// automaticamente.
	//
	private String gravaArquivoAssinatura() {
		int aux, doIt;
		String str = null, sNome = null, sTemp = null;

		String msg = "Salvar arquivo de estatsticas (.sig)?";
		doIt = JOptionPane.showConfirmDialog(null, msg,
										"Salvar", JOptionPane.YES_NO_OPTION);
		if (doIt == 0) {	// Sim
      str = "Entre nome do arquivo (sem extenso): ";
			sNome = JOptionPane.showInputDialog(str);
			if (sNome != null) {
				sTemp = sNome;
				sNome += ".sig";
				System.out.println("Salvar arquivo : " + sNome);
				try {
					FileWriter writer = new FileWriter(sNome);
					BufferedWriter buf_writer = new BufferedWriter(writer);
					String ln = "";
					int nt, nb, offsetTema;
					int indMin, indMax, indMed, indVar, indCov, indSTD;

					// Identifica a quantidade de temas definidos pelo usurio
					//
					indMin = indMax = indMed = indVar = indCov = indSTD = 0;

					// Nmero de temas
					//
					nt = sTemas.length;

					// Nmero de bandas
					//
					nb = vetorImagens.size();

					// Grava nmero de temas
					//
					ln += nt;
					grava_linha(ln, buf_writer);

					// Grava quantidade de bandas
					//
					ln = "";
					ln += nb;
					grava_linha(ln, buf_writer);					

					// Percorre todos os temas
					//
					for (int indT = 0; indT < nt; indT++) {
						ln = "";

						// Grava nome do tema
						//
						ln += sTemas[indT];
						grava_linha(ln, buf_writer);

						// Percorre todas as bandas
						//
						for (int indB = 0; indB < nb ; indB++ ){
							ln = "";

							// Grava nome do arquivo imagem da banda
							//
							ln += vetorNomeArquivos.elementAt(indB);
							grava_linha(ln, buf_writer);
							ln = "";

							// Grava valor mnimo
							//
							ln += vetorMinimo.elementAt(indMin++);
							grava_linha(ln, buf_writer);
							ln = "";

							// Grava valor mximo
							//
							ln += vetorMaximo.elementAt(indMax++);
							grava_linha(ln, buf_writer);
							ln = "";

							// Grava valor da mdia
							//
							ln += vetorMedia.elementAt(indMed++);
							grava_linha(ln, buf_writer);
							ln = "";

							// Grava valor da varincia
							//
							ln += vetorVar.elementAt(indVar++);
							grava_linha(ln, buf_writer);
							ln = "";

							// Grava valor do desvio padro
							//
							ln += vetorSTD.elementAt(indSTD++);
							grava_linha(ln, buf_writer);
						}

						// Grava valor de covarincia				
						//
						ln = "";
						ln += "Covariancias";
						grava_linha(ln, buf_writer);
						aux = 0;
						for (int i = 0; i < nb - 1; i++)
							aux += (i + 1);
						for (int i = 0; i < aux; i++) {
							ln = "";
							ln += vetorCov.elementAt(indCov++);
							grava_linha(ln, buf_writer);
						}
					}
					buf_writer.close();
				} catch (IOException e) {
					String str1 = "AssinaturaEspectral::Erro ";
					String str2 = "na gravao do arquivo "	+ sNome;
					JOptionPane.showMessageDialog(desk, str1 + str2);
				}
			} 
		}
		return sTemp;
	}

	private void grava_linha(String texto, BufferedWriter buf_writer) {
		try	{
			buf_writer.write(texto);
			buf_writer.newLine();		
		} catch (IOException e) {
			String str1 = "AssinaturaEspectral::Erro ";
			String str2 = "na gravao da linha " + texto;
			JOptionPane.showMessageDialog(desk, str1 + str2);
		} 
	}

	// Grava arquivos com valores normalizados (.rna) e cabealho (.hdr).
	// Para o usurio, a gravao desses arquivos  transparente. Ela  feita
	// automaticamente.
	//
	// O arquivo de cabealho contm apenas informaes sobre a quantidade de
	// temas usados, o nmero de bandas, nome das bandas e o nome de cada tema.
	//
	private void gravaArquivoRNA(String sNome) {
		int nTemas, nBandas, i;
		String ln = "", sHeader;
		FileWriter writer = null, header = null;
		BufferedWriter buf_writer = null, buf_header = null;

		sHeader = sNome + ".hdr";
		sNome += ".rna";
		System.out.println("AssinaturaEspectral::Salvar arquivo RNA: " + sNome);
		System.out.println("AssinaturaEspectral::Salvar arquivo RNA: " + sHeader);
		try {
			writer = new FileWriter(sNome);
			header = new FileWriter(sHeader);
			buf_writer = new BufferedWriter(writer);
			buf_header = new BufferedWriter(header);
			ln = "";

			// *************************************************
			// **** Informaes do arquivo cabealho (.hdr) ****
			// *************************************************

			// Nmero de bandas
			//
			nBandas = vetorImagens.size();

			// Grava nmero de bandas
			//
			ln = "";
			ln += nBandas;
			grava_linha(ln, buf_header);

			// Grava nome do arquivo imagem da banda
			//
			ln = "";
			for (int indB = 0; indB < nBandas ; indB++ ){
				ln += vetorNomeArquivos.elementAt(indB);
				grava_linha(ln, buf_header);
				ln = "";
			}

			// Nmero de temas
			//
			nTemas = sTemas.length;

			// Grava nmero de temas
			//
			ln += nTemas;
			grava_linha(ln, buf_header);

			// Grava nomes dos temas
			//
			for (int indT = 0; indT < nTemas; indT++) {
				ln = "";
				ln += sTemas[indT];
				grava_linha(ln, buf_header);
			}

			// ****************************************************************
			// **** Informaes do arquivo com valores normalizados (.rna) ****
			// ****************************************************************

			// Grava valores normalizados dos pixels
			//
			i = 0;
			while (i < vetorNorm.size()) {
				ln = "";
				for (int j = 0; j <= nBandas; j++) {
					ln += vetorNorm.elementAt(i);
					ln += " ";
					i++;
				}
				grava_linha(ln, buf_writer);
			}

			buf_writer.close();
			buf_header.close();
		} catch (IOException e) {
				String arqs = sNome + ", " + sHeader;
				String str1 = "AssinaturaEspectral::Erro ";
				String str2 = "na gravao dos arquivos: "	+ arqs;
				JOptionPane.showMessageDialog(desk, str1 + str2);
		}
	}

}
