/******************************************************************************
CPrincipal.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

	Calcula imagens como componentes principais

Entrada:

	Vector com bandas.

Sada:

	Imagens com as principais componentes gravadas em disco.
	A classe mostra apenas a primeira componente.
	Resultados da anlise so impressos no prompt do DOS.
	A classe pode ser modificada para gravar os resultados em arquivo texto.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/

package pID;

// Pacotes genricos
import java.io.File;
import java.util.Vector;
import java.io.FileWriter;
import java.io.IOException;
import java.io.BufferedWriter;
import java.text.DecimalFormat;

// Pacote do AWT
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;

// pacotes do Swing
import javax.swing.JInternalFrame;
import javax.swing.JFileChooser;
import javax.swing.JDesktopPane;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.ImageIcon;
import javax.swing.JLabel;

// Leitura/gravao em modo imediato - para J2SE 1.4+
import javax.imageio.ImageIO;

public class CPrincipal {

	// ATRIBUTOS

	private	int componente[][];

	// N = nmero de pixels da banda
	//
	private int N;

	// Atributos para salvar em arquivo de resultados
	//
	String elemento, sAutoValores[], sAutoVetores[][], sPorcentoIndiv[],
				 sPorcentoAcum[], sCorrelacao[][], prefixo;

	// CONSTRUTOR DA CLASSE
	//
	public CPrincipal(Vector src, int w, int h) {

		int lin, col;

		// O try  usado no clculo de Matrizes
		//
		try {

			// Objeto Matriz calcula autovalores a partir da Matriz de covarincia, 
			// calculada com objeto Multivariada
			// Primeiro, calcula matrizes de var-covar e correlao
			//
			Multivariada multiv = new Multivariada(src);
			Matriz m_cov = multiv.retornaVarCovar();
			lin = m_cov.getNrow();
			col = m_cov.getNcol();

			// Arrays para resultados
			//
			double[][] a_autoValores = new double[lin][1];
			double[][] a_autoVetores = new double[lin][col];
			double[][] a_autoVal = new double[lin+1][col];

			// Array para mostrar correlao entre cada CP (colunas) 
			// e cada banda (linhas)
			//
			double[][] a_correl = new double[lin][col];

			// CRIA INSTNCIA DE MATRIZES
			//
			Matriz m_autoValores = new Matriz(a_autoValores);
			Matriz m_autoVetores = new Matriz(a_autoVetores);
			Matriz m_autoVal		 = new Matriz(a_autoVal);
			Matriz m_correl			 = new Matriz(a_correl);

			// Armazena em a_autoVal autovalores na primeira linha e
			// autovetores em linhas subsequentes
			//
			try {
					m_autoVal = m_cov.autoValores();
			} catch (IllegalArgumentException ie) {
				String msg = "Autovalores: ";
				String titulo = "Clculo Matricial";
				JOptionPane.showMessageDialog(null, msg + ie, titulo, 0);
			}

			// Pega autovalores
			//
			m_autoValores = m_autoVal.retornaValores();

			// Pega autovetores
			//
			m_autoVetores = m_autoVal.retornaVetores();

			// Ordena valores de autovalores (componentes principais) e
			// autovetores
			//
			int[] ordem = m_autoValores.ordenaAutoValores();
			m_autoVetores.ordenaAutoVetores(ordem);

			// Armazena valores para gerao de arquivo com resultados
			//
			String padrao = "#.###";
			DecimalFormat formato = null;
			formato = new DecimalFormat(padrao);
			String elemento;

			sAutoValores = new String[lin];
			for (int i = 0; i < lin; i++) {
				elemento = formato.format(m_autoValores.getElement(i, 0));
				sAutoValores[i] = elemento;
			}

			sAutoVetores = new String[lin][col];
			for (int i = 0; i < lin; i++)
				for (int j = 0; j < col; j++) {
					elemento = formato.format(m_autoVetores.getElement(i, j));
					sAutoVetores[i][j] = elemento;
				}

			System.out.println("autovalores:");
			m_autoValores.imprimeTransposta();
			m_autoVetores.imprimeMatriz();

			// Calcula a varincia explicada por cada componente principal
			// (autovalores)
			//
			double[] porcentVaria = m_autoValores.porcentagemVariancia();
			sPorcentoIndiv = new String[col];
			String msg1 = "Porcentagem de variancia total explicada ";
			String msg2 = "por cada componente:";
			System.out.println("\n" + msg1 + msg2);
			for (int i = 0; i < col; i++) {
				elemento = formato.format((double)porcentVaria[i]);
				sPorcentoIndiv[i] = elemento;
				System.out.print(elemento + "\t");
			}
			System.out.println();
	
			System.out.println("\nPorcentagem de variancia acumulada:");
			sPorcentoAcum = new String[col];
			double soma = 0.0d;
			for (int i = 0; i < col; i++) {
				soma += porcentVaria[i];
				elemento = formato.format((double)soma);
				sPorcentoAcum[i] = elemento;
				System.out.print(elemento + "\t");
			}

			System.out.println();
		
			// Calcula a correlao entre cada CP e cada banda
			// Pega valores de varincia da matriz de var-covar
			//
			double[] a_variancia = new double[col];
			for (int i = 0; i < col; i++)
				a_variancia[i] = m_cov.getElement(i, i);

			m_correl.correlacaoComponentePrincipal(a_variancia, m_autoValores, m_autoVetores);
			System.out.println("\nGrau de correlacao (linha = banda/coluna = CP)");
			m_correl.imprimeMatriz();

			sCorrelacao = new String[lin][col];
			for (int i = 0; i < lin; i++)
				for (int j = 0; j < col; j++) {
					elemento = formato.format(m_correl.getElement(i, j));
					sCorrelacao[i][j] = elemento;
				}


			// Calcula imagens como componentes principais e mostra resultado
			//
			calculaCPrincipal(src, m_autoVetores, w, h);

		} catch (IllegalArgumentException ie) {
			String msg = "Matriz: ";
			String titulo = "Componente Principal::Clculo Matricial";
			JOptionPane.showMessageDialog(null, msg + ie, titulo, 0);
		}
	}		// fim construtor

	// MTODOS PBLICOS

	// Mostra apenas a primeira componente
	//
	public void mostraComponente(JDesktopPane desktop, int w, int h) {

		JInternalFrame frame = new JInternalFrame("Componente #1",
													          true, //resizable
													          true, //closable
													          true, //maximizable
													          true);//iconifiable
		desktop.add(frame);

		int tipo = BufferedImage.TYPE_BYTE_GRAY;
		int indice;
		BufferedImage biImagem = new BufferedImage(w, h, tipo); 
		WritableRaster destWR = biImagem.getRaster(); 
		for (int j = 0; j < h; j++)
			for (int i = 0; i < w; i++) {
				indice = j*w + i;
				destWR.setSample(i, j, 0, componente[0][indice]);
			}

		// Mostra a imagem
		//
    JLabel lsrc = new JLabel(new ImageIcon(biImagem));
    frame.getContentPane().add(new JScrollPane(lsrc));
		frame.setSize(w, h);
    frame.setVisible(true);
	}

	// MTODOS PRIVADOS

	private void calculaCPrincipal(Vector vecImagens, Matriz m_autoVetores, 
																	int w, int h) {

		// Cria variveis locais
		//
		double soma = 0.0;
		int nComponentes, nBandas;

		// Nmero de componentes principais (CP) = nmero de bandas
		//
		nBandas = vecImagens.size();
		nComponentes = m_autoVetores.getNrow();

		//  Array duplo para armazenar pixels das bandas.
		//	src[#bandas][N] -> array com valores das bandas originais
		//  componente[#componentes][N] -> array para imagens das CP
		//	largura = src[i].length
		//	altura = src[i][0].length
		//
		int src[][] = new int[nBandas][];
		double a_autoVetores[][] = new double[nComponentes][nComponentes];

		// Armazena valores dos pixels das imagens originais
		//
		for (int k = 0; k < nBandas; k++)
			src[k] = (int[])vecImagens.elementAt(k);

		// Prepara array 'componente' para receber valores de pixels das CP
		//
		N = src[0].length;
		componente = new int[nComponentes][N];

		// Reverte Matriz para array
		//
		a_autoVetores = m_autoVetores.getArrayCopy();

		// Calcula imagens com CP, usando autovetores
		// ref.: Jensen, pg. 179, equao 7-66
		// indices: p = #componentes
		//					k = #bandas
		//					N = #pixels
		//
		for (int p = 0; p < nComponentes; p++)	{
			for (int i = 0; i < N; i++) {
				for (int k = 0; k < nBandas; k++)	{

					// Pega pixels da banda k
					//
					soma += a_autoVetores[k][p]*src[k][i];
				}
				if (soma > 255.0)
					soma = 255.0;
				if (soma < 0.0)
					soma = 0.0;
				componente[p][i] = (int)soma;
				soma = 0.0;
			}
		}

		// Grava as CPs como imagens
		//
		int tipo = BufferedImage.TYPE_BYTE_GRAY;
		int indice;
		String arquivo = null;
		BufferedImage[] biImagem;
		WritableRaster destWR;

		biImagem = new BufferedImage[nComponentes];
		for (int p = 0; p < nComponentes; p++)	{
			biImagem[p] = new BufferedImage(w, h, tipo); 
			destWR = biImagem[p].getRaster(); 
			for (int j = 0; j < h; j++)
				for (int i = 0; i < w; i++) {
					indice = j*w + i;
					destWR.setSample(i, j, 0, componente[p][indice]);
				}
		}

		// Grava imagem.
		//
    boolean status = saveFile(biImagem);
    if (!status) {
			String msg = "Erro de IO no arquivo!!";
			String titulo = "Salva arquivo";
      JOptionPane.showMessageDialog(null, msg, titulo, JOptionPane.ERROR_MESSAGE);
		}

	}		// calculaCPrincipal

	private boolean saveFile(BufferedImage[] dest) {
		File file = null;
    JFileChooser fc;
		String nome;
		int result;

    fc = new JFileChooser();
    fc.setCurrentDirectory(new File("."));
		fc.setDialogTitle("Salva componente principal (*.png)");
		fc.addChoosableFileFilter(new TipoArquivo(".png"));

    // Define nome padro
		//
		File fFile = new File ("componente");
    fc.setSelectedFile(fFile);

    // Abre dialog
		//
    result = fc.showSaveDialog(null);

    if (result == JFileChooser.CANCEL_OPTION) {
			return true;
    } else if (result == JFileChooser.APPROVE_OPTION) {
				fFile = fc.getSelectedFile();
        if (fFile.exists()) {
					int response = JOptionPane.showConfirmDialog (null,
										"Atualiza arquivo existente?","Confirma Atualizao",
										JOptionPane.OK_CANCEL_OPTION,
										JOptionPane.QUESTION_MESSAGE);
          if (response == JOptionPane.CANCEL_OPTION)
						return false;
        }
				nome = fFile.getName();
				String search = ".";
				int i = nome.indexOf(search);
				if (i > 0)
					prefixo = nome.substring(0, i);
				else
					prefixo = nome;
				for (int cp = 0; cp < dest.length; cp++) {
					try {
							prefixo += "_cp" + (cp + 1) + ".png";
							System.out.println("\nGravou arquivo -> " + 
																	fc.getCurrentDirectory() + "\\"+ prefixo + "\n");
							ImageIO.write(dest[cp], "png", new File(prefixo));
					} catch (IOException e) {
							System.out.println("Problema gravando arquivo. ");
							System.exit(0);
					}
				if (i > 0)
					prefixo = nome.substring(0, i);
				else
					prefixo = nome;
				}	// for

				// Grava arquivo com resultados da anlise
				//
				gravaResultadosCP();
    } else {
				return false;
    }
		return true;
  } // saveFile

	private void gravaResultadosCP() {

		String sNome = prefixo + "_res.dcp";
		String ln = "";

		try {
			FileWriter writer = new FileWriter(sNome);
			BufferedWriter buf_writer = new BufferedWriter(writer);
			ln = "Resultado da anlise das componentes principais";
			grava_linha(ln, buf_writer);
			ln = "Autovalores:";
			grava_linha(ln, buf_writer);
			ln = "";
			for (int i = 0; i < sAutoValores.length; i++)	{
				ln += sAutoValores[i] + "\t";
			}
			grava_linha(ln, buf_writer);
			ln = "Matriz de autovetores:";
			grava_linha(ln, buf_writer);
			ln = "";
			for (int i = 0; i < sAutoVetores.length; i++) {
				for (int j = 0; j < sAutoVetores[0].length; j++) {
					ln += sAutoVetores[i][j] + "\t";
				}
				grava_linha(ln, buf_writer);
				ln = "";
			}
			String msg1 = "Porcentagem de varincia total explicada ";
			String msg2 = "por cada componente:";
			ln = msg1 + msg2;
			grava_linha(ln, buf_writer);
			ln = "";
			for (int i = 0; i < sPorcentoIndiv.length; i++)	{
				ln += sPorcentoIndiv[i] + "\t";
			}
			grava_linha(ln, buf_writer);
			ln = "Porcentagem de varincia acumulada:";
			grava_linha(ln, buf_writer);
			ln = "";
			for (int i = 0; i < sPorcentoAcum.length; i++)	{
				ln += sPorcentoAcum[i] + "\t";
			}
			grava_linha(ln, buf_writer);
			ln = "Grau de correlao (linha = banda/coluna = CP)";
			grava_linha(ln, buf_writer);
			ln = "";
			for (int i = 0; i < sCorrelacao.length; i++) {
				for (int j = 0; j < sCorrelacao[0].length; j++) {
					ln += sCorrelacao[i][j] + "\t";
				}
				grava_linha(ln, buf_writer);
				ln = "";
			}
			buf_writer.close();
		} catch (IOException e) {
				String str = "Erro na gravao do arquivo "
												+ sNome;
				String titulo = "Componente Principal";
				JOptionPane.showMessageDialog(null, str, titulo, 0);
		}
	}

	private void grava_linha(String texto, BufferedWriter buf_writer) {
		try	{
			buf_writer.write(texto);
			buf_writer.newLine();		
		} catch (IOException e) {
			String str = "Erro na gravao da linha " + texto;
			String titulo = "Componente Principal";
			JOptionPane.showMessageDialog(null, str, titulo, 0);
		} 
	}

}
