/******************************************************************************
Cluster.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

	Esta classe realiza a anlise de agrupamento no supervisionado 
	("cluster analysis"), a partir das estatsticas multivariadas das bandas
	escolhidas para a classificao.

	Mtodos disponveis:

	1. Heurstico.
	2. Leader.
	3. Joiner.
	4. K-mdias.
	5. Fuzzy c-mdias (FCM).

Entrada:

	Parmetros para cada tipo de mtodo. Verificar abaixo.

Sada:

	Vetor com pixels classificados.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/

package pID;

// Pacotes genricos
import java.text.DecimalFormat;
import java.util.Random;

// Pacotes do Swing
import javax.swing.JDesktopPane;
import javax.swing.JOptionPane;

public class Cluster {

	// ATRIBUTOS

	// Estes trs atributos so usados no mtodo doisClusters:
	//
	private double[] centroideX, centroideY;
	private double distanciaEuclidiana;

	// Nmero de grupos (clusters) desejado
	//
	private int nGrupos;

	// Para o caso do algoritmo heurstico, o dimetro do grupo tem que
	// ser definido.
	//
	private int diametroGrupo;

	// Para o FCM, fornecer os seguintes parmetros:
	// 1. Nmero mximo de iteraes. O algoritmo pra se no convergir
	//		aps maxIteracao.
	//
	private int maxIteracao;

	// 2. Valor de parada (epsilon) e expoente de ponderao ou 
	//		'fuzziness'
	//
	private double epsilon, fuzziness;
	
  // CONSTRUTOR

	public Cluster() {
  }

	// MTODOS PBLICOS

	// Define quantidade de grupos para algoritmo k-mdias.
	//
	public void paramKMedias(final JDesktopPane desktop) {
		String msg;

    try {
			String sGrupos = JOptionPane.showInputDialog(null, "Nmero de grupos: ");
			if (sGrupos != null)
				nGrupos = Integer.parseInt(sGrupos);
			else {
				nGrupos = 2;
				msg = "Assumindo 2 grupos";
				JOptionPane.showMessageDialog(desktop, msg);
			}
    } catch (NumberFormatException e) {
			msg = "Parmetro invlido";
			JOptionPane.showMessageDialog(desktop, msg);
		}
	}

	// Define quantidade de grupos e dimetro para algoritmo heurstico.
	//
	public void paramHeuristico(final JDesktopPane desktop) {
		boolean valorGrupo = false;
		boolean valorDiametro = false;

		do	{
	    try {
				String sGrupos = JOptionPane.showInputDialog(null, "Nmero de grupos: ");
				if (sGrupos != null)
					nGrupos = Integer.parseInt(sGrupos);
				else 
					nGrupos = 2;
				valorGrupo = true;
	    } catch (NumberFormatException e) {
				String str = "Parmetro invlido";
				JOptionPane.showMessageDialog(desktop, str);
			}
		} while(!valorGrupo);

		do	{
	    try {
				String sDiam = JOptionPane.showInputDialog(null, "Dimetro do grupo: ");
				if (sDiam != null)
					diametroGrupo = Integer.parseInt(sDiam);
				else
					diametroGrupo = 10;
				valorDiametro = true;
	    } catch (NumberFormatException e) {
				String str = "Parmetro invlido";
				JOptionPane.showMessageDialog(desktop, str);
			}
		} while(!valorDiametro);

	}

	// Define parmetros para o fuzzy c-means:
	//  maxIteracao - nmero mximo de iteraes.
	//	epsilon		-	valor de parada requerido pelo algoritmo.
	//							Um valor sugerido  0.01 (Bezdek et al., 1984).
	//	fuzziness - nmero fuzzy 'm', ou expoente de ponderao.
	//
	public void paramFCM(final JDesktopPane desktop) {
		boolean valorGrupo = false;
		boolean valorIteracao = false;
		boolean valorEpsilon = false;
		boolean valorFuzziness = false;

		do	{
	    try {
				String sGrupos = JOptionPane.showInputDialog(null, "Nmero de grupos: ");
				if (sGrupos != null)
					nGrupos = Integer.parseInt(sGrupos);
				else
					nGrupos = 2;
				valorGrupo = true;
	    } catch (NumberFormatException e) {
				String str = "Parmetro invlido";
				JOptionPane.showMessageDialog(desktop, str);
			}
		} while(!valorGrupo);

		do	{
	    try {
				String sMaxIteracao = JOptionPane.showInputDialog(null, 
													"Nmero mximo de iteraes: ");
				if (sMaxIteracao != null)
					maxIteracao = Integer.parseInt(sMaxIteracao);
				else
					maxIteracao = 40;
				valorIteracao = true;
	    } catch (NumberFormatException e) {
				String str = "Parmetro invlido";
				JOptionPane.showMessageDialog(desktop, str);
			}
		} while(!valorIteracao);

		do	{
	    try {
				String sEpsilon = JOptionPane.showInputDialog(null, 
															"Valor de parada (epsilon): ");
				if (sEpsilon != null)
					epsilon = Double.parseDouble(sEpsilon);
				else
					epsilon = 1.5;
				valorEpsilon = true;
	    } catch (NumberFormatException e) {
				String str = "Parmetro invlido";
				JOptionPane.showMessageDialog(desktop, str);
			}
		} while(!valorEpsilon);

		do	{
	    try {
				String sFuzziness = JOptionPane.showInputDialog(null, 
															"Valor 'fuzziness' (m): ");
				if (sFuzziness != null)
					fuzziness = Double.parseDouble(sFuzziness);
				else
					fuzziness = 1.5;
				valorFuzziness = true;
	    } catch (NumberFormatException e) {
				String str = "Parmetro invlido";
				JOptionPane.showMessageDialog(desktop, str);
			}
		} while(!valorFuzziness);
	}

	/*
	 *  Heuristico
	 *
	 *  Fonte: Helmuth Spth, Cluster analysis algorithms. 
	 *				 Chichester,WS,England:Ellis Horwood Ltd,	1982. 226p.
	 *
	 * 	Algoritmo heurstico, o mais simples, considera cada objeto s uma vez 
	 *  e o aloca a um grupo.  
	 *
   *  Existem dois mtodos:
	 *	
	 *	1. Para matriz de nmeros reais: double variavel[m][n].
   *  2. Para matriz de nmeros inteiros: int variavel[m][n].
   *
   *  Sada:
   *    grupos[m] -  o vetor com os nmeros das parties (clusters). 
   *
	 */

	// Mtodo para array double
	//
	public int[] leader(double[][] variavel, int nmax, double rho) {
	
		int[] F = new int[nmax];	
		int n, r, m, p;
		double d, t;
		boolean flag;

		m = variavel.length;
		p = variavel[0].length;
		int[] grupos = new int[m];
		
		n = 1;
		F[0] = 0;
		for (int i = 0; i < m; i++) {
			grupos[i] = 0;
			flag = true;
			for (int j = 0; j < n; j++) {
				r = F[j];
				d = 0.0d;
				for (int k = 0; k < p; k++) {
					t = variavel[r][k] - variavel[i][k];
					d += t*t;					
				} // 1
				if (Math.sqrt(d) <= rho) {
					grupos[i] = j+1;
					flag = false; 
					break;
				}
			}
			if (flag) {
				if (n != nmax) {
					n++;
					grupos[i] = n;
					F[n-1] = i;
				}
			}
		}
		return grupos;
	}	// double leader

	// Mtodo para array int
	//
	public int[] leader(int[][] variavel, int nmax, double rho) {
	
		int[] F = new int[nmax];	
		int n, r, m, p;
		double d, t;
		boolean flag;

		m = variavel.length;
		p = variavel[0].length;
		int[] grupos = new int[m];
		
		n = 1;
		F[0] = 0;
		for (int i = 0; i < m; i++) {
			grupos[i] = 0;
			flag = true;
			for (int j = 0; j < n; j++) {
				r = F[j];
				d = 0.0d;
				for (int k = 0; k < p; k++) {
					t = (double)variavel[r][k] - (double)variavel[i][k];
					d += t*t;					
				} // 1
				if (Math.sqrt(d) <= rho) {
					grupos[i] = j+1;
					flag = false; 
					break;
				}
			}
			if (flag) {
				if (n != nmax) {
					n++;
					grupos[i] = n;
					F[n-1] = i;
				}
			}
		}
		return grupos;
	}	// int leader

	/*
	 *  Heuristico
	 *
	 *  Fonte: Helmuth Spth, Cluster analysis algorithms. 
	 *				 Chichester,WS,England:Ellis Horwood Ltd,	1982. 226p.
	 *
	 * 	Uma modificao do algoritmo anterior (HEURISTICO), usando mtrica 
   *  Euclidiana.
	 *
   *  Existem dois mtodos: 
	 *
   *    1. Para matriz de nmeros reais: double variavel[m][n].
   *    2. Para matriz de nmeros inteiros: int  variavel[m][n].
   *
   *  Parmetros de entrada:
	 *
   *    TAU       - um valor de	corte na alocao; 
   *    NMAX      - nmero mximo de clusters a ser formado.
   *
   *  Sada:
	 *
   *    grupos[m] -  o vetor com os nmeros das parties (clusters). 
   *
	 */

	// Mtodo para array double
	//
	public int[] joiner(double[][] variavel, int nmax, double tau) {
		int m = variavel.length;
		int p = variavel[0].length;
		int[] grupos = new int[m];
		double[][] s = new double[nmax][p];
		int[] q = new int[nmax];
		int r, u, n = 0;

		double h, f = 0.0, t, v, w;
		boolean flag;
		
		for (int i = 0; i < m; i++)
			grupos[i] = 0;
		do {
			flag = false;
			n++;
			if (n > nmax) {
				return grupos;
			}
			r = -1;
			u = 0;
			h = 0.0;
			for (int i = 1; i <= m-1; i++) {
				if (grupos[i-1] == 0) {
					u = i-1;
					for (int j = i+1; j <= m; j++) {
						if (grupos[j-1] == 0) {
							f = 0.0;
							for (int k = 0; k < p; k++) {
								t = variavel[i-1][k] - variavel[j-1][k];
								f += t*t;													
							}
							if (f > h) {
								h = f;
								r = i-1;
							}
						}
					}
				}
			}
			if (r == -1) {
				if (u != 0) {
					grupos[u] = n;
					return grupos;
				} else
					return grupos;
			}
			grupos[r] = n;
			q[n-1] = 1;
			for (int k = 0; k < p; k++) {
				s[n-1][k] = variavel[r][k];
			}
			do {
				flag = false;
				h = 1.e30;	
				for (int i = 1; i <= m; i++ ) {
					if (grupos[i-1] == 0) {
						f = 0.0;
						for (int k = 0; k < p; k++) {
							t = s[n-1][k] - variavel[i-1][k];
							f += t*t;				
						}
						if (f < h) {
							h = f;
							r = i-1;
						}
					}
				}	
				if (Math.sqrt(h) <= tau) {
					grupos[r] = n;
					u = q[n-1];
					v = (double)u;
					w = 1.0/(v + 1.0);
					q[n-1] = u + 1;
					for (int k = 0; k < p; k++) {
						s[n-1][k] = w*(v*s[n-1][k] + variavel[r][k]);
					}
					flag = true;
				}
			} while (flag);
		} while (Math.sqrt(h) > tau);
		return grupos;
	}	// double joiner

	// Mtodo para array int
	//
	public int[] joiner(int[][] variavel, int nmax, int tau) {
		int m = variavel.length;
		int p = variavel[0].length;
		int[] grupos = new int[m];
		double[][] s = new double[nmax][p];
		int[] q = new int[nmax];
		int r, u, n = 0;

		double h, f = 0.0, t, v, w;
		boolean flag;

		for (int i = 0; i < m; i++)
			grupos[i] = 0;	
		do {
			flag = false;
			n++;
			if (n > nmax) {
				return grupos;
			}
			r = -1;
			u = 0;
			h = 0.0;
			for (int i = 1; i <= m-1; i++) {
				if (grupos[i-1] == 0) {
					u = i-1;
					for (int j = i+1; j <= m; j++) {
						if (grupos[j-1] == 0) {
							f = 0.0;
							for (int k = 0; k < p; k++) {
								t = (double)variavel[i-1][k] - (double)variavel[j-1][k];
								f += t*t;													
							}	
							if (f > h) {
								h = f;
								r = i-1;
							}
						}
					}
				}
			}
			if (r == -1) {
				if (u != 0) {
					grupos[u] = n;
					return grupos;
				} else
					return grupos;
			}
			grupos[r] = n;
			q[n-1] = 1;
			for (int k = 0; k < p; k++) {
				s[n-1][k] = (double)variavel[r][k];
			}
			do {
				flag = false;
				h = 1.e30;	
				for (int i = 1; i <= m; i++ ) {
					if (grupos[i-1] == 0) {
						f = 0.0;
						for (int k = 0; k < p; k++) {
							t = s[n-1][k] - (double)variavel[i-1][k];
							f += t*t;				
						}
						if (f < h) {
							h = f;
							r = i-1;
						}
					}
				}
				if (Math.sqrt(h) <= (double)tau) {
					grupos[r] = n;
					u = q[n-1];
					v = (double)u;
					w = 1.0/(v + 1.0);
					q[n-1] = u + 1;
					for (int k = 0; k < p; k++) {
						s[n-1][k] = w*(v*s[n-1][k] + (double)variavel[r][k]);
					}
					flag = true;
				}
			} while (flag);
		} while (Math.sqrt(h) > (double)tau);
		return grupos;
	}	// int joiner

	/*
	 *  DoisGrupos
	 *
	 *  Fonte: Helmuth Spth, Cluster analysis algorithms. 
	 *				 Chichester,WS,England:Ellis Horwood Ltd,	1982. 226p.
	 * 
	 * 	Este algoritmo gera exatos dois clusters
	 */

	public int[] doisGrupos(double[][] variavel, int idr) {
		int m = variavel.length;
		int p = variavel[0].length;
		int[] grupos = new int[m];
		centroideX = new double[p];
		centroideY = new double[p];
		int ic, jc, g1, g2, ip, i1, i2;
		double dmax, h, h1, h2, f, d1, d2;
		boolean flag, flag15;

		f = 0.0;
		ic = jc = i1 = i2 = 0;
		for (int i = 0; i < m; i++)
			grupos[i] = 0;
		if (m <= 2)
			return grupos;
		dmax = 0.0;
		for (int i = 1; i <= m-1; i++) {
			for (int j = i+1; j <= m; j++) {
				h = 0.0;
				for (int k = 0; k < p; k++) {
					f = variavel[i-1][k] - variavel[j-1][k];
					h += f*f;													
				}
				if (h > dmax) {
					dmax = h;
					ic = i-1;
					jc = j-1;
				}
			}
		}
		for (int k = 0; k < p; k++) {
			centroideX[k] = variavel[ic][k];
			centroideY[k] = variavel[jc][k];
		}
		grupos[ic] = 1;
		grupos[jc] = 2;
		g1 = 1;
		g2 = 1;
		ip = 2;
		flag = true;
		do {
			flag15 = true;
			h = 0.0;
			for (int k = 0; k < p; k++) {
				f = centroideX[k] - centroideY[k];
				h += f*f;
			}		// 7
			distanciaEuclidiana = Math.sqrt(h);
			if (idr != 0) {
				for (int i = 0; i < m; i++)
					System.out.print("  "+grupos[i]);
				System.out.println();
				System.out.println("DS = "+distanciaEuclidiana);
			}
			ip++;	
			if (ip > m)	{
				return grupos;
			}
			d1 = 1.e30;			
			d2 = d1;
			for (int i = 1; i <= m; i++) {
				if (grupos[i-1] == 0) {
					h1 = 0.0;
					h2 = 0.0;
					for (int k = 0; k < p; k++) {
						h = variavel[i-1][k];
						f = centroideX[k] - h;
						h1 += f*f;
						f = centroideY[k] - h;
						h2 += f*f;
					}
					if (h1 < d1) {
						d1 = h1;
						i1 = i-1;
					}
					if (h2 < d2) {
						d2 = h2;
						i2 = i-1;
					}
				}
			}	
			if (d1 < d2) {
				grupos[i1] = 1;
				h = (double)g1;
				f = 1.0/(h + 1.0);
				for (int k = 0; k < p; k++) {
					centroideX[k] = (h*centroideX[k] + variavel[i1][k])*f;
				}
				g1++;
				flag15 = false;
			}
			if (flag15)	{
				grupos[i2] = 2;
				h = (double)g2;
				f = 1.0/(h + 1.0);
				for (int k = 0; k < p; k++) {
					centroideY[k] = (h*centroideY[k] + variavel[i2][k])*f;
				}
				g2++;
			}
		}	while (flag);
		return grupos;
	}	// DoisGrupos

	// Estas trs funes retornam valores apenas do DoisClusters
	// Coordenada X do centride
	//
	public double[] retornaCentroideX() {
		return centroideX;
	}

	// Coordenada Y do centride
	//
	public double[] retornaCentroideY() {
		return centroideY;
	}

	// Distncia entre os dois clusters
	//
	public double retornaDistancia() {
		return distanciaEuclidiana;
	}

	/*
	 *
	 * k-mdias
	 *
	 * Fonte: Helmuth Spth,
	 *			  Cluster Analysis Algorithms	for Data Reduction and 
	 *				Classification of Objects, Ellis Horwood, 1982
	 *
   *  Existem dois mtodos: 
	 *
   *    1. Para matriz de nmeros reais: double variavel[m][n].
   *    2. Para matriz de nmeros inteiros: int  variavel[m][n].
   *
   *  Parmetros de entrada:
	 *
   *    variavel - array com pixels a serem classificados.
   *    nc       - nmero de grupos a ser formado.
	 *		idr			 - controla a impresso das iteraes.
   *
   *  Sada:
	 *
   *    grupos[m] -  o vetor com os nmeros das parties (clusters). 
   *
	 *
	*/

	// Mtodo para array double
	//
	public int[] kMeans(double[][] variavel, int[] grupos, int nc, int idr) {
		int m = variavel.length;
		int n = variavel[0].length;
		int r, u, v = 0, w = 0, i, it;
		int[] q = new int[nc];
		double a, b, soma_sdQ, f, g, h, t = 0.0;
		double[] sdQ = new double[nc];						// sdQ = soma dos quadrados
		double[][] centroide = new double[nc][n];
		boolean flag = true;

    if (m <= 0) {
			System.out.println("KMEANS - Erro fatal!  M <= 0.");
			System.exit(0);
		}

    if (n <= 0) {
			System.out.println("KMEANS - Erro fatal!  N <= 0.");
			System.exit(0);
		}

    if (nc <= 0) {
			System.out.println("KMEANS - Erro fatal!  NC <= 0.");
			System.exit(0);
		}

    if ( nc > m ) {
			System.out.println("KMEANS - Erro fatal!  NC > M.");
			System.exit(0);
		}

		// Assegura que a atribuio de grupos est certa
		//
		for (int ic = 0; ic < m; ic++) {
			if (grupos[ic] < 1 || grupos[ic] > nc) {
				return grupos;
			}
		}

		// Se existe s um grupo, ento finalizar
		//
    if (nc == 1) {
			return grupos;
		}

		// Determina a populao de grupos
		//
		v = 0;
		w = 0;
		for (int j = 0; j < nc; j++) {
			q[j] = 0;
      sdQ[j] = 0.0;
			for (int k = 0; k < n; k++) {
				centroide[j][k] = 0.0;
			} 
		}

		// Conta o nmero de grupos vazios
		//
		for (int ic = 1; ic <= m; ic++) {
			r = grupos[ic-1];
      if (r < 1 || r > nc) {
				return grupos;
			}
      q[r-1]++;
			for (int k = 0; k < n; k++) {
				centroide[r-1][k] = centroide[r-1][k] + variavel[ic-1][k];
			} 
		}

		// Determina o centride de cada grupo (clculo da mdia).
		//
		for (int j = 0; j < nc; j++) {
			r = q[j];
      if (r == 0) {
				return grupos;
			}
      f = 1.0/(double)r;
			for (int k = 0; k < n; k++) {
				centroide[j][k] *= f;
			}
		}

		// Clculo da soma dos quadrados
		//
		for (int ic = 0; ic < m; ic++) {
			r = grupos[ic];
      f = 0.0;
			for (int k = 0; k < n; k++) {
				t = centroide[r-1][k] - variavel[ic][k];
        f += t*t;
			}
      sdQ[r-1] += f;
			if (sdQ[r-1] < 0)	{
				System.out.println("sdQ["+(r-1)+"] = "+sdQ[r-1]);
			}
		}	
    soma_sdQ = 0.0;
		for (int j = 0; j < nc; j++) {
			soma_sdQ += sdQ[j];
		}	

		// Se todos os pixels foram examinados sem mudanas, nada mais resta
		// para ser feito
		//
		i = -1;
		it = 0;
		do {		// 10
	    i++;
      if (i > m - 1)
         i -= m;
      if (it == m - 1) {

				// Imprime coordenadas dos centrides, quadrados das somas (sdQ) e 
				// soma dos quadrados (soma_sdQ)
				//
				hMeans_Imprime(centroide, sdQ, soma_sdQ, n, nc);
				return grupos;
			}
      r = grupos[i];
      u = q[r-1];
      if (u > 1) {
				h = (double)u;
		    h /= (h - 1.0);
		    f = 0.0;
				for (int k = 0; k < n; k++) {
					t = centroide[r-1][k] - variavel[i][k];
	        f += t*t;
				}
				a = h*f;
				b = 1.e30;
				for (int j = 0; j < nc; j++) {
					if (j != r-1) {
						u = q[j];
						h = (double)u;
						h /= (h + 1.0);
						f = 0.0;
						for (int k = 0; k < n; k++) {
							t = centroide[j][k] - variavel[i][k];
						  f += t*t;
						}
						f = h*f;
						if (f <= b) {
							b = f;
							v = j + 1;
							w = u;
						} // if (f <= b)
					} // if (j != r)
				}	
	      if (b >= a) {
		      it++;
				} else {
			    it = 0;	
		      sdQ[r-1] -= a;
		      sdQ[v-1] += b;
		      soma_sdQ = (soma_sdQ - a) + b;
		      h = (double)(q[r-1]);
		      g = (double)w;
		      a = 1.0/(h - 1.0);
		      b = 1.0/(g + 1.0);
					for (int k = 0; k < n; k++) {
						f = variavel[i][k];
						centroide[r-1][k] = (h*centroide[r-1][k] - f)*a;
            centroide[v-1][k] = (g*centroide[v-1][k] + f)*b;
					  f += t*t;
					} // 15
		      grupos[i] = v;
		      q[r-1]--;
		      q[v-1]++;

					// Imprime iteraes, se idr = 1
					//
				  if (idr == 1) {
						System.out.println("i = "+i+", soma_sdQ = "+soma_sdQ);
						for (int j = 0; j < m; j++) {
							System.out.print(grupos[j] + ",");
						}
						System.out.println();
					}

				}		// else if (b >= a)
			}	// if (u > 1)
		}	while (flag); 

		System.out.println("return: apos while");
		return grupos;
	}		// double KMeans

	// Mtodo para array int
	//
	public int[] kMeans(int[][] variavel, int[] grupos, int nc, int idr) {
		int m = variavel.length;
		int n = variavel[0].length;
		int r, u, v = 0, w = 0, i, it;
		int[] q = new int[nc];
		double a, b, soma_sdQ, f, g, h, t = 0.0;
		double[] sdQ = new double[nc];						// sdQ = soma dos quadrados
		double[][] centroide = new double[nc][n];
		boolean flag = true;

    if (m <= 0) {
			System.out.println("KMEANS - Fatal error!  M <= 0.");
			System.exit(0);
		}

    if (n <= 0) {
			System.out.println("KMEANS - Fatal error!  N <= 0.");
			System.exit(0);
		}

    if (nc <= 0) {
			System.out.println("KMEANS - Fatal error!  NC <= 0.");
			System.exit(0);
		}

    if ( nc > m ) {
			System.out.println("KMEANS - Fatal error!  NC > M.");
			System.exit(0);
		}

		// Assegura que a atribuio de grupos est certa
		//
		for (int ic = 0; ic < m; ic++) {
			if (grupos[ic] < 1 || grupos[ic] > nc) {
				return grupos;
			}
		}

		// Se existe s um grupo, ento finalizar
		//
    if (nc == 1) {
			return grupos;
		}

		// Determina a populao de grupos
		//
		v = 0;
		w = 0;
		for (int j = 0; j < nc; j++) {
			q[j] = 0;
      sdQ[j] = 0.0;
			for (int k = 0; k < n; k++) {
				centroide[j][k] = 0.0;
			}
		}

		// Conta o nmero de grupos vazios
		//
		for (int ic = 1; ic <= m; ic++) {
			r = grupos[ic-1];
      if (r < 1 || r > nc) {
				return grupos;
			}
      q[r-1]++;
			for (int k = 0; k < n; k++) {
				centroide[r-1][k] = centroide[r-1][k] + (double)variavel[ic-1][k];
			} 
		}

		// Determina o centride de cada grupo (clculo da mdia).
		//
		for (int j = 0; j < nc; j++) {
			r = q[j];
      if (r == 0) {
				return grupos;
			}
      f = 1.0/(double)r;
			for (int k = 0; k < n; k++) {
				centroide[j][k] *= f;
			} 
		}	

		// Clculo da soma dos quadrados
		//
		for (int ic = 0; ic < m; ic++) {
			r = grupos[ic];
      f = 0.0;
			for (int k = 0; k < n; k++) {
				t = centroide[r-1][k] - (double)variavel[ic][k];
        f += t*t;
			}
      sdQ[r-1] += f;
			if (sdQ[r-1] < 0)	{
				System.out.println("sdQ["+(r-1)+"] = "+sdQ[r-1]);
			}
		}
    soma_sdQ = 0.0;
		for (int j = 0; j < nc; j++) {
			soma_sdQ += sdQ[j];
		}

		// Se todos os pixels foram examinados sem mudanas, nada mais resta
		// para ser feito
		//
		i = -1;
		it = 0;
		do {		// 10
	    i++;
      if (i > m - 1)
         i -= m;
      if (it == m - 1) {

				// Imprime coordenadas dos centrides, quadrados das somas (sdQ) e 
				// soma dos quadrados (soma_sdQ)
				//
				hMeans_Imprime(centroide, sdQ, soma_sdQ, n, nc);
				return grupos;
			}
      r = grupos[i];
      u = q[r-1];
      if (u > 1) {
				h = (double)u;
		    h /= (h - 1.0);
		    f = 0.0;
				for (int k = 0; k < n; k++) {
					t = centroide[r-1][k] - (double)variavel[i][k];
	        f += t*t;
				} // 11
				a = h*f;
				b = 1.e30;
				for (int j = 0; j < nc; j++) {
					if (j != r-1) {
						u = q[j];
						h = (double)u;
						h /= (h + 1.0);
						f = 0.0;
						for (int k = 0; k < n; k++) {
							t = centroide[j][k] - (double)variavel[i][k];
						  f += t*t;
						}
						f = h*f;
						if (f <= b) {
							b = f;
							v = j + 1;
							w = u;
						} // if (f <= b)
					} // if (j != r)
				}
	      if (b >= a) {	
		      it++;
				} else {
			    it = 0;
		      sdQ[r-1] -= a;
		      sdQ[v-1] += b;
		      soma_sdQ = (soma_sdQ - a) + b;
		      h = (double)(q[r-1]);
		      g = (double)w;
		      a = 1.0/(h - 1.0);
		      b = 1.0/(g + 1.0);
					for (int k = 0; k < n; k++) {
						f = (double)variavel[i][k];
						centroide[r-1][k] = (h*centroide[r-1][k] - f)*a;
            centroide[v-1][k] = (g*centroide[v-1][k] + f)*b;
					  f += t*t;
					}
		      grupos[i] = v;
		      q[r-1]--;
		      q[v-1]++;

					// Imprime iteraes, se idr = 1
					//
				  if (idr == 1) {
						System.out.println("i = "+i+", soma_sdQ = "+soma_sdQ);
						for (int j = 0; j < m; j++) {
							System.out.print(grupos[j] + ",");
						}
						System.out.println();
					}

				}		// else if (b >= a)
			}	// if (u > 1)
		}	while (flag);
		System.out.println("return: apos while");
		return grupos;
	}		// int KMeans

	public int retornaGrupos() {
		return nGrupos;
	}	

	public int retornaDiametro() {
		return diametroGrupo;
	}	

	public int retornaMaximoIteracoes() {
		return maxIteracao;
	}	

	public double retornaEpsilon() {
		return epsilon;
	}	

	public double retornaFuzziness() {
		return fuzziness;
	}	

	// Algoritmo fuzzy c-means (FCM).
	// Agrupa casos usando o algoritmo fuzzy c-means.
	// O algoritmo converge, ou pra, quando o valor psilon, 
	// conforme anunciado no algoritmo de Bezdek et al. (1984),  
	// alcanado ou quando o nmero mximo de iteraes  alcanado.
	// Nesse caso, o algoritmo no convergiu.
  //
	// Parmetros de entrada:
	//	variavel    - array com variveis (colunas) e casos (linhas) a serem agrupados.
	//	grupos      -	array com a partio inicial dos grupos. O mtodo
	//						    retorna nesse array a partio final.
	//	numGrupos   - nmero de grupos (#clusters).
	//  maxIteracao - nmero mximo de iteraes
	//	epsilon		  -	valor de parada requerido pelo algoritmo.
	//							  Um valor sugerido  0.01 (Bezdek et al., 1984).
	//	fuzziness   - nmero fuzzy 'm', ou expoente de ponderao, conforme 
	//							  algoritmo FCM (Bezdek et al., 1984). Os autores sugerem
	//							  para 'm' um valor	no intervalo [1.5, 3].
  //
	// Parmetro de sada:
	//	O mtodo retorna o array de inteiros 'grupos' contendo os
	//	grupos classificados.

	public int[] fuzzyCMeans(double[][] variavel, int[] grupos, int numGrupos, 
		int maxIteracoes, double aEpsilon, double aFuzziness) {

		// L nmero de casos = #linhas
		//
		int N = variavel.length;
		int iteracao = 0;

		// L nmero de variveis = #colunas
		//
		int numVariaveis = variavel[0].length;

		// Define objeto Matriz m_mu (N x numGrupos) para armazenar os nmeros 
		// fuzzy (fuzzy membership) ou graus de pertinncia - mu.
		// Para cada caso existem tantos mu's quantos grupos definidos,
		// sendo que os mu's devem satisfazer duas condies:
		// (Bezdek et al., 1984) (Usamos uma forma transposta  de Bezek):
		//	(1) 0 < soma mu(k,i) < N, para i = 1,...,numGrupos e k = 1,...,N. Onde
		//			soma = somatrio. Ou seja, a soma dos mus de todos os casos de 
		//			uma famlia de parties (soma nas colunas)  um valor positivo
		//			e menor que o nmero de casos existentes.
		//	(2) soma mu(k,i) = 1, para k = 1,...,N e i = 1,...,numGrupos. Ou seja,
		//			a soma dos mus de cada caso (soma nas linhas) deve ser igual a um.
		//
		double a_mu[][] = new double[N][numGrupos];
		double a_muPrime[][] = new double[N][numGrupos];

		// Array para armazenar distncias entre casos e centrides
		//
		double a_dist[][] = new double[N][numGrupos];

		// Array para centrides
		//
		double a_centroides[][] = new double[numGrupos][numVariaveis];

		// CRIA INSTNCIA DE MATRIZ
		//
		Matriz m_mu  = new Matriz(a_mu);
		Matriz m_muPrime  = new Matriz(a_muPrime);
		Matriz m_dist  = new Matriz(a_dist);
		Matriz m_centroides  = new Matriz(a_centroides);

		// Varivel lgica de parada do algoritmo
		//
		boolean terminou = false;

		// Expoente usado para calcular novos valores de mu.
		// Equao 11b de Bezdek
		//
		double exp = 2/(aFuzziness - 1);

		// Inicia aleatoriamente os primeiros valores de mu
		//
		Random rand_mu = new Random();
		double soma = 0.0, mu = 0.0;
		for (int i = 0; i < N; i++)	{
			for (int c = 0; c < numGrupos; c++)	{

					// Cada caso do grupo recebe um mu aleatrio.
					//
					mu = 0.01 + rand_mu.nextFloat();
					m_mu.setElement(i, c, mu);
					soma += mu;
				}

				// Normaliza resultado para que soma de mus = 1
				//
				for (int c = 0; c < numGrupos; c++) {
					mu = m_mu.getElement(i, c);
					mu /= soma;
					m_mu.setElement(i, c, mu);
				}
				soma = 0.0f;
		}

		//		System.out.println("\nmu inicial");
		//		m_mu.imprimeMatriz();

		do {
			iteracao++;
			//System.out.println("\n******* ITERACAO " + iteracao + "********\n");

			// Calcula centrides
			//
			m_centroides = calculaCentroides(m_mu, variavel, aFuzziness);
			//System.out.println("\ncentroides");
			//m_centroides.imprimeMatriz();

			// Calcula distncias dos centrides para os casos.
			// Mtrica usada: distncia euclidiana
			//
			m_dist = calculaDistancia(m_centroides, variavel);
			//System.out.println("\ndistancia centroides");
			//m_dist.imprimeMatriz(9);

			// Calcula novos valores de mu [U(k+1)]
			//
			m_muPrime = atualizaValoresMu(m_dist, exp);

			//System.out.println("m_muPrime");
			//m_muPrime.imprimeMatriz(9);

			// Compara mu(k+1) com m(k).
			// Se mu(k+1) - mu(k) < epsilon, o algoritmo pra.
			// Seno, faz mu(k) = mu(k+1) e retorna ao clculo de centrides
			//
			terminou = comparaMu(m_mu, m_muPrime, aEpsilon);

			// Calcula valores de mu considerando valores anteriores (m_muPrime)
			//
				m_mu = m_muPrime.copy();
		} while ((!terminou) && (iteracao < maxIteracoes));
	
		// Atribui casos aos grupos somente se grau de pertinncia difusa
		// for maior que .70.
		// Criamos um grupo a mais. Esse ltimo grupo conter os valores 
		// fuzzy no classificados como um grupo "puro." Estamos usando
		// um valor de corte = .70 para os grupos "puros."
		//
		for (int i = 0; i < N; i++) {
			for (int c = 0; c < numGrupos; c++) {
				if (m_muPrime.getElement(i, c) > 0.70) 
					grupos[i] = c + 1;
				else 
					grupos[i] = numGrupos + 1;
			}
		}

		System.out.println("\nO algoritmo convergiu em " + iteracao + " iteracoes");

		return grupos;
	}

	// MTODOS PRIVADOS

  // Fuzzy c-means:
	// Mtodo para calcular os centrides das funes de pertinncia
	//
	private Matriz calculaCentroides(Matriz m_mu, double[][] variavel, double aFuzziness) {

		double numerador = 0.0, denominador = 0.0, mu = 0.0;
		double a_muExp[][] = m_mu.getArrayCopy();
		int indice;

		// O nmero de colunas de m_mu define o nmero de grupos
		//
		int lin = m_mu.getNrow();
		int col = m_mu.getNcol();
		int numVariaveis = variavel[0].length;
		double a_centroide[][] = new double[col][numVariaveis];
		double a_valores[][] = new double[lin][col*numVariaveis];

		// CRIA INSTNCIA DE MATRIZ
		//
		Matriz m_valores  = new Matriz(a_valores);
		Matriz m_centroide  = new Matriz(a_centroide);

		// Eleva cada valor mu ao expoente fuzziness (mu^fuzziness)
		//
		for(int i = 0; i < lin; i++)
  		for(int j = 0; j < col; j++) {
				mu = m_mu.getElement(i, j);
				a_muExp[i][j] = Math.pow(mu, aFuzziness);
			}

		// Multiplica mu^fuzziness por valores das variveis
		//
		for(int i = 0; i < lin; i++) {
			indice = 0;
  		for(int j = 0; j < col; j++) {
				for (int v = 0; v < numVariaveis; v++) {
					m_valores.setElement(i, indice++, a_muExp[i][j]*variavel[i][v]);
				}
			}
			indice = 0;
		}

		//System.out.println("\nmu(k)*Variavel");
		//m_valores.imprimeMatriz(9);

		// Calcula os centrides
		// Percorre todos os grupos (= col)
		//
		indice = 0;
		for(int c = 0; c < col; c++) {

			// Percorre todas as variveis
			//
			for (int v = 0; v < numVariaveis; v++) {

				// Percorre todos os elementos dos grupos
				//
				for (int i = c; i < lin; i += col) {
					numerador += m_valores.getElement(i, indice);
					denominador += a_muExp[i][c];
				}
				m_centroide.setElement(c, v, numerador/denominador);
				indice++;
				numerador = denominador = 0.0;
			}
		}
		return m_centroide;
	}

  // Fuzzy c-means:
	// Mtodo para calcular distncia dos casos para os centrides.
	//
  private Matriz calculaDistancia(Matriz m_centroides, double[][] variavel)	{

		int numGrupos = m_centroides.getNrow();

		// Nmero de casos = #linhas variavel
		//
		int N = variavel.length;

		// Nmero de variveis = #colunas variavel
		//
		int numVariaveis = variavel[0].length;
		double a_distancia[][] = new double[N][numGrupos];
		double [] pontoCentroide, pontoCaso;
		pontoCentroide = new double[numVariaveis];
		pontoCaso = new double[numVariaveis];
		double dist = 0.0;

		// CRIA INSTNCIA DE MATRIZ
		//
		Matriz m_distancia  = new Matriz(a_distancia);

		// Para cada caso e grupo
		//
		for(int i = 0; i < N; i++) {
			for(int c = 0; c < numGrupos; c++) {
				for(int v = 0; v < numVariaveis; v++) {
					pontoCentroide[v] = m_centroides.getElement(c, v);
					pontoCaso[v] = variavel[i][v];
				}
				for(int v = 0; v < numVariaveis; v++) 
					dist += (pontoCentroide[v]-pontoCaso[v])*(pontoCentroide[v]-pontoCaso[v]);
				m_distancia.setElement(i, c, Math.sqrt(dist));
				dist = 0.0;
			}
		}
		return m_distancia;
	}

  // Fuzzy c-means:
	// Mtodo para calcular novos valores de mu.
	// Equao 11b de Bezdek
	//
  private Matriz atualizaValoresMu(Matriz m_dist, double exp)	{

		// Nmero de casos
		//
		int N = m_dist.getNrow();

		// Nmero de variveis
		//
		int numGrupos = m_dist.getNcol();
		double denominador = 0.0, termo = 0.0;
		double[][] mu = new double[N][numGrupos];

		// CRIA INSTNCIA DE MATRIZ
		//
		Matriz m_mu = new Matriz(mu);

		for(int i = 0; i < N; i++) {
			for(int c = 0; c < numGrupos; c++) {
				termo = m_dist.getElement(i, c);
				for(int c1 = 0; c1 < numGrupos; c1++) {
					denominador += Math.pow(termo/m_dist.getElement(i, c1), exp);
				}
				m_mu.setElement(i, c, 1/denominador);
				denominador = 0.0;
			}
		}

		return m_mu;
	}

  // Fuzzy c-means:
	// Mtodo para determinar a parada do algoritmo
	//
	private boolean comparaMu(Matriz m_Uk1, Matriz m_Uk, double aEpsilon) {
		int N = m_Uk1.getNrow();
		int numGrupos = m_Uk1.getNcol();
		boolean terminou = false;
		double[] maximo = new double[numGrupos];
		double maximoGlobal = -1.e30, uk1, uk;

		for (int c = 0; c < numGrupos; c++)
			maximo[c] = -1.e30;

		for (int c = 0; c < numGrupos; c++) {
			for (int i = 0; i < N; i++) {
				uk1 = m_Uk1.getElement(i, c);
				uk = m_Uk.getElement(i, c);
				maximo[c] = Math.max(maximo[c], (uk - uk1)); 
				//System.out.print((uk - uk1)+", ");
			}
			//System.out.println();
		}

		//System.out.println("\nvetor de maximas");
		for (int c = 0; c < numGrupos; c++) {
			maximoGlobal = Math.max(maximo[c], maximoGlobal);
			//System.out.print(maximo[c]+", ");
		}
		//System.out.println();

		if (maximoGlobal < aEpsilon)
			return terminou = true;

		return terminou;
	}

	// A seguinte rotina serve para o KMeans
	//
	private void hMeans_Imprime(double[][] centroide, double[] sdQ, double soma_sdQ, 
		int n, int nc) {

		String padrao = "#0.0";
		DecimalFormat formato = null;
		formato = new DecimalFormat(padrao);
		String elemento;

		// Centroide[j][k] = coordenadas dos clusters (centrides)
		//
    System.out.println();
		for (int k = 0; k < n; k++)	{
			for (int j = 0; j < nc; j++) {
				elemento = formato.format((double)centroide[j][k]);
				System.out.print(elemento + "\t");
			}
			System.out.println();
		}
    System.out.println();

		// sdQ[j] = soma do quadrado das distncias entre os membros
		//				  do j-simo cluster e seus centrides.
		for (int j = 0; j < nc; j++) {
			elemento = formato.format((double)sdQ[j]);
			System.out.print(elemento + "\t");
		}

		// soma_sdQ = soma dos sdQ[j]'s
		//
		elemento = formato.format((double)soma_sdQ);
		System.out.print(elemento + "\t");
		System.out.println();
	}

}
