/******************************************************************************
DistanciaMinima.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

	Classifica imagem pelo mtodo da distncia mnima da mdia.
	O clculo da distncia Euclidiana se faz de duas maneiras:

	1. Usando o tringulo de Pitgoras;
	2. Usando a distncia "entorno da quadra" ou city block, com a diferena
		 do valor absoluto.

	A distncia Euclidiana  mais cara computacionalmente.

Entrada:

	Arquivo .SIG, com informaes estatsticas e nome dos arquivos com as 
	bandas a serem classificadas.
	BufferedImage das bandas selecionadas.

Sada:

	Imagem classificada gravada em disco, com os grupos.
	Conforme o mtodo 'gravaClasse', somente sete grupos so identificados.
	Para aceitar um nmero maior de classes, modificar o mtodo citado.
	Arquivo .CSV, com informaes sobre os grupos classificados.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/

package pID;

// Pacotes genricos
import java.io.File;
import java.util.Vector;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.BufferedWriter;
import java.io.BufferedReader;
import java.text.DecimalFormat;
import java.util.StringTokenizer;

// Pacotes do AWT
import java.awt.Component;
import java.awt.Container;
import java.awt.GridLayout;
import java.awt.image.Raster;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

// Pacotes do Swing
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JButton;
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.swing.JDesktopPane;
import javax.swing.ButtonGroup;
import javax.swing.JRadioButton;

// Peitura em modo imediato - para J2SE 1.4+
import javax.imageio.ImageIO;

public class DistanciaMinima {

	// ATRIBUTOS

	// sTemas = guarda os temas definidos pelo usurio.
	//
	private String[] sTemas;

	// bandas - BufferedImage das bandas selecionadas.
	//
	private BufferedImage[] banda;

	// Nome do arquivo imagem
	//
	private String arquivoImagem;

	private JDesktopPane desk;

	// EsTe atributo  um vetor de vetores para as mdias
	// por tema e banda. O algoritmo da distncia mnima s
	// usa as mdias das reas de treinamento.
	//
	private Vector vetorMedBanda;

	// CONSTRUTOR DA CLASSE

	public DistanciaMinima(){ }

	// MTODOS PBLICOS DA CLASSE

	// EsTe mtodo apenas abre o arquivo com as assinaturas
	// espectrais (.sig) calculadas pelo objeto AssinaturaEspectral
	// com as estatsticas multivariadas de cada tema a ser classificado.
	//
	public File abreArquivoAssinaturas(JDesktopPane desktop) {
		int tam;
		File arquivoAssinaturas = null;
		String aFile, sufixo;

		desk = desktop;
		JFileChooser fileChooser = new JFileChooser(".");
		fileChooser.setDialogTitle("Abre arquivo com assinaturas (*.sig)");
		fileChooser.addChoosableFileFilter(new TipoArquivo(".sig"));
    int status = fileChooser.showOpenDialog(null);
    if (status == JFileChooser.APPROVE_OPTION) {
			arquivoAssinaturas = fileChooser.getSelectedFile();

			// Verifica que abriu arquivo do tipo .sig
			aFile = arquivoAssinaturas.getName();
			tam = aFile.length();
			sufixo = aFile.substring((tam-3), tam);
			System.out.println("sufixo = " + sufixo);
			if (!sufixo.equals("sig")) {
				String str = "Arquivo tem que ser do tipo .sig!";
				String titulo = "Distncia Mnima";
				JOptionPane.showMessageDialog(desk, str, titulo, 0);
				arquivoAssinaturas = null;
			}
		} else 
			if (status == JFileChooser.CANCEL_OPTION)
				arquivoAssinaturas = null;

		return arquivoAssinaturas;
	}

	// Este mtodo l e abre os arquivos das imagens, 
	// armazenando-os em BufferedImage.
	// Como os nomes das imagens aparecem de maneira repetida
	// no arquivo .sig para cada tema, precisamos de ler apenas
	// um tema para acessar os valores necessrios.
	//
	public void abreBandas(File arquivoAssinaturas) {
		int tam, numTemas, numBandas = 1;
		FileReader estatFile = null;
		String line, aFile, nomeBanda[] = null;
		StringTokenizer st = null;
		File file;

		try {
			aFile = arquivoAssinaturas.getName();
			estatFile = new FileReader(aFile);
			BufferedReader br = new BufferedReader(estatFile);

			// L a primeira linha do arquivo e pega nmero de temas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numTemas = Integer.parseInt(st.nextToken());

			// L a segunda linha do arquivo e pega nmero de bandas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numBandas = Integer.parseInt(st.nextToken());
			nomeBanda = new String[numBandas];

			// Percorre apenas um tema
			// l linha com nome do tema
			//
			line = br.readLine();
			st = new StringTokenizer(line);

			// Percorre todas as bandas
			//
			for (int nb = 0; nb < numBandas; nb++) {

				// L linha com nome da banda
				//
				line = br.readLine();
				st = new StringTokenizer(line);
				nomeBanda[nb] = st.nextToken();
				System.out.println("\nBanda: " + nomeBanda[nb]);

				// L linha com valor mnimo do tema na banda
				//
				line = br.readLine();
				st = new StringTokenizer(line);

				// L linha com valor mximo do tema na banda
				//
				line = br.readLine();
				st = new StringTokenizer(line);

				// L linha com mdia do tema na banda
				//
				line = br.readLine();
				st = new StringTokenizer(line);

				// L linha com valor de varincia
				//
				line = br.readLine();
				st = new StringTokenizer(line);

				// L linha com valor do desvio padro
				//
				line = br.readLine();
				st = new StringTokenizer(line);
			} // fim bandas
			estatFile.close();
    } catch (IOException e) {
				String msg = "abreBandas: problema com imagem: " + arquivoAssinaturas;
				String titulo = "Distncia Mnima";
				JOptionPane.showMessageDialog(desk, msg, titulo, 0);
        System.exit(0);
    }

		arquivoImagem = nomeBanda[0];

		// Armazena pixels em BufferedImage
		//
		banda = new BufferedImage[numBandas];
		for (int i = 0; i < numBandas; i++) {
	    file = new File(nomeBanda[i]);
			try {
				banda[i] = ImageIO.read(file);
				System.out.println("banda["+i+"]: w = "+banda[i].getWidth()+
												 " h = "+banda[i].getHeight());
		  } catch(Exception e) {
				String msg = "abreImagens: problema lendo: " + file;
				String titulo = "Distncia Mnima";
				JOptionPane.showMessageDialog(desk, msg, titulo, 0);
			}
		}
	}

	// Este mtodo classifica os pixels de cada banda de acordo
	// com as assinaturas espectrais das amostras definidas pelo 
	// usurio usando o mtodo da distncia mnima da mdia.
	// A rigor, a regra de deciso do mtodo s precisa da mdia
	// do tema.
	//
	public void classificaImagem(File arquivoAssinaturas) {
		int numTemas, numBandas, minimo, maximo, aux;
		double media, var, cov, dLimiar = 0.0;
		FileReader estatFile = null;
		String line, aFile, nomeBanda, covar, msg;
		Vector vetorBandas, vetorMedia;
		StringTokenizer st = null;

    try {
			msg = "Valor do limiar: ";
			String sLimiar = JOptionPane.showInputDialog(null, msg);
			if (sLimiar != null)
				dLimiar = Double.parseDouble(sLimiar);
			else {
				dLimiar = 255.0;
				msg = "Assumindo limiar de 255.0";
				String titulo = "Distncia Mnima";
				JOptionPane.showMessageDialog(desk, msg, titulo, 0);
			}
    } catch (NumberFormatException e) {
				msg = "Parmetro invlido";
				String titulo = "Distncia Mnima";
				JOptionPane.showMessageDialog(desk, msg, titulo, 0);
		}

		try {
			vetorMedBanda = new Vector();
			vetorMedia = new Vector();
			aFile = arquivoAssinaturas.getName();
			estatFile = new FileReader(aFile);
			BufferedReader br = new BufferedReader(estatFile);

			// L a primeira linha do arquivo e pega nmero de temas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numTemas = Integer.parseInt(st.nextToken());

			// L a segunda linha do arquivo e pega nmero de bandas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numBandas = Integer.parseInt(st.nextToken());
			sTemas = new String[numTemas];

			// Percorre todos os temas
			//
			for (int nt = 0; nt < numTemas; nt++) {

				// L linha com nome do tema
				//
				line = br.readLine();
				st = new StringTokenizer(line);
				sTemas[nt] = st.nextToken();
				//System.out.println("\nTema: " + sTemas[nt]);

				// Percorre todas as bandas
				//
				for (int nb = 0; nb < numBandas; nb++) {

					// L linha com nome da banda
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					nomeBanda = st.nextToken();
					//System.out.println("\nBanda: " + nomeBanda);

					// L linha com valor mnimo do tema na banda
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					minimo = Integer.parseInt(st.nextToken());
					//System.out.println("minimo: " + minimo);

					// L linha com valor mximo do tema na banda
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					maximo = Integer.parseInt(st.nextToken());
					//System.out.println("maximo: " + maximo);

					// L linha com mdia do tema na banda
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					media = Double.parseDouble(st.nextToken());
					vetorMedia.addElement(new Double(media));
					//System.out.println("media: " + media);

					// L linha com valor de varincia
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					var = Double.parseDouble(st.nextToken());
					//System.out.println("var: " + var);

					// L linha com valor do desvio padro
					//
					line = br.readLine();
					st = new StringTokenizer(line);
				} // fim bandas

				// L linhas com nome e valores das covarincias
				//
				line = br.readLine();
				st = new StringTokenizer(line);
				covar = st.nextToken();
				//System.out.println("==> " + covar);
				aux = 0;
				for (int i = 0; i < numBandas - 1; i++)
					aux += (i + 1);
				for (int i = 0; i < aux; i++) {
					line = br.readLine();
					st = new StringTokenizer(line);
					cov = Double.parseDouble(st.nextToken());
					//System.out.println(cov);
				}
				vetorMedBanda.addElement(vetorMedia);
				vetorMedia = new Vector();
			}	// fim temas
    } catch (IOException e) {
				msg = "classificaImagem: problema lendo: " + arquivoAssinaturas;
				String titulo = "Distncia Mnima";
				JOptionPane.showMessageDialog(desk, msg, titulo, 0);
    }

		// Escolhe mtrica de clculo de distncia mnima
		//
		final JFrame frame = new JFrame("Mtrica de distncia");
		final JPanel metricaPanel = new JPanel();
		final ButtonGroup metricaGroup = new ButtonGroup();
		final double fdLimiar = dLimiar;
		JRadioButton radioBtn;
		metricaPanel.add(radioBtn = new JRadioButton("Euclidiana", true));
		radioBtn.setActionCommand("Euclidiana");
		metricaGroup.add(radioBtn);
		metricaPanel.add(radioBtn = new JRadioButton("Bloco"));
		radioBtn.setActionCommand("Bloco");
		metricaGroup.add(radioBtn);
		
		JPanel btnPanel = new JPanel();
		JButton okBtn = new JButton("OK");
		btnPanel.add(okBtn);
		JButton cancelBtn = new JButton("Cancela");
		btnPanel.add(cancelBtn);

		Container cPaineis = frame.getContentPane();
		cPaineis.setLayout(new GridLayout(3, 1));
		cPaineis.add(metricaPanel);
		cPaineis.add(btnPanel);

		okBtn.addActionListener(new ActionListener() {
			String sMet;
			public void actionPerformed(ActionEvent ae) {
				String sMetrica = metricaGroup.getSelection().getActionCommand();
				Component[] cMetrica = metricaPanel.getComponents();
				for (int i = 0; i < cMetrica.length; i++)	{
					JRadioButton btnMetrica = (JRadioButton)cMetrica[i];
					if (btnMetrica.isSelected()) {
						sMet = btnMetrica.getText();
					}
				}
				frame.setVisible(false);
				System.out.println("metrica = "+ sMet);
				if (sMet.equals("Euclidiana"))
					calculaRegraEuclidiana(fdLimiar);
				else if (sMet.equals("Bloco"))
					calculaRegraCityBlock(fdLimiar);
			}
		});

		cancelBtn.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				frame.setVisible(false);
			}
		});		
		frame.setSize(200, 200);
		frame.setVisible(true);
	}
	
	// MTODOS PRIVADOS DA CLASSE

	// Este mtodo usa regras de deciso do mtodo da distncia
	// mnima da mdia, mtrica Euclidiana, para decidir para qual tema 
	// (classe) um pixel pertence.
	//
	private void calculaRegraEuclidiana(double limiar) {
		int w, h, nb, nt, tema, pixel[], tipo, contaClasse[], percent;
		double media[][], dif2[][], soma[], distancia[], minDist;
		String arquivo = null;
		Vector vecMedia;
		Raster[] srcR;
		
		// Variveis usadas para gravar arquivo .csv
		//
		String[] cores = {"Vermelho", "Verde", "Azul", "Rosa", 
											"Turquesa", "Amarelo", "Violeta"};
		String ln, pV = ";", grupo, cor, pixels, porcento;
		FileWriter writer;
		BufferedWriter buf_writer; 
		String arquivoCSV = null;
				
		// Pega tempo inicial
		//
		long e_time = System.currentTimeMillis();

		nb = banda.length;
		nt = sTemas.length;
		w = banda[0].getWidth();
		h = banda[0].getHeight();
		media = new double[nt][nb];
		dif2 = new double[nt][nb];
		soma = new double[nt];
		distancia = new double[nt];
		contaClasse = new int[nt];
		pixel = new int[nb];
		srcR = new Raster[nb];

		// Prepara imagem para gravar classificao.
		//
		tipo = BufferedImage.TYPE_3BYTE_BGR;
		BufferedImage imagemClasses = null;
		imagemClasses = new BufferedImage(w, h, tipo); 
		WritableRaster outWR = imagemClasses.getRaster(); 

		vecMedia = new Vector();
    System.out.println("DistanciaMinima::regras: #bandas = " 
												+ nb + ", #temas = " + nt);

		// L os valores armazenados em Vector para array,
		// fazendo a devida converso de valores.
		//
		for (int t = 0; t < nt; t++) {
			vecMedia = (Vector)vetorMedBanda.elementAt(t);
		  //System.out.println("Tema "+(i+1));
			for (int b = 0; b < nb; b++) {
				Double med  = (Double)vecMedia.elementAt(b);
				media[t][b] = med.doubleValue();
			  /*System.out.println("Banda["+(b+1)+"]: media = "
											+ media[t][b]);*/
			}
		}
		
		for (int b = 0; b < nb; b++)
			srcR[b] = banda[b].getRaster();

		// Agora que temos os valores das mdias,
		// podemos aplicar a regra de deciso para cada pixel.
		//
		// Esta implementao usa a mtrica Euclidiana para o clculo da
		// distncia mnima da mdia
		//
		for (int lin = 0; lin < h; lin++) {
			for (int col = 0; col < w; col++) {

				// Acessa pixel das bandas.
				//
				for (int b = 0; b < nb; b++) 
					pixel[b] = srcR[b].getSample(col, lin, 0);

				// Acessa mdia do tema em cada banda.
				//
				for (int t = 0; t < nt; t++)
					for (int b = 0; b < nb; b++) 
						dif2[t][b] = (pixel[b] - media[t][b])*(pixel[b] - media[t][b]);
				for (int t = 0; t < nt; t++)
					soma[t] = 0.0;
				for (int t = 0; t < nt; t++)
					for (int b = 0; b < nb; b++) 
						soma[t] += dif2[t][b];
				//
				// Clculo da mtrica Euclidiana
				//
				minDist = Math.sqrt(soma[0]);
				for (int t = 0; t < nt; t++) {
					distancia[t] = Math.sqrt(soma[t]);	// mtrica Euclidiana
					minDist = Math.min(distancia[t], minDist);
				}
				tema = 0;
				for (int t = 0; t < nt; t++) {
					if (minDist == distancia[t])
						tema = t;
				}	
				
				// Grava pixel classificado como classe 't'.
				//
				if (minDist < limiar)
					gravaClasse(col, lin, tema, outWR);

				// Conta pixels da classe 't'.
				//
				contaClasse[tema]++;
			}	//col
		}	//lin

		// Grava imagem classificada.
		//
		try {
			String search = ".";
			int i = arquivoImagem.indexOf(search);
			arquivo = arquivoImagem.substring(0, i);
			arquivoCSV = arquivo;
			arquivo += "_minDistE.png";
			System.out.println("\nImagem classificada: '" + arquivo);
			ImageIO.write(imagemClasses, "png", new File(arquivo));
		} catch (IOException e) {
			String msg = "gravaImagem: problema gravando arquivo: " + arquivo;
			String titulo = "Distncia Mnima";
			JOptionPane.showMessageDialog(desk, msg, titulo, 0);
		}
		System.out.println("\nDistanciaMinima::Pixels classificado por classe:");

		// Grava arquivo .csv
		//
		arquivoCSV += "_" + String.valueOf(nt) + "_minDistE" + ".csv";
		try {
				writer = new FileWriter(arquivoCSV);
				buf_writer = new BufferedWriter(writer);
				for (int t = 0; t < nt; t++) {
					System.out.print("\nClasse '" + sTemas[t] + "' = " + contaClasse[t]);
					percent = (contaClasse[t]*100)/(w*h);
					System.out.println(" ("+percent+"%)");
					ln = "";
					grupo = "Classe " + String.valueOf(t + 1);
					cor = cores[t];
					pixels = String.valueOf(contaClasse[t]);
					porcento = String.valueOf(percent);
					ln = grupo + pV + sTemas[t] + pV + cor + pV + pixels + pV + porcento;
					grava_linha(ln, buf_writer);
				}
				buf_writer.close();
				System.out.println("\nCriando arquivo: " + arquivoCSV + "\n");
		} catch (IOException e) {
				String msg = "Erro na gravao do arquivo " + arquivoCSV;
				String titulo = "Distncia Mnima";
				JOptionPane.showMessageDialog(desk, msg, titulo, 0);
		}

		// Pega tempo final
		//
	  e_time = System.currentTimeMillis() - e_time;
	  DecimalFormat df = new DecimalFormat("0.00");
	  System.out.println("Tempo de execucao: " + df.format(e_time) + " milisseg.");
	} // calculaRegraEuclidiana

	// Este mtodo usa regras de deciso do mtodo da distncia
	// mnima da mdia, mtrica da quadra (city block), para decidir para qual
	// tema (classe) um pixel pertence.
	//
	private void calculaRegraCityBlock(double limiar) {
		int w, h, nb, nt, tema, pixel[], tipo, contaClasse[], percent;
		double media[][], dif2[][], soma[], distancia[], minDist;
		String arquivo = null;
		Vector vecMedia;
		Raster[] srcR;
		
		// Variveis usadas para gravar arquivo .csv
		//
		String[] cores = {"Vermelho", "Verde", "Azul", "Rosa", 
											"Turquesa", "Amarelo", "Violeta"};
		String ln, pV = ";", grupo, cor, pixels, porcento;
		FileWriter writer;
		BufferedWriter buf_writer; 
		String arquivoCSV = null;
				
		// Pega tempo inicial
		//
		long e_time = System.currentTimeMillis();

		nb = banda.length;
		nt = sTemas.length;
		w = banda[0].getWidth();
		h = banda[0].getHeight();
		media = new double[nt][nb];
		dif2 = new double[nt][nb];
		soma = new double[nt];
		distancia = new double[nt];
		contaClasse = new int[nt];
		pixel = new int[nb];
		srcR = new Raster[nb];

		// Prepara imagem para gravar classificao.
		//
		tipo = BufferedImage.TYPE_3BYTE_BGR;
		BufferedImage imagemClasses = null;
		imagemClasses = new BufferedImage(w, h, tipo); 
		WritableRaster outWR = imagemClasses.getRaster(); 

		vecMedia = new Vector();
    System.out.println("DistanciaMinima::regras: #bandas = " 
												+ nb + ", #temas = " + nt);

		// L os valores armazenados em Vector para array,
		// fazendo a devida converso de valores.
		//
		for (int t = 0; t < nt; t++) {
			vecMedia = (Vector)vetorMedBanda.elementAt(t);
		  //System.out.println("Tema "+(i+1));
			for (int b = 0; b < nb; b++) {
				Double med  = (Double)vecMedia.elementAt(b);
				media[t][b] = med.doubleValue();
			  /*System.out.println("Banda["+(b+1)+"]: media = "
											+ media[t][b]);*/
			}
		}
		
		for (int b = 0; b < nb; b++)
			srcR[b] = banda[b].getRaster();

		// Agora que temos os valores das mdias,
		// podemos aplicar a regra de deciso para cada pixel.
		//
		// Esta implementao usa a mtrica de quadra para o clculo da
		// distncia mnima da mdia
		//
		for (int lin = 0; lin < h; lin++) {
			for (int col = 0; col < w; col++) {

				// Acessa pixel das bandas.
				//
				for (int b = 0; b < nb; b++) 
					pixel[b] = srcR[b].getSample(col, lin, 0);

				// Acessa mdia do tema em cada banda.
				//
				for (int t = 0; t < nt; t++)
					for (int b = 0; b < nb; b++) 
						dif2[t][b] = Math.abs((pixel[b] - media[t][b]));
				for (int t = 0; t < nt; t++)
					soma[t] = 0.0;
				for (int t = 0; t < nt; t++)
					for (int b = 0; b < nb; b++) 
						soma[t] += dif2[t][b];
				//
				// Clculo da mtrica da quadra
				//
				minDist = soma[0];
				for (int t = 0; t < nt; t++) {
					distancia[t] = soma[t];	// mtrica da quadra
					minDist = Math.min(distancia[t], minDist);
				}
				tema = 0;
				for (int t = 0; t < nt; t++) {
					if (minDist == distancia[t])
						tema = t;
				}

				// Grava pixel classificado como classe 't'.
				//
				if (minDist < limiar)
					gravaClasse(col, lin, tema, outWR);

				// Conta pixels da classe 't'.
				//
				contaClasse[tema]++;
			}	//col
		}	//lin

		// Grava imagem classificada.
		//
		try {
			String search = ".";
			int i = arquivoImagem.indexOf(search);
			arquivo = arquivoImagem.substring(0, i);
			arquivoCSV = arquivo;
			arquivo += "_minDistCB.png";
			System.out.println("\nImagem classificada: '" + arquivo);
			ImageIO.write(imagemClasses, "png", new File(arquivo));
		} catch (IOException e) {
			String msg = "gravaImagem: problema gravando arquivo: " + arquivo;
			String titulo = "Distncia Mnima";
			JOptionPane.showMessageDialog(desk, msg, titulo, 0);
		}
		System.out.println("\nDistanciaMinima::Pixels classificado por classe:");

		// Grava arquivo .csv
		//
		arquivoCSV += "_" + String.valueOf(nt) + "_minDistCB" + ".csv";
		try {
				writer = new FileWriter(arquivoCSV);
				buf_writer = new BufferedWriter(writer);
				for (int t = 0; t < nt; t++) {
					System.out.print("\nClasse '" + sTemas[t] + "' = " + contaClasse[t]);
					percent = (contaClasse[t]*100)/(w*h);
					System.out.println(" ("+percent+"%)");
					ln = "";
					grupo = "Classe " + String.valueOf(t + 1);
					cor = cores[t];
					pixels = String.valueOf(contaClasse[t]);
					porcento = String.valueOf(percent);
					ln = grupo + pV + sTemas[t] + pV + cor + pV + pixels + pV + porcento;
					grava_linha(ln, buf_writer);
				}
				buf_writer.close();
				System.out.println("\nCriando arquivo: " + arquivoCSV + "\n");
		} catch (IOException e) {
				String msg = "Erro na gravao do arquivo " + arquivoCSV;
				String titulo = "Distncia Mnima";
				JOptionPane.showMessageDialog(desk, msg, titulo, 0);
		}

		// Pega tempo final
		//
	  e_time = System.currentTimeMillis() - e_time;
	  DecimalFormat df = new DecimalFormat("0.00");
	  System.out.println("Tempo de execucao: " + df.format(e_time) + " milisseg.");

	} // calculaRegraCityBlock

	private void gravaClasse(int col, int lin, int tema,
		WritableRaster outWR) {
		switch (tema) {
			case 0:
				outWR.setSample(col, lin, 0, 255);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 0);
				break;
			case 1:
				outWR.setSample(col, lin, 0, 0);
				outWR.setSample(col, lin, 1, 255);
				outWR.setSample(col, lin, 2, 0);
				break;
			case 2:
				outWR.setSample(col, lin, 0, 0);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 255);
				break;
			case 3:
				outWR.setSample(col, lin, 0, 255);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 255);
				break;
			case 4:
				outWR.setSample(col, lin, 0, 0);
				outWR.setSample(col, lin, 1, 255);
				outWR.setSample(col, lin, 2, 255);
				break;
			case 5:
				outWR.setSample(col, lin, 0, 255);
				outWR.setSample(col, lin, 1, 255);
				outWR.setSample(col, lin, 2, 0);
				break;
			case 6:
				outWR.setSample(col, lin, 0, 128);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 255);
				break;
		}
	}

	private void grava_linha(String texto, BufferedWriter buf_writer) {
		try	{
			buf_writer.write(texto);
			buf_writer.newLine();		
		} catch (IOException e) {
			String msg = "Erro na gravao da linha " + texto;
			String titulo = "Distncia Mnima";
			JOptionPane.showMessageDialog(desk, msg, titulo, 0);
		} 
	}
}
