/******************************************************************************
FalsaCor.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

 Classe chamada por EscolheBandas.
 Gera imagem sinttica em falsa cor.
 Antes, faz realce nas bandas usando a equalizao de histograma.
  
Entrada:

 Trs bandas espectrais, escolhidas na classe EscolheBandas.

Sada:

 Uma imagem em falsa cor.
 A imagem gerada  um mapeamento da ordem de entrada das bandas
 nos canais R, G e B.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/

package pID;

// pacotes genricos
import java.io.File;
import java.io.IOException;

// pacotes do AWT
import java.awt.image.Raster;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;

// pacote Swing para interface grfica
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;

// leitura em modo imediato - para J2SE 1.4+
import javax.imageio.ImageIO;

public class FalsaCor {

	// CONSTRUTOR
	//
  public FalsaCor(String[] path, String[] arquivos) {

		// Define objeto BufferedImage para encapsular as imagens
		//
    BufferedImage src1 = null, src2 = null, src3 = null, dest = null;
    int nc1, nc2, nc3, w1, h1, w2, h2, w3, h3, tipo;
		String aFile1 = null, aFile2 = null, aFile3 = null;

    System.out.println("\nFalsaCor::bandas escolhidas:");
    System.out.println("FalsaCor::length:" + arquivos.length);
		for (int i = 0; i < arquivos.length; i++) {
	    System.out.println("banda = " + arquivos[i]);
		}

		if (arquivos == null) {
			String msg = "Abrir arquivo imagem!";
			String titulo = "Falsa cor";
			JOptionPane.showMessageDialog(null, msg, titulo, 0);
		} else {
			aFile1 = path[0] + "\\" + arquivos[0];
			aFile2 = path[1] + "\\" + arquivos[1];
			aFile3 = path[2] + "\\" + arquivos[2];

	    // Certifica que as imagens existem
			//
		  File file1 = new File(aFile1);
			try {
				src1 = ImageIO.read(file1);
			} catch(Exception e) {
				System.out.println("Falsa cor::Imagem '" + aFile1 + "' nao existe.");
				System.exit(0);
			}
			File file2 = new File(aFile2);
			try {
				src2 = ImageIO.read(file2);
			} catch(Exception e) {
				System.out.println("Falsa cor::Imagem '" + aFile2 + "' nao existe.");
				System.exit(0);
			}
			File file3 = new File(aFile3);
			try {
				src3 = ImageIO.read(file3);
			} catch(Exception e) {
				System.out.println("Falsa cor::Imagem '" + aFile3 + "' nao existe.");
				System.exit(0);
			}

			// Certifica que imagens tm mesma dimenso
			//
			w1 = src1.getWidth();
			h1 = src1.getHeight();
			w2 = src2.getWidth();
			h2 = src2.getHeight();
			w3 = src3.getWidth();
			h3 = src3.getHeight();
			if ((w1 != w2) || (w1 != w3) || (w2 != w3)) {
				System.out.println("Falsa cor::Imagens com larguras diferentes.");
				System.exit(0);
			}
			if ((h1 != h2) || (h1 != h3) || (h2 != h3)) {
				System.out.println("Falsa cor::Imagens com alturas diferentes.");
				System.exit(0);
			}

			// Antes de fazer composio colorida, faz realce nas bandas
			//
			src1 = equaliza(src1);
			src2 = equaliza(src2);
			src3 = equaliza(src3);

			// Realiza composio colorida
			//
			tipo = BufferedImage.TYPE_3BYTE_BGR;
			dest = new BufferedImage(w1, h1, tipo);
			Raster srcR1 = src1.getRaster();
			Raster srcR2 = src2.getRaster();
			Raster srcR3 = src3.getRaster();
			WritableRaster destWR = dest.getRaster();

			for (int i = 0; i < h1; i++) {
					for (int j = 0; j < w1; j++) {
							nc1 = srcR1.getSample(j, i, 0);
							nc2 = srcR2.getSample(j, i, 0);
							nc3 = srcR3.getSample(j, i, 0);
							destWR.setSample(j, i, 0, nc1);
							destWR.setSample(j, i, 1, nc2);
							destWR.setSample(j, i, 2, nc3);
					}
			}

      // Grava imagem.
			//
      boolean status = saveFile(dest);
      if (!status) {
				String msg = "Erro de IO no arquivo!!";
				String titulo = "Salva arquivo";
        JOptionPane.showMessageDialog (
              null, msg, titulo, JOptionPane.ERROR_MESSAGE);
			}
		}	// else arquivos == null
  }		// fim construtor

	// MTODOS PRIVADOS

	//
  // Transformao no linear - Equalizao de histograma
  //
  private BufferedImage equaliza(BufferedImage src) {
		int dim, histoEq[] = new int[256], hist[] = new int[256], tipo, 
				w, h, pixelOut, nc;
    double soma = 0.0;

		Raster srcR = src.getRaster();
    w = src.getWidth();
    h = src.getHeight();

		// Calcula histograma
		//
    for (int i = 0; i < h; i++) {
			for (int j = 0; j < w; j++) {
				nc = srcR.getSample(j, i, 0);
        hist[nc]++;
      }
    }
    
		dim = w*h;
    for (int i = 0; i < 256; i++) {
				soma = 0.0;
				for (int j = 0; j <= i; j++) {
          soma += (double)hist[j]/dim;
        }
        histoEq[i] = (int)(255*soma + .5);
    }
			
    tipo = BufferedImage.TYPE_BYTE_GRAY;
    BufferedImage eqlImage = new BufferedImage(w, h, tipo);
    WritableRaster eqlWR = eqlImage.getRaster();

		for (int i = 0; i < h; i++) {
				for (int j = 0; j < w; j++) {
					nc = srcR.getSample(j, i, 0);
          pixelOut = histoEq[nc];
					eqlWR.setSample(j, i, 0, pixelOut);
				}
    }
		return eqlImage;
	}

	//
	// Salva imagem falsa cor
	//
	private boolean saveFile(BufferedImage dest) {
		File file = null;
    JFileChooser fc;
		String nome;
		int result;

    fc = new JFileChooser();
    fc.setCurrentDirectory(new File("."));
		fc.setDialogTitle("Salva arquivo falsa cor (*.png)");
		fc.addChoosableFileFilter(new TipoArquivo(".png"));

    // Define nome inicial
		//
		File fFile = new File ("falsacor.png");
    fc.setSelectedFile(fFile);

    // Abre dialog
		//
    result = fc.showSaveDialog(null);

    if (result == JFileChooser.CANCEL_OPTION) {
			return true;
    } else if (result == JFileChooser.APPROVE_OPTION) {
				fFile = fc.getSelectedFile();
        if (fFile.exists()) {
					int response = JOptionPane.showConfirmDialog (null,
										"Atualiza arquivo existente?","Confirma Atualizao",
										JOptionPane.OK_CANCEL_OPTION,
										JOptionPane.QUESTION_MESSAGE);
          if (response == JOptionPane.CANCEL_OPTION)
						return false;
        }
				try {
						nome = fFile.getName();
						System.out.println("\nGravou arquivo -> " + nome + "\n");
						ImageIO.write(dest, "png", new File(nome));
				} catch (IOException e) {
						System.out.println("Problema gravando arquivo. ");
						System.exit(0);
				}
    } else {
				return false;
    }
		return true;
  } // saveFile
}
