/******************************************************************************
Histograma.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

 Calcula o histograma de uma imagem.
 Desenha o histograma usando a biblioteca JFreeChart, verso 0.9.8
 Se uma verso mais nova do JFreeChart for usada, algumas funes
 podem no funcionar.
  
Entrada:

 Vetor com pixels da imagem.

Sada:

 Painel com histograma.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/

package pID;

// Pacotes do AWT
import java.awt.image.BufferedImage;
import java.awt.GradientPaint;
import java.awt.Graphics2D;
import java.awt.Color;
import java.awt.Paint;

// Pacotes do JFreeChart
import org.jfree.data.XYSeriesCollection;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartFrame;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.TextTitle;
import org.jfree.chart.plot.Plot;
import org.jfree.data.XYDataset;
import org.jfree.data.XYSeries;

// Pacotes do Swing
import javax.swing.JInternalFrame;
import javax.swing.JDesktopPane;
import javax.swing.JScrollPane;
import javax.swing.ImageIcon;
import javax.swing.JLabel;

public class Histograma {

	// ATRIBUTOS

	// Array para armazenar a freqncia de valores dos pixels.
	//
	private int[] histo = new int[256];

	// CONSTRUTOR DA CLASSE

	// int[] src -> vetor contendo os pixels da imagem.
	// Os valores dos pixels esto no intervalo inteiro [0, 255].
	//
	public Histograma(int[] src) {
		int N;		// Quantidade de pixels na imagem

		N = src.length;
		for (int i = 0; i < N; i++) {
			histo[src[i]]++;
		}
	}

	// MTODOS PBLICOS

	// Retorna a freqncia do nvel de cinza 'valor'
	//
	public int retornaFrequencia(int valor) {

		// Certifica se 'valor' est no intervalo [0, 255]
		//
		if ((valor >= 0) && (valor < 256))	{
			return histo[valor];
		} else
			return 0;
	}

	// Retorna a freqncia acumulada at o nvel de cinza 'valor'
	//
	public long retornaFrequenciaAcumulada(int valor) {

		long acumulado = 0;

		// Certifica se 'valor' est no intervalo [0, 255]
		//
		if ((valor >= 0) && (valor < 256))	{
			for (int i = 0; i <= valor; i++)
				acumulado += histo[i];
		} else
			acumulado = 0;
		return acumulado;
	}

	// Mostra o histograma
	//
	public void mostraHistograma(JDesktopPane desktop,
															 String nomeArquivoImagem) {

		JInternalFrame jiFrame = new JInternalFrame("Histo",
													          true, //resizable
													          true, //closable
													          true, //maximizable
													          true);//iconifiable

		jiFrame.setTitle("Histograma");
		jiFrame.setSize(300, 300);
	  jiFrame.setVisible(true);
    desktop.add(jiFrame);
	
		BufferedImage histo = criaImagemHisto(nomeArquivoImagem);
		JLabel lsrc = new JLabel(new ImageIcon(histo));
		jiFrame.getContentPane().add(new JScrollPane(lsrc));
	}

	// MTODOS PRIVADOS

	private BufferedImage criaImagemHisto(String nomeArquivoImagem) {
					
		BufferedImage bi = null;
		int tipo, x, y, lasty;

		tipo = BufferedImage.TYPE_BYTE_GRAY;
		bi = new BufferedImage(300, 300, tipo);

		// Define contexto grfico para colocar desenho da imagem
		//
		Graphics2D g2 = bi.createGraphics();

		// Cria o arquivo de dados...	
		XYSeries dataset = new XYSeries("Densidade probabilidade");
		for (int i = 0; i < 256; i++) {
			dataset.add(i, histo[i]);
		}

		XYDataset xyDataSet = new XYSeriesCollection(dataset);

		// Cria o grfico...
		//
		JFreeChart chart = ChartFactory.createAreaXYChart("Histograma",
												"nvel de cinza",
												"Freqncia",
												xyDataSet,
												false,				// mostra a legenda FDP
												true,
												true);

		// Cria e mostra um frame...
		//
		ChartFrame jiFrame = new ChartFrame("Histograma", chart);

		// Personaliza apresentao do histograma
		// Adiciona sub-ttulo
		//
		TextTitle subtitulo = new TextTitle(nomeArquivoImagem);
		chart.addSubtitle(subtitulo);

		// Cor de fundo do quadro 
		//
		Paint pq = new GradientPaint(0, 0, Color.white, 1000, 0, Color.blue);
		chart.setBackgroundPaint(pq);

		// Cor de fundo da rea do grfico
		//
		Plot plot = chart.getPlot();
		Paint pc = new GradientPaint(0, 0, Color.green, 1000, 0, Color.white);
		plot.setBackgroundPaint(pc);

		bi = chart.createBufferedImage(300, 300);

		return bi;
	}

}
