/******************************************************************************
Imagem.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

	Define a classe principal para processamento de imagens digitais:
	o objetivo  tornar disponvel os pixels de um arquivo imagem no
	vetor "vetorPixels[]"
  
Entrada:

	Arquivo com a imagem a ser usada para anlise.

Sada:

	Vetor com os pixels da imagem.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/

package pID;

// Pacote genrico
import java.io.File;

// Pacotes do AWT
import java.awt.image.BufferedImage;
import java.awt.image.Raster;

// Pacotes do Swing
import javax.swing.JInternalFrame;
import javax.swing.JFileChooser;
import javax.swing.JDesktopPane;
import javax.swing.JScrollPane;
import javax.swing.ImageIcon;
import javax.swing.JLabel;

// Leitura em modo imediato - para J2SE 1.4+
import javax.imageio.ImageIO;

public class Imagem {

	// ATRIBUTOS

	// Array de inteiros para armazenar os pixels da imagem
	//
	private int vetorPixels[];

	// altura e largura da imagem
	//
	private int altura, largura;

	// Nmero de bandas da imagem
	//
	private int bandas;

	// Nome do arquivo imagem
	//
	File arquivoImagem;

	// Nome da BufferedImage
	//
	BufferedImage biImagem = null;

	// CONSTRUTOR DA CLASSE

	//	L o arquivo imagem da lista de parmetro e disponibiliza os valores dos
	//	pixels no array int[] pixels
	//
	public Imagem(){

		JFileChooser fileChooser = new JFileChooser(".");
		fileChooser.setDialogTitle("Abre imagem");
		fileChooser.addChoosableFileFilter(new TipoImagem());
    int status = fileChooser.showOpenDialog(null);
    if (status == JFileChooser.APPROVE_OPTION) {
			arquivoImagem = fileChooser.getSelectedFile();
			abreImagem();
			altura = biImagem.getHeight();
			largura = biImagem.getWidth();
			matrizDePixels();
			System.out.println("Arquivo: " + arquivoImagem.getParent() 
														+ " --- " + arquivoImagem.getName());
			System.out.println("#bandas = " + bandas);
		} else 
			if (status == JFileChooser.CANCEL_OPTION)
				arquivoImagem = null;
	}

	// MTODOS PBLICOS DA CLASSE

	// Retorna banda como um vetor coluna (uma varivel)
	//
	public int[] retornaVetorPixels() {
		return vetorPixels;
	}

	public int retornaLargura() {
		return largura;
	}	

	public int retornaAltura() {
		return altura;
	}	

	public int retornaBandas() {
		return bandas;
	}	

	public String retornaNomeArquivoImagem() {
		if (arquivoImagem != null) {
			return arquivoImagem.getName();
		} else
			return null;
	}	

	public String retornaPathArquivoImagem() {
		if (arquivoImagem != null) {
			return arquivoImagem.getParent();
		} else
			return null;
	}	

	public void mostraImagem(JDesktopPane desktop) {

		JInternalFrame frame = new JInternalFrame("Internal Frame",
													          true, //resizable
													          true, //closable
													          true, //maximizable
													          true);//iconifiable
		desktop.add(frame);

		// Mostra a imagem
		//
    JLabel lsrc = new JLabel(new ImageIcon(biImagem));
    frame.getContentPane().add(new JScrollPane(lsrc));
		frame.setSize(largura, altura);
		frame.setTitle(arquivoImagem.getName());
    frame.setVisible(true);
	}

	// MTODOS PRIVADOS DA CLASSE

	// Abre arquivo imagem
	//
	private void abreImagem() {
		try {
			biImagem = ImageIO.read(arquivoImagem);
			bandas = biImagem.getSampleModel().getNumBands();
	  } catch(Exception e) {
			biImagem = null;
	    System.out.println("Problema lendo '" + arquivoImagem + "'.");
	  }
	}

	// L os valores de pixels do arquivo imagem e os armazena numa matriz
	//
	private void matrizDePixels() {

		int indice;
		Raster srcR = biImagem.getRaster();
		vetorPixels = new int[largura*altura];

	  for (int i = 0; i < altura; i++)
			for (int j = 0; j < largura; j++) {
				indice = i*largura + j;
				vetorPixels[indice] = srcR.getSample(j, i, 0);
			}
	}

}
