/******************************************************************************
MaxVer.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

	Classifica imagem pelo mtodo da mxima verossimilhana.
  
Entrada:

	Arquivo .SIG, com informaes estatsticas e nome dos arquivos com as 
	bandas a serem classificadas.
	BufferedImage das bandas selecionadas.

Sada:

	Imagem classificada gravada em disco, com os grupos.
	Conforme o mtodo 'gravaClasse', somente sete grupos so identificados.
	Para aceitar um nmero maior de classes, modificar o mtodo citado.
	Arquivo .CSV, com informaes sobre os grupos classificados.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/

package pID;

// Pacotes genricos
import java.io.File;
import java.util.Vector;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.text.DecimalFormat;
import java.util.StringTokenizer;

// Pacotes do AWT
import java.awt.image.Raster;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;

// Pacotes do Swing
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.swing.JDesktopPane;

// Leitura em modo imediato - para J2SE 1.4+
import javax.imageio.ImageIO;

public class MaxVer {

	// ATRIBUTOS

	// sTemas = guarda os temas definidos pelo usurio.
	//
	private String[] sTemas;

	// bandas - BufferedImage das bandas selecionadas.
	//
	private BufferedImage[] banda;

	// Nome do arquivo imagem
	//
	String arquivoImagem;

	// Estes atributos sero vetores de vetores para as mdias
	// e varincias por tema e banda
	//
	private Vector vetorMedBanda, vetorVarBanda, vetorCovarBanda;

	// CONSTRUTOR DA CLASSE

	public MaxVer(){ }

	// MTODOS PBLICOS DA CLASSE

	// Este mtodo apenas abre o arquivo com as assinaturas
	// espectrais (.sig) calculadas pelo objeto AssinaturaEspectral
	// com as estatsticas multivariadas de cada tema a ser classificado.
	//
	public File abreArquivoAssinaturas(JDesktopPane desk) {
		int tam;
		File arquivoAssinaturas = null;
		String aFile, sufixo;

		JFileChooser fileChooser = new JFileChooser(".");
		fileChooser.setDialogTitle("Abre arquivo com assinaturas (*.sig)");
		fileChooser.addChoosableFileFilter(new TipoArquivo(".sig"));
    int status = fileChooser.showOpenDialog(null);
    if (status == JFileChooser.APPROVE_OPTION) {
			arquivoAssinaturas = fileChooser.getSelectedFile();

			// Verifica que abriu arquivo do tipo .sig
			aFile = arquivoAssinaturas.getName();
			tam = aFile.length();
			sufixo = aFile.substring((tam-3), tam);
			System.out.println("sufixo = " + sufixo);
			if (!sufixo.equals("sig")) {
				String str = "Arquivo tem que ser do tipo .sig!";
				JOptionPane.showMessageDialog(desk, str);
				arquivoAssinaturas = null;
			}
		} else 
			if (status == JFileChooser.CANCEL_OPTION)
				arquivoAssinaturas = null;

		return arquivoAssinaturas;
	}

	// Este mtodo l e abre os arquivos das imagens, 
	// armazenando-os em BufferedImage.
	// Como os nomes das imagens aparecem de maneira repetida
	// no arquivo .sig para cada tema, precisamos de ler apenas
	// um tema para acessar os valores necessrios.
	//
	public void abreBandas(File arquivoAssinaturas) {
		int tam, numTemas, numBandas = 1;
		FileReader estatFile = null;
		String line, aFile, nomeBanda[] = null;
		StringTokenizer st = null;
		File file;

		try {
			aFile = arquivoAssinaturas.getName();
			estatFile = new FileReader(aFile);
			BufferedReader br = new BufferedReader(estatFile);

			// L a primeira linha do arquivo e pega nmero de temas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numTemas = Integer.parseInt(st.nextToken());

			// L a segunda linha do arquivo e pega nmero de bandas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numBandas = Integer.parseInt(st.nextToken());
			nomeBanda = new String[numBandas];

			// Percorre apenas um tema
			// L linha com nome do tema
			//
			line = br.readLine();
			st = new StringTokenizer(line);

			// Percorre todas as bandas
			//
			for (int nb = 0; nb < numBandas; nb++) {

				// L linha com nome da banda
				//
				line = br.readLine();
				st = new StringTokenizer(line);
				nomeBanda[nb] = st.nextToken();
				System.out.println("\nBanda: " + nomeBanda[nb]);

				// L linha com valor mnimo do tema na banda
				//
				line = br.readLine();
				st = new StringTokenizer(line);

				// L linha com valor mximo do tema na banda
				//
				line = br.readLine();
				st = new StringTokenizer(line);

				// L linha com mdia do tema na banda
				//
				line = br.readLine();
				st = new StringTokenizer(line);

				// L linha com valor de varincia
				//
				line = br.readLine();
				st = new StringTokenizer(line);

				// L linha com valor do desvio padro
				//
				line = br.readLine();
				st = new StringTokenizer(line);
			} // fim bandas
    } catch (IOException e) {
				String msg = "MaxVer::abreBandas: problema com imagem: ";
        System.out.println(msg + arquivoAssinaturas);
        System.exit(0);
    }

		arquivoImagem = nomeBanda[0];

		// Armazena pixels em BufferedImage
		//
		banda = new BufferedImage[numBandas];
		for (int i = 0; i < numBandas; i++) {
	    file = new File(nomeBanda[i]);
			try {
				banda[i] = ImageIO.read(file);
				System.out.println("banda["+i+"]: w = "+banda[i].getWidth()+
												 " h = "+banda[i].getHeight());
		  } catch(Exception e) {
		    System.out.println("MaxVer::abreImagens: Problema lendo '" +
														file + "'.");
			}
		}
	}

	// Este mtodo classifica os pixels de cada banda de acordo
	// com as assinaturas espectrais das amostras definidas pelo 
	// usurio usando o mtodo da mxima verossimilhana.
	// A regra de deciso do mtodo precisa da mdia, varincia e
	// covarincia do tema em cada classe.
	//
	public void classificaImagem(File arquivoAssinaturas) {
		int numTemas, numBandas, minimo, maximo, aux;
		double media, var, cov;
		FileReader estatFile = null;
		String line, aFile, nomeBanda, covar;
		Vector vetorBandas, vetorMedia, vetorVar, vetorCovar;
		StringTokenizer st = null;

		try {
			vetorMedBanda = new Vector();
			vetorVarBanda = new Vector();
			vetorCovarBanda = new Vector();
			vetorMedia = new Vector();
			vetorVar = new Vector();
			vetorCovar = new Vector();
			aFile = arquivoAssinaturas.getName();
			estatFile = new FileReader(aFile);
			BufferedReader br = new BufferedReader(estatFile);

			// L a primeira linha do arquivo e pega nmero de temas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numTemas = Integer.parseInt(st.nextToken());

			// L a segunda linha do arquivo e pega nmero de bandas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numBandas = Integer.parseInt(st.nextToken());
			sTemas = new String[numTemas];

			// Percorre todos os temas
			//
			for (int nt = 0; nt < numTemas; nt++) {

				// L linha com nome do tema
				//
				line = br.readLine();
				st = new StringTokenizer(line);
				sTemas[nt] = st.nextToken();

				// Percorre todas as bandas
				//
				for (int nb = 0; nb < numBandas; nb++) {

					// L linha com nome da banda
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					nomeBanda = st.nextToken();

					// L linha com valor mnimo do tema na banda
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					minimo = Integer.parseInt(st.nextToken());

					// L linha com valor mximo do tema na banda
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					maximo = Integer.parseInt(st.nextToken());

					// L linha com mdia do tema na banda
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					media = Double.parseDouble(st.nextToken());
					vetorMedia.addElement(new Double(media));

					// L linha com valor de varincia
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					var = Double.parseDouble(st.nextToken());
					vetorVar.addElement(new Double(var));

					// L linha com valor do desvio padro
					//
					line = br.readLine();
					st = new StringTokenizer(line);
				} // fim bandas

				// L linhas com nome e valores das covarincias
				//
				line = br.readLine();
				st = new StringTokenizer(line);
				covar = st.nextToken();
				aux = 0;
				for (int i = 0; i < numBandas - 1; i++)
					aux += (i + 1);
				for (int i = 0; i < aux; i++) {
					line = br.readLine();
					st = new StringTokenizer(line);
					cov = Double.parseDouble(st.nextToken());
					vetorCovar.addElement(new Double(cov));
				}
				vetorCovarBanda.addElement(vetorCovar);
				vetorMedBanda.addElement(vetorMedia);
				vetorVarBanda.addElement(vetorVar);
				vetorMedia = new Vector();
				vetorCovar = new Vector();
				vetorVar = new Vector();
			}	// fim temas
			estatFile.close();
    } catch (IOException e) {
				String msg = "MaxVer::classificaImagem: problema com imagem: ";
        System.out.println(msg + arquivoAssinaturas);
        System.exit(0);
    }

		calculaRegraDeDecisao();
	}
	
	// MTODOS PRIVADOS DA CLASSE

	// Este mtodo usa regras de deciso do mtodo da mxima 
	// verossimilhana para decidir se um pixel pertence a um 
	// tema (classe) ou no.
	// Nesse mtodo, todos os pixels so classificados.
	// No final, so gravados dois arquivos: a imagem classificada e um arquivo 
	// .csv, com as categorias, suas cores, nmeros de pixels classificados e 
	// porcentagem.
	//
	private void calculaRegraDeDecisao() {
		int nb, nt, w, h, pixel, tipo, contaClasse[], percent, aux, tema;
		double[][] media, var, covar;
		double fator1, fator2, pmax, p[];
		double[][][] a_Cov;
		String arquivo = null;
		Vector vecMedia, vecVar, vecCovar;
		Raster[] srcR;

		// Variveis usadas para gravar arquivo .csv
		//
		String[] cores = {"Vermelho", "Verde", "Azul", "Rosa", 
											"Turquesa", "Amarelo", "Violeta"};
		String ln, pV = ";", grupo, cor, pixels, porcento;
		FileWriter writer;
		BufferedWriter buf_writer;
		String arquivoCSV = null;
		
		nb = banda.length;
		nt = sTemas.length;
		w = banda[0].getWidth();
		h = banda[0].getHeight();
		media = new double[nt][nb];
		var = new double[nt][nb];
		contaClasse = new int[nt];
		p = new double[nt];
		srcR = new Raster[nb];
		vecMedia = vecVar = vecCovar = new Vector();


		// Pega tempo inicial
		//
		long e_time = System.currentTimeMillis();


		// A quantidade de valores de covarincia  a mesma
		// em cada banda. Precisamos desse valor para
		// definir a segunda dimenso do vetor de covarincias.
		//
		vecCovar = (Vector)vetorCovarBanda.elementAt(0);
		covar = new double[nt][vecCovar.size()];

		// Prepara imagem para gravar classificao.
		//
		tipo = BufferedImage.TYPE_3BYTE_BGR;
		BufferedImage imagemClasses = null;
		imagemClasses = new BufferedImage(w, h, tipo); 
		WritableRaster outWR = imagemClasses.getRaster(); 

		// L os valores armazenados em 'Vectors' para 'arrays',
		// fazendo a devida converso de valores.
		// Percorre todos os temas.
		//
		for (int t = 0; t < nt; t++) {
			vecMedia = (Vector)vetorMedBanda.elementAt(t);
			vecCovar = (Vector)vetorCovarBanda.elementAt(t);
			vecVar = (Vector)vetorVarBanda.elementAt(t);

			// Percorre todas as bandas.
			//
			for (int b = 0; b < nb; b++) {
				Double med  = (Double)vecMedia.elementAt(b);
				Double vari  = (Double)vecVar.elementAt(b);
				media[t][b] = med.doubleValue();
				var[t][b] = vari.doubleValue();
			}	// banda

			// L valores das covarincias
			//
			aux = 0;
			for (int b = 0; b < nb - 1; b++)
				aux += (b + 1);
			for (int b = 0; b < aux; b++) {
				Double cov  = (Double)vecCovar.elementAt(b);
				covar[t][b] = cov.doubleValue();
			}	// banda
		}	// tema

		// Monta matriz de varincia e covarincia
		//
		a_Cov = new double[nt][nb][nb];
		
		// Percorre todos os temas.
		//
		for (int t = 0; t < nt; t++) {

			// Monta a diagonal
			//
			for (int k = 0; k < nb; k++)
				a_Cov[t][k][k] = var[t][k];

			// Matriz triangular inferior.
			//
			aux = 0;
			for (int k = 0; k < nb-1; k++)
				for (int i = k+1; i < nb; i++) {
					a_Cov[t][i][k] = covar[t][aux++];
				}

			// Matriz triangular superior.
			//
			aux = 0;
			for (int lin = 0; lin < nb-1; lin++)
				for (int col = lin+1; col < nb; col++)
					a_Cov[t][lin][col] = covar[t][aux++];
		}

		// Imprime matriz de var-covar.
		//
		System.out.println("Matriz de var-covar");
		String padrao = "##.##";
		DecimalFormat formato = null;
		formato = new DecimalFormat(padrao);
		for (int t = 0; t < nt; t++) {
			for (int lin = 0; lin < nb; lin++){
				for (int col = 0; col < nb; col++)
					System.out.print("\t" + formato.format(a_Cov[t][lin][col]));
				System.out.print("\n");				
			}
			System.out.print("\n");
		}

		for (int b = 0; b < nb; b++)
			srcR[b] = banda[b].getRaster();

		// Agora que temos os valores das mdias e matriz de
		// covarincias, podemos aplicar a regra de deciso 
		// para cada vetor de medidas de pixels para todas as
		// bandas.

		// x = vetor de medidas de pixels.
		//
		double[][] a_x = new double[nb][1];

		// mc = vetor de mdias da classe (tema) c.
		//
		double[][] a_mc = new double[nb][1];

		// Arrays para operaes intermedirias.
		//
		double[][] a_dif = new double[nb][1];
		double[][] a_transposta = new double[1][nb];
		double[][] a_fator = new double[1][nb];
		double[][] a_multi = new double[1][1];

		// a_covar = array de var-covar.
		//
		double[][] a_covar = new double[nb][nb];
		double[][] a_inversaCovar = new double[nb][nb];

		// Cria instncia de matrizes.
		//
		Matriz m_x, m_mc, m_covar, m_dif, m_inversaCovar, 
					 m_transposta, m_fator, m_multi;

		// Define objetos Matriz.
		//
		m_x = new Matriz(a_x);
		m_mc = new Matriz(a_mc);
		m_covar = new Matriz(a_covar);
		m_dif = new Matriz(a_dif);
		m_inversaCovar = new Matriz(a_inversaCovar);
		m_transposta = new Matriz(a_transposta);
		m_fator = new Matriz(a_fator);
		m_multi = new Matriz(a_multi);

		double dpixel;

		for (int lin = 0; lin < h; lin++) {
			for (int col = 0; col < w; col++) {

				// Define vetor de medidas.
				//
				for (int b = 0; b < nb; b++) {
					dpixel = srcR[b].getSample(col, lin, 0);
					m_x.setElement(b, 0, dpixel);
				}

				// Define vetor de mdias.
				//
				for (int t = 0; t < nt; t++) {
					for (int b = 0; b < nb; b++)
							m_mc.setElement(b, 0, media[t][b]);
					m_dif = m_x.minus(m_mc);

					// Define matriz de var-covar.
					//
					for (int linha = 0; linha < nb; linha++) {
						for (int coluna = 0; coluna < nb; coluna++) {
							m_covar.setElement(linha, coluna, a_Cov[t][linha][coluna]);
						}
					}

					// Calcula primeiro fator da equao 8-22, Jensen, pg. 230 ou
					// equao 11.3 do livro PID: Mtodos Multivariados com Java.
					// -0.5*ln(det(m_covar))
					//
					fator1 = -0.5*Math.log(m_covar.determinant());

					// Calcula o segundo fator:
					// 0.5*(X - Mc)'*inveresa(Vc)*(X - Mc).
					// 1. Pega a transposta da diferena entre
					//		o vetor de medidas e o vetor de mdias.
					//
					m_transposta = m_dif.transpose();

					// 2. Acha a inversa da matriz de var-covar.
					//
					m_inversaCovar = m_covar.inverse();

					// 3. A multiplicao das trs matrizes no segundo fator
					//		ser feita usando a propriedade da associatividade
					//		da multiplicao de matrizes. Primeiro, vamos 
					//		multiplicar a transposta da diferena pela inversa
					//		da matriz var-covar; em seguida, multiplicamos esse
					//		resultado pela matriz de diferena.
					//
					m_fator = m_transposta.times(m_inversaCovar);
					m_multi = m_fator.times(m_dif);

					// Por fim, acha o resultado final do segundo fator.
					//
					fator2 = 0.5*m_multi.getElement(0, 0);

					// O resultado final  colocado no array p.
					//
					p[t] = fator1 - fator2;
				}	// tema

				// Acha o valor mximo dos p's.
				//
				pmax = p[0];
				for (int t = 0; t < nt; t++)
					pmax = Math.max(p[t], pmax);

				// Acha o tema correspondente para classificar o pixel.
				//
				tema = 0;
				for (int t = 0; t < nt; t++) {
					if (pmax == p[t])
						tema = t;
				}			

				// Grava pixel classificado como classe 't'.
				//
				gravaClasse(col, lin, tema, outWR);

				// Conta pixels da classe 't'.
				//
				contaClasse[tema]++;
			}	// col
		}	// lin

		
		// Pega tempo final
		//
	  e_time = System.currentTimeMillis() - e_time;
	  java.text.DecimalFormat df = 
					new java.text.DecimalFormat("0.00");
	  System.out.println("Tempo de execucao: " + df.format(e_time) + " milisseg.");
		
		
		/*
		System.out.println("vetor de medidas");
		m_x.imprimeMatriz();
		System.out.println("vetor de medias");
		m_mc.imprimeMatriz();
		System.out.println("vetor dif");
		m_dif.imprimeMatriz();
		System.out.println("matriz var-covar");
		m_covar.imprimeMatriz();
		//double det = m_covar.determinant();
		System.out.println("determinante da matriz var-covar: "
						+ m_covar.determinant());
		System.out.println("Inversa da matriz var-covar");
		m_inversaCovar = m_covar.inverse();
		m_inversaCovar.imprimeMatriz();

		System.out.println("Fator1");
		double fator1 = -0.5*Math.log(m_covar.determinant());
		System.out.println(fator1);
		
		System.out.println("Fator2");
		m_transposta = m_dif.transpose();
		m_fator = m_transposta.times(m_inversaCovar);
		m_multi = m_fator.times(m_dif);
		double fator2 = 0.5*m_multi.getElement(0, 0);
		System.out.println(fator2);
		
		System.out.println("'pc' da classe 1");
		System.out.println(fator1 - fator2);
		*/
		
		// Grava imagem classificada.
		//
		try {
			String search = ".";
			int i = arquivoImagem.indexOf(search);
			arquivo = arquivoImagem.substring(0, i);
			arquivoCSV = arquivo;
			arquivo += "_maxver.png";
			System.out.println("Imagem classificada: '" + arquivo);
			ImageIO.write(imagemClasses, "png", new File(arquivo));
		} catch (IOException e) {
			String msg = "MaxVer::problema gravando arquivo -> ";
			msg +=  "'" + arquivo + "'";
			System.out.println(msg);
		}

		System.out.println("\nMaxVer::Pixels classificado por classe:");

		// Grava arquivo .csv
		//
		arquivoCSV += "_" + String.valueOf(nt) + "_maxver" + ".csv";
		try {
				writer = new FileWriter(arquivoCSV);
				buf_writer = new BufferedWriter(writer);
				for (int t = 0; t < nt; t++) {
					System.out.print("\nClasse '" + sTemas[t] + "' = " + contaClasse[t]);
					percent = (contaClasse[t]*100)/(w*h);
					System.out.println(" ("+percent+"%)");
					ln = "";
					grupo = "Classe " + String.valueOf(t + 1);
					cor = cores[t];
					pixels = String.valueOf(contaClasse[t]);
					porcento = String.valueOf(percent);
					ln = grupo + pV + sTemas[t] + pV + cor + pV + pixels + pV + porcento;
					grava_linha(ln, buf_writer);
				}
				buf_writer.close();
				System.out.println("\nCriando arquivo: " + arquivoCSV + "\n");
		} catch (IOException e) {
				String msg = "MaxVer::Erro na gravao do arquivo " + arquivoCSV;
				JOptionPane.showMessageDialog(null, msg);
				System.exit(0);
		}
	}

	private void gravaClasse(int col, int lin, int tema,
		WritableRaster outWR) {
		switch (tema) {
			case 0:	// Vermelho
				outWR.setSample(col, lin, 0, 255);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 0);
				break;
			case 1:	// Verde
				outWR.setSample(col, lin, 0, 0);
				outWR.setSample(col, lin, 1, 255);
				outWR.setSample(col, lin, 2, 0);
				break;
			case 2:	// Azul
				outWR.setSample(col, lin, 0, 0);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 255);
				break;
			case 3:	// Rosa
				outWR.setSample(col, lin, 0, 255);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 255);
				break;
			case 4:	// Turquesa
				outWR.setSample(col, lin, 0, 0);
				outWR.setSample(col, lin, 1, 255);
				outWR.setSample(col, lin, 2, 255);
				break;
			case 5:	// Amarelo
				outWR.setSample(col, lin, 0, 255);
				outWR.setSample(col, lin, 1, 255);
				outWR.setSample(col, lin, 2, 0);
				break;
			case 6:	// Violeta
				outWR.setSample(col, lin, 0, 128);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 255);
				break;
		}
	}

	private void grava_linha(String texto, BufferedWriter buf_writer) {
		try	{
			buf_writer.write(texto);
			buf_writer.newLine();		
		} catch (IOException e) {
			String str = "MaxVer::Erro na gravao da linha " + texto;
			System.out.println(str);
			System.exit(0);
		} 
	}
}
