/******************************************************************************
Multivariada.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

	Calcula estatstica multivariada para imagens monocromticas.

Entrada:

	Vector com as bandas.

Sada:

	Valores estatsticos multivariados.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/

package pID;

// Pacotes genricos
import java.util.Vector;
import java.text.DecimalFormat;

// Pacotes do AWT
import java.awt.BorderLayout;
import java.awt.Font;

// Pacote Swing para interface grafica
import javax.swing.SwingConstants;
import javax.swing.JInternalFrame;
import javax.swing.JDesktopPane;
import javax.swing.JScrollPane;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JTable;

public class Multivariada {

	// ATRIBUTOS

	// Adotamos a conveno de atribuir 'a_' para identificar arrays e
	// 'm_' para identificar objetos tipo Matriz
	//
	private	int nBandas;

	// N = nmero total de pixels da banda
	//
	private int N;

	double[][] a_minimo, a_maximo, a_media, a_variancia, a_desvioPadrao, 
						 a_covariancia, a_correlacao;
	Matriz	m_covariancia, m_correlacao;

	// CONSTRUTOR DA CLASSE

  public Multivariada(Vector vecImagens) {

		// Cria variveis locais
		//
		double[][] a_somaBanda, a_soma2Banda, a_somaB1xB2, a_segundoMomento, 
							 a_somaProduto;
		int nc, nc_k, nc_k1;
		double max, min, soma, soma2, total, med, seg, var, term1, term2;

		// CRIA INSTNCIA DE MATRIZES
		//
		Matriz m_minimo, m_maximo, m_somaBanda, m_soma2Banda, m_somaB1xB2,
					 m_media, m_variancia, m_desvioPadrao, m_segundoMomento, m_somaProduto;

		// Nmero de bandas = nmero de variveis
		//
		nBandas = vecImagens.size();

		//	Array duplo para armazenar pixels da imagem:
		//	src[#bandas][N]
		//	N = src[i].length
		//
		int src[][] = new int[nBandas][];

		// Armazena valores dos pixels das bandas
		//
		for (int k = 0; k < nBandas; k++) {
			src[k] = (int[])vecImagens.elementAt(k);
		}

		// Define os arrays
		//
		a_media = new double[nBandas][1];
		a_minimo = new double[nBandas][1];
		a_maximo = new double[nBandas][1];
		a_somaB1xB2 = new double[nBandas][nBandas];
		a_somaBanda = new double[nBandas][1];
		a_variancia = new double[nBandas][1];
		a_correlacao = new double[nBandas][nBandas];
		a_soma2Banda = new double[nBandas][1];
		a_somaProduto = new double[nBandas][nBandas];
		a_covariancia = new double[nBandas][nBandas];
		a_desvioPadrao = new double[nBandas][1];
		a_segundoMomento = new double[nBandas][1];

		// Define objetos Matriz
		//
		m_media  = new Matriz(a_media);
		m_minimo  = new Matriz(a_minimo);
		m_maximo  = new Matriz(a_maximo);
		m_somaB1xB2 = new Matriz(a_somaB1xB2);
		m_somaBanda = new Matriz(a_somaBanda);
		m_variancia  = new Matriz(a_variancia);
		m_correlacao  = new Matriz(a_correlacao);
		m_soma2Banda = new Matriz(a_soma2Banda);
		m_somaProduto  = new Matriz(a_somaProduto);
		m_covariancia  = new Matriz(a_covariancia);
		m_desvioPadrao  = new Matriz(a_desvioPadrao);
		m_segundoMomento  = new Matriz(a_segundoMomento);

		// Inicia variveis
		//
		for (int k = 0; k < nBandas; k++) {
			m_minimo.setElement(k, 0, 255.0);
			m_maximo.setElement(k, 0, 0.0);
		}
		nc = 0;
		min = 255.0;
		max = 0.0;
		soma = 0.0;
		soma2 = 0.0;
		med = 0.0;
		var = 0.0;
		seg = 0.0;

		// Como as bandas tm mesma dimenso, pega valor da primeira banda
		//
		N = src[0].length;

		// Acha mximos, mnimos e acumula valores para clculos univariados
		//
		for (int k = 0; k < nBandas; k++)	{
			for (int i = 0; i < N; i++) {

				// Pega pixels da banda k
				//
				nc = src[k][i];

				// Acha mximos e mnimos
				//
				min = Math.min((double)nc, m_minimo.getElement(k, 0));
				m_minimo.setElement(k, 0, min);

				max = Math.max((double)nc, m_maximo.getElement(k, 0));
				m_maximo.setElement(k, 0, max);

				// soma - valores acumulados para calcular mdia da banda k
				// soma2 - soma ao quadrado para calcular varincia da banda k
				//
        soma += (double)nc;
        soma2 += (double)nc*(double)nc;
			}
			m_somaBanda.setElement(k, 0, soma);
			m_soma2Banda.setElement(k, 0, soma2);
			soma = 0.0;
			soma2 = 0.0;
		}

    // Calcula valores univariados da banda k
		//
		total = (double)N;
		for (int k = 0; k < nBandas; k++)	{
			m_media.setElement(k, 0, m_somaBanda.getElement(k, 0)/total);
			m_segundoMomento.setElement(k, 0, m_soma2Banda.getElement(k, 0)/total);
			med = m_media.getElement(k, 0);
			seg = m_segundoMomento.getElement(k, 0);
			m_variancia.setElement(k, 0, seg - med*med);
			var = m_variancia.getElement(k, 0);
			m_desvioPadrao.setElement(k, 0, Math.sqrt(var));
		}

		// Valores multivariados
		// Calcula soma no corrigida dos produtos, covarincia e correlao
		// Calcula primeiro a soma no corrigida dos produtos entre as bandas
		// Como a matriz resultante  simtrica, calculamos apenas a parte inferior
		//
		soma = 0.0;
		for (int k = 0; k < nBandas; k++)	{
			for (int m = k; m < nBandas; m++)	{
				for (int i = 0; i < N; i++) {

					// Pega pixels das bandas k e k+1
					//
					nc_k  = src[k][i];
					nc_k1 = src[m][i];
					soma += nc_k*nc_k1;
				}
				m_somaB1xB2.setElement(k, m, soma);

				// Atribui mesmo valor na parte inferior da matriz
				//
				m_somaB1xB2.setElement(m, k, soma);
				soma = 0.0;
			}
		}

		// Calcula a soma no corrigida dos produtos
		//
		soma = 0.0;
		term1 = 0.0;
		term2 = 0.0;
		for (int k = 0; k < nBandas; k++)	{
			for (int m = k; m < nBandas; m++)	{
				term1 = m_somaBanda.getElement(k, 0);
				term2 = m_somaBanda.getElement(m, 0);
				soma = m_somaB1xB2.getElement(k, m) - ((term1*term2)/total);
				m_somaProduto.setElement(k, m, soma);

				// Atribui mesmo valor na parte inferior da matriz
				//
				m_somaProduto.setElement(m, k, soma);
			}
		}

		// Calcula matriz de varincia-covarincia e correlao
		//
		term1 = 0.0;
		term2 = 0.0;
		soma = 0.0;
		soma2 = 0.0;
		for (int k = 0; k < nBandas; k++)	{
			for (int m = k; m < nBandas; m++)	{
				if (k == m) {
					m_covariancia.setElement(k, m, m_variancia.getElement(k, 0));
					m_correlacao.setElement(k, m, 1.0);
				}
				soma2 = m_somaProduto.getElement(k, m)/(total - 1);
				m_covariancia.setElement(k, m, soma2);

				// Calcula correlao
				//
				term1 = m_desvioPadrao.getElement(k, 0);
				term2 = m_desvioPadrao.getElement(m, 0);
				soma = m_covariancia.getElement(k, m)/(term1*term2);
				m_correlacao.setElement(k, m, soma);

				// Atribui mesmo valor na parte inferior da matriz
				//
				m_covariancia.setElement(m, k, soma2);
				m_correlacao.setElement(m, k, soma);
			}
		}

	}		// CONSTRUTOR

	// MTODOS PBLICOS

	public void mostraResultados(JDesktopPane desktop) {

		JInternalFrame jiFrame = new JInternalFrame("Multivariada",
													          true, //resizable
													          true, //closable
													          true, //maximizable
													          true);//iconifiable

		jiFrame.setTitle("Estatstica Multivariada");
		jiFrame.setSize(230, 250);
	  jiFrame.setVisible(true);
    desktop.add(jiFrame);

		String padrao = "#.###";
		DecimalFormat formato = null;
		formato = new DecimalFormat(padrao);
		String elemento;

		// Cria nomes para colunas
		//
		String columnNames[] = new String[nBandas];
		for (int k = 1; k <= nBandas; k++)
			columnNames[k-1] = "Banda" + k;

   	// Reverte de Matriz para array
		//
   	a_covariancia = m_covariancia.getArrayCopy();
   	a_correlacao = m_correlacao.getArrayCopy();

		// Insere os dados na tabela
		// Valores da matriz de covarincia
		//
		String dataValues[][] =	new String[2*nBandas][nBandas];
		for (int lin = 0; lin < nBandas; lin++)
			for (int col = 0; col < nBandas; col++) {
				if (lin >= col) {
 					elemento = formato.format(a_covariancia[lin][col]);
					dataValues[lin][col] = elemento;
				} else
					dataValues[lin][col] = "";
		}

		// Valores da matriz de correlao
		//
		for (int lin = nBandas; lin < 2*nBandas; lin++)
			for (int col = 0; col < nBandas; col++) {
				if ((lin - nBandas) >= col) {
 					elemento = formato.format(a_correlacao[lin-nBandas][col]);
					dataValues[lin][col] = elemento;
				} else
					dataValues[lin][col] = "";
			}

		// Cria JTable para mostrar os dados
		//
		JTable table = new JTable(dataValues, columnNames);

		// Adiciona a tabela em um painel com barras de rolamento
		//
		JScrollPane scrollPane = new JScrollPane(table);
		jiFrame.add(scrollPane, BorderLayout.CENTER);
		String strTitulo = "Matriz de varincia-covarincia e correlao";
		JLabel lbTitulo = new JLabel(strTitulo, SwingConstants.CENTER);
		lbTitulo.setFont(new Font("Helvetica", Font.PLAIN,12));
		jiFrame.getContentPane().add(lbTitulo, BorderLayout.NORTH);
	}

	// Retorna matriz de varincia-covarincia
	//
	public Matriz retornaVarCovar() {
		return m_covariancia;
	}

	// Retorna matriz de correlao
	//
	public Matriz retornaCorrelacao() {
		return m_correlacao;
	}

}
