/******************************************************************************
Paralelepipedo.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

	Classifica imagem pelo mtodo do paraleleppedo.
	O procedimento padro  usar um desvio padro para
	definir a rea do paraleleppedo.
  
Entrada:

	Arquivo .SIG, com informaes estatsticas e nome dos arquivos com as 
	bandas a serem classificadas.
	BufferedImage das bandas selecionadas.

Sada:

	Imagem classificada gravada em disco, com os grupos.
	Conforme o mtodo 'gravaClasse', somente sete grupos so identificados.
	Para aceitar um nmero maior de classes, modificar o mtodo citado.
	Arquivo .CSV, com informaes sobre os grupos classificados.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/

package pID;

// Pacotes genricos
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.BufferedWriter;
import java.io.BufferedReader;

import java.util.Vector;
import java.util.StringTokenizer;

// Pacotes do AWT
import java.awt.image.Raster;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;

// Pacotes do Swing
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.swing.JDesktopPane;

// Leitura em modo imediato - para J2SE 1.4+
import javax.imageio.ImageIO;

public class Paralelepipedo {

	// ATRIBUTOS

	// sTemas = guarda os temas definidos pelo usurio.
	//
	private String[] sTemas;

	// bandas - BufferedImage das bandas selecionadas.
	//
	private BufferedImage[] banda;

	// Nome do arquivo imagem
	//
	String arquivoImagem;

	// Estes atributos sero vetores de vetores para as mdias
	// e varincias por tema e banda
	//
	private Vector vetorMedBanda, vetorSTDBanda;

	// CONSTRUTOR DA CLASSE

	public Paralelepipedo(){ }

	// MTODOS PBLICOS DA CLASSE

	// Este mtodo apenas abre o arquivo com as assinaturas
	// espectrais (.sig) calculadas pelo objeto AssinaturaEspectral
	// com as estatsticas multivariadas de cada tema a ser classificado.
	//
	public File abreArquivoAssinaturas(JDesktopPane desk) {
		int tam;
		File arquivoAssinaturas = null;
		String aFile, sufixo;

		JFileChooser fileChooser = new JFileChooser(".");
		fileChooser.setDialogTitle("Abre arquivo com assinaturas (*.sig)");
		fileChooser.addChoosableFileFilter(new TipoArquivo(".sig"));
    int status = fileChooser.showOpenDialog(null);
    if (status == JFileChooser.APPROVE_OPTION) {
			arquivoAssinaturas = fileChooser.getSelectedFile();

			// Verifica que abriu arquivo do tipo .sig
			//
			aFile = arquivoAssinaturas.getName();
			tam = aFile.length();
			sufixo = aFile.substring((tam-3), tam);
			System.out.println("sufixo = " + sufixo);
			if (!sufixo.equals("sig")) {
				String str = "Arquivo tem que ser do tipo .sig!";
				JOptionPane.showMessageDialog(desk, str);
				arquivoAssinaturas = null;
			}
		} else 
			if (status == JFileChooser.CANCEL_OPTION)
				arquivoAssinaturas = null;

		return arquivoAssinaturas;
	}

	// Este mtodo l e abre os arquivos das imagens, 
	// armazenando-os em BufferedImage.
	// Como os nomes das imagens aparecem de maneira repetida
	// no arquivo .sig para cada tema, precisamos de ler apenas
	// um tema para acessar os valores necessrios.
	//
	public void abreBandas(File arquivoAssinaturas) {
		int tam, numTemas, numBandas = 1;
		FileReader estatFile = null;
		String line, aFile, nomeBanda[] = null;
		StringTokenizer st = null;
		File file;

		try {
			aFile = arquivoAssinaturas.getName();
			estatFile = new FileReader(aFile);
			BufferedReader br = new BufferedReader(estatFile);

			// L a primeira linha do arquivo e pega nmero de temas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numTemas = Integer.parseInt(st.nextToken());

			// L a segunda linha do arquivo e pega nmero de bandas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numBandas = Integer.parseInt(st.nextToken());
			nomeBanda = new String[numBandas];

			// Percorre apenas um tema
			// l linha com nome do tema
			//
			line = br.readLine();
			st = new StringTokenizer(line);

			// Percorre todas as bandas
			//
			for (int nb = 0; nb < numBandas; nb++) {

				// L linha com nome da banda
				//
				line = br.readLine();
				st = new StringTokenizer(line);
				nomeBanda[nb] = st.nextToken();
				System.out.println("\nBanda: " + nomeBanda[nb]);

				// L linha com valor mnimo do tema na banda
				//
				line = br.readLine();
				st = new StringTokenizer(line);

				// L linha com valor mximo do tema na banda
				//
				line = br.readLine();
				st = new StringTokenizer(line);

				// L linha com mdia do tema na banda
				//
				line = br.readLine();
				st = new StringTokenizer(line);

				// L linha com valor de varincia
				//
				line = br.readLine();
				st = new StringTokenizer(line);

				// L linha com valor do desvio padro
				//
				line = br.readLine();
				st = new StringTokenizer(line);
			} // fim bandas
    } catch (IOException e) {
				String msg = "Paralelepipedo::abreBandas: problema com imagem: ";
        System.out.println(msg + arquivoAssinaturas);
        System.exit(0);
    }

		arquivoImagem = nomeBanda[0];

		// Armazena pixels em BufferedImage
		//
		banda = new BufferedImage[numBandas];
		for (int i = 0; i < numBandas; i++) {
	    file = new File(nomeBanda[i]);
			try {
				banda[i] = ImageIO.read(file);
				System.out.println("banda["+i+"]: w = "+banda[i].getWidth()+
												 " h = "+banda[i].getHeight());
		  } catch(Exception e) {
		    System.out.println("Paralelepipedo::abreImagens: Problema lendo '" +
														file + "'.");
			}
		}
	}

	// Este mtodo classifica os pixels de cada banda de acordo
	// com as assinaturas espectrais das amostras definidas pelo 
	// usurio usando o mtodo do paraleleppedo.
	// A rigor, a regra de deciso do mtodo s precisa da mdia
	// do tema numa classe e o seu desvio padro, que nada mais
	//  do que a raiz quadrada da varincia.
	//
	public void classificaImagem(File arquivoAssinaturas) {
		int numTemas, numBandas, minimo, maximo, aux;
		double media, std, cov;
		FileReader estatFile = null;
		String line, aFile, nomeBanda, covar;
		Vector vetorBandas, vetorMedia, vetorSTD;
		StringTokenizer st = null;

		try {
			vetorMedBanda = new Vector();
			vetorSTDBanda = new Vector();
			vetorMedia = new Vector();
			vetorSTD = new Vector();
			aFile = arquivoAssinaturas.getName();
			estatFile = new FileReader(aFile);
			BufferedReader br = new BufferedReader(estatFile);

			// L a primeira linha do arquivo e pega nmero de temas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numTemas = Integer.parseInt(st.nextToken());

			// L a segunda linha do arquivo e pega nmero de bandas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numBandas = Integer.parseInt(st.nextToken());
			sTemas = new String[numTemas];

			// Percorre todos os temas
			//
			for (int nt = 0; nt < numTemas; nt++) {

				// L linha com nome do tema
				//
				line = br.readLine();
				st = new StringTokenizer(line);
				sTemas[nt] = st.nextToken();

				// Percorre todas as bandas
				//
				for (int nb = 0; nb < numBandas; nb++) {

					// L linha com nome da banda
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					nomeBanda = st.nextToken();

					// L linha com valor mnimo do tema na banda
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					minimo = Integer.parseInt(st.nextToken());

					// L linha com valor mximo do tema na banda
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					maximo = Integer.parseInt(st.nextToken());

					// L linha com mdia do tema na banda
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					media = Double.parseDouble(st.nextToken());
					vetorMedia.addElement(new Double(media));

					// L linha com valor de varincia
					//
					line = br.readLine();
					st = new StringTokenizer(line);

					// L linha com valor do desvio padro
					//
					line = br.readLine();
					st = new StringTokenizer(line);
					std = Double.parseDouble(st.nextToken());
					vetorSTD.addElement(new Double(std));
				} // fim bandas

				// L linhas com nome e valores das covarincias
				//
				line = br.readLine();
				st = new StringTokenizer(line);
				covar = st.nextToken();
				aux = 0;
				for (int i = 0; i < numBandas - 1; i++)
					aux += (i + 1);
				for (int i = 0; i < aux; i++) {
					line = br.readLine();
					st = new StringTokenizer(line);
					cov = Double.parseDouble(st.nextToken());
				}
				vetorMedBanda.addElement(vetorMedia);
				vetorSTDBanda.addElement(vetorSTD);
				vetorMedia = new Vector();
				vetorSTD = new Vector();
			}	// fim temas
    } catch (IOException e) {
				String msg = "Paralelepipedo::classificaImagem: problema com imagem: ";
        System.out.println(msg + arquivoAssinaturas);
        System.exit(0);
    }
		calculaRegraDeDecisao();
	}
	
	// MTODOS PRIVADOS DA CLASSE

	// Este mtodo usa regras de deciso do mtodo do paraleleppedo
	// para decidir se um pixel pertence a um tema (classe) ou no.
	// Pixels que no pertencem a nenhuma classe definida pelo
	// analista ser alocado na classe 'no classificado,' 
	// recebendo a cor preta na imagem classificada.
	//
	private void calculaRegraDeDecisao() {
		int t, nb, w, h, nt, pixel, tipo, contaClasse[], percent;
		double sup, inf, media[][], std[][];
		boolean tautologia = true;
		String arquivo = null;
		Vector vecMedia, vecSTD;
		Raster[] srcR;
		
		// Variveis usadas para gravar arquivo .csv
		//
		String[] cores = {"Vermelho", "Verde", "Azul", "Rosa", 
											"Turquesa", "Amarelo", "Violeta"};
		String ln, pV = ";", grupo, cor, pixels, porcento;
		FileWriter writer;
		BufferedWriter buf_writer;
		String arquivoCSV = null;
		
		nb = banda.length;
		nt = sTemas.length;
		w = banda[0].getWidth();
		h = banda[0].getHeight();
		media = new double[nt][nb];
		std = new double[nt][nb];
		contaClasse = new int[nt];
		srcR = new Raster[nb];

		// Prepara imagem para gravar classificao.
		//
		tipo = BufferedImage.TYPE_3BYTE_BGR;
		BufferedImage imagemClasses = null;
		imagemClasses = new BufferedImage(w, h, tipo); 
		WritableRaster outWR = imagemClasses.getRaster(); 

		vecMedia = vecSTD = new Vector();
    System.out.println("Paralelepipedo::Regras: #bandas = " + nb + ", #temas = " + nt);

		// L os valores armazenados em vectors para arrays,
		// fazendo a devida converso de valores.
		//
		for (int i = 0; i < nt; i++) {
			vecMedia = (Vector)vetorMedBanda.elementAt(i);
			vecSTD = (Vector)vetorSTDBanda.elementAt(i);
			for (int j = 0; j < nb; j++) {
				Double med  = (Double)vecMedia.elementAt(j);
				Double desv  = (Double)vecSTD.elementAt(j);
				media[i][j] = med.doubleValue();
				std[i][j] = desv.doubleValue();
			}
		}
		
		for (int b = 0; b < nb; b++)
			srcR[b] = banda[b].getRaster();

		// Agora que temos os valores das mdias e desvio padro,
		// podemos aplicar a regra de deciso para cada pixel.
		// L valores dos pixels e aplica a regra de deciso.
		// Assumimos um paraleleppedo com altura de -1dp a +1dp.
		//
		for (int lin = 0; lin < h; lin++) {
			for (int col = 0; col < w; col++) {

				// Acessa um tema.
				//
				for (t = 0; t < nt; t++) {
					tautologia = true;

					// Acessa a banda correspondente.
					//
					for (int b = 0; b < nb; b++) {
						pixel = srcR[b].getSample(col, lin, 0);
	 					inf = media[t][b] - std[t][b];
						sup = media[t][b] + std[t][b];
						if ((pixel >= inf) && (pixel <= sup))
							tautologia = true;
						else
							tautologia = false;
					}
					if (tautologia) {

						// Grava pixel classificado como classe 't'.
						//
						gravaClasse(col, lin, t, outWR);

						// Conta pixels da classe 't'.
						//
						contaClasse[t]++;

						// No preciso percorrer mais os temas.
						//
						continue;
					}
				}	// tema
			}	//col
		}	//lin

		// Grava imagem classificada.
		//
		try {
			String search = ".";
			int i = arquivoImagem.indexOf(search);
			arquivo = arquivoImagem.substring(0, i);
			arquivoCSV = arquivo;
			arquivo += "_Piped.png";
			System.out.println("\nParalelepipedo::Imagem classificada: '" + arquivo);
			ImageIO.write(imagemClasses, "png", new File(arquivo));
		} catch (IOException e) {
			String msg = "Paralelepipedo::problema gravando arquivo -> ";
			msg +=  "'" + arquivo + "'";
			System.out.println(msg);
		}
		/*System.out.println("\nPixels classificado por classe:");
		for (t = 0; t < nt; t++) {
			System.out.print("\nParalelepipedo::Classe '" + sTemas[t] + "' = " 
												+ contaClasse[t]);
			percent = (contaClasse[t]*100)/(w*h);
			System.out.println(" ("+percent+"%)");
		}*/

		// Grava arquivo .csv
		//
		arquivoCSV += "_" + String.valueOf(nt) + "_Piped" + ".csv";
		try {
				writer = new FileWriter(arquivoCSV);
				buf_writer = new BufferedWriter(writer);
				for (t = 0; t < nt; t++) {
					System.out.print("\nClasse '" + sTemas[t] + "' = " + contaClasse[t]);
					percent = (contaClasse[t]*100)/(w*h);
					System.out.println(" ("+percent+"%)");
					ln = "";
					grupo = "Classe " + String.valueOf(t + 1);
					cor = cores[t];
					pixels = String.valueOf(contaClasse[t]);
					porcento = String.valueOf(percent);
					ln = grupo + pV + sTemas[t] + pV + cor + pV + pixels + pV + porcento;
					grava_linha(ln, buf_writer);
				}
				buf_writer.close();
				System.out.println("\nCriando arquivo: " + arquivoCSV + "\n");
		} catch (IOException e) {
				String msg = "nDistanciaMinima::Erro na gravao do arquivo " 
											+ arquivoCSV;
				JOptionPane.showMessageDialog(null, msg);
				System.exit(0);
		}

	}

	private void gravaClasse(int col, int lin, int tema,
		WritableRaster outWR) {
		switch (tema) {
			case 0:
				outWR.setSample(col, lin, 0, 255);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 0);
				break;
			case 1:
				outWR.setSample(col, lin, 0, 0);
				outWR.setSample(col, lin, 1, 255);
				outWR.setSample(col, lin, 2, 0);
				break;
			case 2:
				outWR.setSample(col, lin, 0, 0);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 255);
				break;
			case 3:
				outWR.setSample(col, lin, 0, 255);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 255);
				break;
			case 4:
				outWR.setSample(col, lin, 0, 0);
				outWR.setSample(col, lin, 1, 255);
				outWR.setSample(col, lin, 2, 255);
				break;
			case 5:
				outWR.setSample(col, lin, 0, 255);
				outWR.setSample(col, lin, 1, 255);
				outWR.setSample(col, lin, 2, 0);
				break;
			case 6:
				outWR.setSample(col, lin, 0, 128);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 255);
				break;
		}
	}

	private void grava_linha(String texto, BufferedWriter buf_writer) {
		try	{
			buf_writer.write(texto);
			buf_writer.newLine();		
		} catch (IOException e) {
			String str = "MaxVer::Erro na gravao da linha " + texto;
			System.out.println(str);
			System.exit(0);
		} 
	}
}
