/******************************************************************************
Regress.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

	Resolve modelo de regresso linear simples: y = a + b*x
  
Entrada:

	Vector com as duas bandas.

Sada:

	Desenha o diagrama de espalhamento e escreve a equao 
	de regresso.
	Calcula coeficiente de correlao 'r' e coeficiente de
	determinao 'R**2' entre duas bandas monocromticas.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/

package pID;

// Pacotes genricos
import Jama.Matrix;

import java.util.Vector;
import java.text.DecimalFormat;

// Pacotes do AWT
import java.awt.Color;
import java.awt.Rectangle;
import java.awt.Graphics2D;
import java.awt.BasicStroke;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;

// Pacote Swing para interface grfica
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.ImageIcon;
import javax.swing.JScrollPane;
import javax.swing.JOptionPane;
import javax.swing.JDesktopPane;
import javax.swing.JInternalFrame;

public class Regress {

	// ATRIBUTOS DA CLASSE

	private static int[] x, y;
	private static int xN, yN;
  private static double
									coef[],				// valores dos pesos das variveis independentes
									r,						// coeficiente de correlao de Pearson
									rQuad,				// coeficiente de determinao, R^2
									aB[], aA[][],	// arrays para matrizes
									somaYSqr;			// soma dos quadrados de y
	private static int iMaxBanda1, iMaxBanda2;

	// CONSTRUTOR

  public Regress(Vector vecImagens) {

		// Armazena valores dos pixels das imagens originais
		//
		x = (int[])vecImagens.elementAt(0);		
		y = (int[])vecImagens.elementAt(1);

		// L dimenso dos vetores. Se diferentes, emite exceo
		//
		xN = x.length;	// quantidade de observaes
    yN = y.length;

		if (xN != yN)	{
			String msg = "Vetores com dimenses diferentes!";
			String titulo = "Regresso Linear";
			JOptionPane.showMessageDialog(null, msg, titulo, 0);
		}

		// Define dimenso do array dos coeficentes a e b
		//
		coef = new double[2];

		// Constroi array aA, para definir matriz A. Essa matriz  SEMPRE quadrada.
		// o array  sempre uma dimenso maior que o nmero de variveis 
		// independentes, neste caso, uma varivel independente
		//
		aA = new double[2][2];
		aB = new double[2];

		// Armazena os valores no array aA[2][2]
		//
		iMaxBanda1 = 0;
		aA[0][0] = xN;
		for (int nO = 0; nO < xN; nO++)	{
			aA[0][1] += x[nO];
			iMaxBanda1 = Math.max(iMaxBanda1, x[nO]);
		}	
		aA[1][0] = aA[0][1];
		for (int nO = 0; nO < xN; nO++)	{
			aA[1][1] += x[nO]*x[nO];
		}
		
		// Armazena as somas e produtos das variveis dependentes e independentes
		// Esses valores vo formar o vetor 'b' da equao linear.
		//
		somaYSqr = 0.0;				// valor usado no clculo de r
		iMaxBanda2 = 0;
		for (int nO = 0; nO < yN; nO++)	{
			aB[0] += y[nO];
			somaYSqr += y[nO]*y[nO];
			iMaxBanda2 = Math.max(iMaxBanda2, y[nO]);
		}
		for (int nO = 0; nO < xN; nO++)	{
			aB[1] += x[nO]*y[nO];
		}		

		// Resolve sistema de equaes lineares A.x = b
		//
		Matrix mA = new Matrix(aA);

		// Define o vetor b
		//
		Matrix mb = new Matrix(2, 1);
		for (int v = 0; v < 2; v++) {
			mb.set(v, 0, aB[v]);
		}

		// Resolve o sistema de equaes lineares
		//
		Matrix mx = mA.solve(mb);

		// Armazena os valores dos pesos das variveis independentes
		//
		for (int v = 0; v < 2; v++) {
			coef[v] = mx.get(v, 0);
		}		
		calculaCoefRegressao();
		imprimeResultado();
	}

	// Traa o grfico de espalhamento entre as duas imagens.
  // Para reduzir o nmero de iteraes, os pixels so amostrado em 1/5.
	//
	public void mostraGrafico(JDesktopPane desktop, String img1, String img2, 
		int w_X, int h_X) {
	
		JInternalFrame frame = new JInternalFrame("Regresso",
													          true, //resizable
													          true, //closable
													          true, //maximizable
													          true);//iconifiable
		desktop.add(frame);

		//int w = w_X;
		//int h = h_X;
		int w = 300;
		int h = 300;

		BufferedImage biImagem = criaImagem(img1, img2, w, h);

		// Mostra a imagem
		//
    JLabel lsrc = new JLabel(new ImageIcon(biImagem));
    frame.getContentPane().add(new JScrollPane(lsrc));
		frame.setSize(w, h);
    frame.setVisible(true);
	}
	
	// MTODOS PRIVADOS

	private BufferedImage criaImagem(String imgX, String imgY, 
		int w_X, int h_X) {

    int xx, yy, winW, winH, taxaX, taxaY, p_X, p_Y, iLin;
    double x1, y1, x2, y2;
		BufferedImage imagem = null;

    winW = w_X; //getSize().width;
    winH = h_X; //getSize().height;
		imagem = new BufferedImage(w_X, h_X, BufferedImage.TYPE_3BYTE_BGR);
		Graphics2D g2 = imagem.createGraphics();

		// Gera imagem com fundo branco
		//
		Rectangle2D fundo = new Rectangle2D.Double(0.0, 0.0, 
																							(double)w_X, (double)h_X);
		g2.setPaint(Color.white);
		g2.fill(fundo);

    taxaX = (int)winW/iMaxBanda1;
    taxaY = (int)winH/iMaxBanda2;
    iLin = 0;
    while (iLin < xN) {
			p_X = x[iLin];
			p_Y = y[iLin];
      yy = winH - p_Y*taxaY;
      xx = p_X*taxaX;
      drawPoint(xx, yy, g2);
      iLin += 5;
    }
		
    // Desenha reta ajustada
		//
    x1 = 0.0;
    y1 = coef[0] + coef[1]*x1;
    x2 = iMaxBanda1;
    y2 = coef[0] + coef[1]*x2;
    g2.setPaint(Color.black);
    g2.setStroke(new BasicStroke(3));
    g2.drawLine((int)x1*taxaX, (int)(winH - y1*taxaY), (int)x2*taxaX, 
                (int)(winH - y2*taxaY));
		
    // Prepara variveis para escrever equao de regresso
		//
    String padrao = "##.###";
    String out1 = null;
    DecimalFormat formato = new DecimalFormat(padrao);
    String regressA = formato.format(coef[0]); 
    String regressr = formato.format(r);
    double R100 = rQuad*100;
    String regressR = formato.format(R100);
		if (r >= 0) {
			String regressB = formato.format(coef[1]);
			out1 = "y = " + regressA + " + " + regressB + "x";
		} else {
			String regressB = formato.format(Math.abs(coef[1]));
			out1 = "y = " + regressA + " - " + regressB + "x";
		}
    String out2 = "r = " + regressr + ", R2 = " + regressR + "%";
    String out3 = "Eixo X = " + imgX;
    String out4 = "Eixo Y = " + imgY;
    g2.drawString(out1, 3, 10); //getInsets().top + 10); 
    g2.drawString(out2, 3, 25); //getInsets().top + 25); 
    g2.drawString(out3, 3, 40); //getInsets().top + 25); 
    g2.drawString(out4, 3, 55); //getInsets().top + 25); 
    g2.drawImage(imagem, 0, 0, null);
		return imagem;
	}
	
  private void drawPoint(int xA, int yA, Graphics2D gContxt) {
    int r1 = 2;
		
    Rectangle point = new Rectangle(xA-r1, yA-r1, 2*r1, 2*r1);
    gContxt.setPaint(Color.red);
    gContxt.fill(point);
    gContxt.setPaint(Color.black);
    gContxt.draw(point);
  }
	
	// Este mtodo s serve para a regresso simples
	//
	private void calculaCoefRegressao() {

		double	mediaY,  
						sqt,					// soma dos quadrados total (para R^2)
						sqr,					// soma dos quadrados da regresso (para R^2)
						yHat,					// valor de y estimado pela regresso
						numerador, denominador, fator1;

		// Para calcular o coeficiente de determinao, R^2, precisa da mdia 
		// da varivel dependente
		//
		mediaY = 0.0;
		for (int i = 0; i < yN; i++)
			mediaY += y[i];
		mediaY /= yN;

		// Calcula soma dos quadrados total (sqt) e soma dos quadrados
		// da regresso (sqr)
		//
		sqt = sqr = 0.0;
		for (int i = 0; i < yN; i++) {
			sqt += (y[i] - mediaY)*(y[i] - mediaY);
			yHat = coef[0] + coef[1]*x[i];
			sqr += (yHat - mediaY)*(yHat - mediaY);
		}
    rQuad = (sqr)/(sqt);

		numerador = aB[1] - (aA[0][1]*aB[0])/xN;
		denominador = aA[1][1] - aA[0][1]*aA[0][1]/xN;
		fator1 = somaYSqr - aB[0]*aB[0]/yN;
    r = numerador/Math.sqrt(denominador*fator1);	
	}

  private void imprimeResultado() {

		// Imprime resultado
		//
    String padrao = "##.###";
    String out1 = null;
    DecimalFormat formato = new DecimalFormat(padrao);
		String regressr = formato.format(r);
    String regressA = formato.format(coef[0]); 
    double R100 = rQuad*100;
    String regressR = formato.format(R100);
		if (r >= 0) {
			String regressB = formato.format(coef[1]);
			out1 = "\n\ty = " + regressA + " + " + regressB + "x";
		} else {
			String regressB = formato.format(Math.abs(coef[1]));
			out1 = "\n\ty = " + regressA + " - " + regressB + "x";
		}      
    String out2 = "\n\tr = " + regressr + ", R2 = " + regressR + "%";
    System.out.println("\nEquacao de regressao:"); 
    System.out.println(out1); 
    System.out.println("\nCoef. de correlacao (r) e de determinacao (R**2):");
    System.out.println(out2 + "\n"); 
  }

}
