/******************************************************************************
RegressMultipla.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

	Classe chamada pela classe BandasRegMulti

	Resolve modelo de regresso llinear mltipla: y = b0 + b1*x + b2*x + ...

	Calcula coeficiente de correlao mltiplo, 'R', e coeficiente de
	determinao mltiplo 'R^2' entre vrias bandas monocromticas.
  
Entrada:

	Dois arrays de int, um com as variveis independentes e outro com a
	varivel dependente.

Sada:

	Equao de regresso.
	Coeficiente de correlao (R) e determinao (R^2) mltiplos.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/

package pID;


// Pacotes genricos
import java.text.DecimalFormat;
import Jama.Matrix;

// Pacote Swing para interface grfica
import javax.swing.JOptionPane;

public class RegressMultipla {

	// ATRIBUTOS DA CLASSE

	private static double x[][], y[];
	private static int xN, yN, xV;
  private static double
									coef[],				// valores dos pesos das variveis independentes
									rMulti,				// coeficiente de correlao mltiplo
									rQuadM,				// coeficiente de determinao, R^2
									aB[], aA[][];	// arrays para matrizes

	// CONSTRUTOR

	public RegressMultipla(int[][] xA, int[] yA) {

		int colunas_aA, linhas_aA;
		
		// L dimenso dos vetores. Se diferentes, emite exceo
		//
		xN = xA[0].length;		// quantidade de observaes
		xV = xA.length;				// quantidade de variveis
    yN = yA.length;

		if (xN != yN)	{
			String msg = "Imagens com dimenses diferentes!";
			String titulo = "Regresso Mltipla";
			JOptionPane.showMessageDialog(null, msg, titulo, 0);
		}

		// Define dimenso dos arrays do atributo
		//
		x = new double[xV][xN];
		y = new double[yN];
		coef = new double[xV + 1];

		// Constroi array aA, para definir matriz A. Essa matriz  SEMPRE quadrada.
		// o array  sempre uma dimenso maior que o nmero de variveis
		// independentes.
		//
		colunas_aA = linhas_aA = xV + 1;
		aA = new double[linhas_aA][colunas_aA];
		aB = new double[linhas_aA];

		// Armazena valores da varivel dependente
		//
		for (int nO = 0; nO < xN; nO++)	{
			y[nO] = (double)yA[nO];
		}
		// Armazena valores das variveis independentes
		//
		for (int nV = 0; nV < xV; nV++) {
			for (int nO = 0; nO < xN; nO++)	{
				x[nV][nO] = (double)xA[nV][nO];
			}
		}

		// Primeiro, armazena os valores da primeira linha e coluna
		//
		aA[0][0] = xN;
		for (int coluna = 1; coluna < colunas_aA; coluna++) {
			for (int nO = 0; nO < xN; nO++)	{
				aA[0][coluna] += x[coluna - 1][nO];
			}	
		}	
		for (int linha = 1; linha < linhas_aA; linha++) {
			aA[linha][0] = aA[0][linha];
		}	

		// Depois, armazena os outros valores
		//
		for (int linha = 1; linha < linhas_aA; linha++) {
			for (int coluna = 1; coluna < colunas_aA; coluna++) {
				for (int nO = 0; nO < xN; nO++)	{
					aA[linha][coluna] += x[linha - 1][nO]*x[coluna - 1][nO];
				}	
			}
		}
		
		// Armazena as somas e produtos das variveis dependentes e independentes
		// Esses valores vo formar o vetor 'b' da equao linear.
		//
		for (int nO = 0; nO < xN; nO++)	{
			aB[0] += y[nO];
		}
		for (int linha = 1; linha < linhas_aA; linha++) {
			for (int nO = 0; nO < xN; nO++)	{
				aB[linha] += x[linha - 1][nO]*y[nO];
			}
		}		

		// Resolve sistema de equaes lineares A.x = b
		//
		Matrix mA = new Matrix(aA);

		// Define o vetor b
		//
		Matrix mb = new Matrix(linhas_aA, 1);
		for (int v = 0; v < linhas_aA; v++) {
			mb.set(v, 0, aB[v]);
		}

		// Resolve o sistema de equaes lineares
		//
		Matrix mx = mA.solve(mb);

		// Armazena os valores dos pesos das variveis independentes
		//
		for (int v = 0; v < linhas_aA; v++) {
			coef[v] = mx.get(v, 0);
		}		
		calculaCoefMultiplo();
		imprimeResultado();

	}

	// MTODOS PRIVADOS

	// Mtodo para regresso mltipla
	// calcula valores de R e R^2 mltiplo
	private void calculaCoefMultiplo() {
		
		double 
			sqt,					// soma dos quadrados total da varivel dependente
			sqr,					// soma dos quadrados resultante da regresso
			yHat,					// valor de y estimado pela regresso
			somaY2, somaY, somaYHat, somaYHat2;
		int nVar;

		nVar = aA.length;

		sqt = sqr = somaY2 = somaY = somaYHat2 = somaYHat = yHat = 0.0;
		for (int i = 0; i < xN; i++) {
			somaY += y[i];
			somaY2 += y[i]*y[i];
			yHat += coef[0];
			for (int j = 1; j < nVar; j++) {
				yHat += coef[j]*x[j - 1][i];
			}
			somaYHat += yHat;
			somaYHat2 += yHat*yHat;
			yHat = 0.0;
		}
		sqt = somaY2 - somaY*somaY/xN;
		sqr = somaYHat2 - somaYHat*somaYHat/xN;
		rQuadM = (sqr/sqt)*100.;
		rMulti = Math.sqrt(rQuadM)/10.;
	}

	private void imprimeResultado() {

		// Imprime resultado
		//
    String padrao = "##.###";
    DecimalFormat formato = new DecimalFormat(padrao);
		String[] sCoef = new String[aB.length];
    System.out.println("\nCoeficientes/pesos das variaveis independentes:\n");
		for (int i = 0; i < aB.length; i++)	{
	    sCoef[i] = formato.format(coef[i]); 
	    System.out.println("b"+i+" = "+formato.format(coef[i])); 
		}
		String regressr = formato.format(rMulti);
    String regressR = formato.format(rQuadM);
    String out2 = "\n\tR = " + regressr + " e R^2 = " + regressR + "%";
    System.out.println("\nCoef. de correlacao multiplo (R) e de determinacao (R^2):");
    System.out.println(out2 + "\n"); 
  }

}
