/******************************************************************************
Univariada.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

	Classe para criar estatsticas univariadas de uma banda.

Entrada:

	Vetor com os pixels da banda.

Sada:

	Estatsticas univariadas.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/

package pID;

// Pacotes genricos
import java.util.Arrays;
import java.text.DecimalFormat;

// Pacotes do Swing
import javax.swing.JLabel;
import java.awt.BorderLayout;
import javax.swing.JScrollPane;
import javax.swing.JDesktopPane;
import javax.swing.BorderFactory;
import javax.swing.JInternalFrame;

public class Univariada {

	// ATRIBUTOS

	// Array com momentos estatsticos da imagem
	//
	private double[] stats = new double[10];

	/*
			stats[0] = mdia
      stats[1] = desvio mdio
      stats[2] = desvio padro
      stats[3] = varincia
      stats[4] = skew
      stats[5] = curt
      stats[6] = valor mnimo
      stats[7] = valor mximo
      stats[8] = moda absoluta.
			stats[9] = mediana
  */

	// CONSTRUTOR DA CLASSE

	// Recebe um vetor coluna com os pixels da banda
	//
	public Univariada(int[] pixels) {
		      
		double ep = 0.0, sum, p;
    int N, pixel, indice, maxFreq;
    int[] frequencia = new int[256];

    for (int i = 0; i < 8; i++) 
			stats[i] = 0.0;
    stats[6] = 255.0;

		// Passo 1: Calcula a mdia, mnimo e mximo
		//
    sum = 0.0;
		N = pixels.length;

		for (int i = 0; i < N; i++) {
				pixel = pixels[i];
				sum += pixel;
        stats[6] = Math.min(stats[6], pixel);
        stats[7] = Math.max(stats[7], pixel);
        frequencia[pixel]++;
    }
    stats[0] = (double)sum/N;

    // Passo 2: Calcula segundo, terceiro e quarto momentos
    //          do desvio da mdia
		//
    for (int i = 0; i < N; i++) {
				sum = pixels[i] - stats[0];
				stats[1] += Math.abs(sum);
        ep += sum;
				stats[3] += (p = sum*sum);
				stats[4] += (p *= sum);
				stats[5] += (p *= sum);
    }
    stats[1] /= N;
    stats[3] = (stats[3] - ep*ep/N)/(N-1);
    stats[2] = Math.sqrt(stats[3]);
    if (stats[3] > 0) {
      stats[4] /= (N*(stats[3])*(stats[2]));
			stats[5] = (stats[5])/(N*stats[3]*stats[3])-3.0;
    } else
      System.out.println("Sem skew/kurtosis. Variancia = 0.");

		// Acha a moda absoluta.
		//
    maxFreq = 0;
    pixel = 1;
    for (int i = 1; i < 256; i++) {
      if (frequencia[i] > maxFreq) {
        maxFreq = frequencia[i];
        pixel = i;
      }
    }
    stats[8] = (double)pixel;

		stats[9] = calculaMediana(pixels);
  }     

	// MTODOS PBLICOS

	public double retornaMedia() {
		return stats[0];
	}

	public double retornaDesvioMedio() {
		return stats[1];
	}

	public double retornaDesvioPadrao() {
		return stats[2];
	}

	public double retornaVariancia() {
		return stats[3];
	}

	public double retornaSkew() {
		return stats[4];
	}

	public double retornaDesvioCurtosis() {
		return stats[5];
	}

	public double retornaValorMinimo() {
		return stats[6];
	}

	public double retornaValorMaximo() {
		return stats[7];
	}

	public double retornaModaAbsoluta() {
		return stats[8];
	}

	public double retornaMediana() {
		return stats[9];
	}

	public void mostraResultados(JDesktopPane desktop, String nomeImagem) {

		JInternalFrame jiFrame = new JInternalFrame("Univariada",
													          true, //resizable
													          true, //closable
													          true, //maximizable
													          true);//iconifiable

		jiFrame.setTitle("Estatstica Univariada");
		jiFrame.setSize(230, 250);
	  jiFrame.setVisible(true);
    desktop.add(jiFrame);

		String padrao = "##.##";
		DecimalFormat formato = null;
		formato = new DecimalFormat(padrao);

		String media = formato.format(stats[0]);
		String desviom = formato.format(stats[1]);
		String desviop = formato.format(stats[2]);
		String varian = formato.format(stats[3]);
		String skew = formato.format(stats[4]);
		String curt = formato.format(stats[5]);
		String min = formato.format(stats[6]);
		String max = formato.format(stats[7]);
		String moda = formato.format(stats[8]);
		String mediana = formato.format(stats[9]);

		String sEstat = 
      "<html>" + "<P>" +
      "Valores da imagem: " + nomeImagem + 
      "<UL>" +
      "  <LI>m&eacute;dia = " + media +
      "  <LI>desvio padr&atilde;o m&eacute;dio = " + desviom +
      "  <LI>desvio padr&atilde;o = " + desviop +
      "  <LI>vari&acirc;ncia = " + varian +
      "  <LI>skewness = " + skew +
      "  <LI>curtosis = " + curt +
      "  <LI>valor m&iacute;nimo = " + min +
      "  <LI>valor m&aacute;ximo = " + max +
      "  <LI>moda absoluta = " + moda +
      "  <LI>mediana = " + mediana +
      "</UL>";

		JLabel lEstat = new JLabel(sEstat, JLabel.CENTER);
		lEstat.setBorder(BorderFactory.createTitledBorder("Estatsticas"));
		jiFrame.getContentPane().add(lEstat, BorderLayout.NORTH);	
	}

	// MTODOS PRIVADOS

  // Calcula valor da mediana
	//
  private double calculaMediana(int[] pixels) {
    int N;
		double valor = 0.0;

		N = pixels.length;

    Arrays.sort(pixels);
    if (N%2 == 0)

      // Nmero par de pixels
			//
      valor = 0.5*(pixels[(int)N/2] + pixels[(int)(N/2)+1]);
    else

      // Nmero mpar de pixels
			//
      valor = pixels[(N+1)/2];
    return valor;
  }

}
