/******************************************************************************
RNA.java

Direitos Autorais Reservados (c) 2009 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob
os termos da Licena Pblica Geral GNU conforme publicada pela Free Software
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais 
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
Temple Street, Suite 330, Boston, MA 02111-1307 USA.

********************************************************************************

Parte integrante do software 'Sofia Imagens - SOFtware Inteligente de Anlise 
                                              de Imagens'.

Este fonte faz parte do livro:

	Processamento de Imagens Digitais: Mtodos Multivariados em Java.
	Jos Iguelmar Miranda.
	Embrapa Informtica Agropecuria.
	Campinas, maio de 2009.

Descrio:

  Esta classe implementa uma rede de neurnios artificiais (RNA), do tipo
	retropropagao do erro.

Entrada:

	Arquivo com extenso .RNA, contendo valores normalizados dos pixels e
	das categorias.

Sada:

	Rede treinada.

Desenvolvido por:

    Jos Iguelmar Miranda

Informaes do CVS:
       $Source$:
       $Revision$:
       $Date$:

******************************************************************************/

package pID;

// Pacotes genricos
import java.io.File;
import java.io.FileWriter;
import java.io.FileReader;
import java.io.PrintStream;
import java.io.IOException;
import java.io.BufferedWriter;
import java.io.BufferedReader;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.FileNotFoundException;
import java.util.StringTokenizer;

// Pacotes da API ANNeF - rede de neurnios artificiais
import ai.annef.Weight;
import ai.annef.ExamplesSet;
import ai.annef.LinearFunction;
import ai.annef.Backpropagation;
import ai.annef.BackpropImporter;
import ai.annef.WeightsCollection;
import ai.annef.QuadraticFunction;
import ai.annef.LayerWasNotCreatedError;
import ai.annef.SymmetricSigmoidFunction;
import ai.annef.TangentHyperbolicFunction;
import ai.annef.AsymmetricSigmoidFunction;

// Pacotes do AWT
import java.awt.image.Raster;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;

// Pacotes Swing
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.swing.JDesktopPane;

// Leitura em modo imediato - para J2SE 1.4+
import javax.imageio.ImageIO;

public class RNA {
	
	// ATRIBUTOS
	//
  private String rnaDir, arquivoTreina = null, arquivoHeader = null;
	private File aFile;
  private Backpropagation backPropNet;
	private WeightsCollection wCol;
  private ExamplesSet eSetTreina;
  static int nCamadas, nEntradas, nSaidas, nOcultos;

	// banda - BufferedImage das bandas selecionadas.
	//
	private BufferedImage[] banda;
	private int numBandas;
	private int numTemas;
	private float[] valorNormalizado;
	private String arquivoImagem;
	private int[]	min_bandaN, max_bandaN;

	// sTemas = guarda os temas definidos pelo usurio, para gravao do
	// arquivo .csv
	//
	private String[] sTemas;

	// CONSTRUTOR
	//
  public RNA () {
		int tam, status;
		String sufixo;
		JFileChooser fileChooser;

		rnaDir = ".\\dadosRNA";

		// Cria, se no existe, diretrio para arquivo de relatrio
		//
		boolean success = new File(rnaDir).mkdir();
		if (success) {
			String msg = "Criou diretrio: " + rnaDir;
			String titulo = "Rede de Neurnios Artificiais";
			JOptionPane.showMessageDialog(null, msg, titulo, 1);
		}

		// ************************************************
		// * Abre arquivo com valores normalizados (.rna) *
		// ************************************************
		//
		fileChooser = new JFileChooser(".");
		fileChooser.setDialogTitle("Abre arquivo de treinamento (*.rna)");
		fileChooser.addChoosableFileFilter(new TipoArquivo(".rna"));
    status = fileChooser.showOpenDialog(null);
    if (status == JFileChooser.APPROVE_OPTION) {
			aFile = fileChooser.getSelectedFile();
			arquivoTreina = aFile.getName();
			tam = arquivoTreina.length();
			sufixo = arquivoTreina.substring((tam-3), tam);
			if (!sufixo.equals("rna")) {
				String str = "Arquivo tem que ser do tipo .rna!";
				JOptionPane.showMessageDialog(null, str);
				aFile = null;
			}
		} else 
			if (status == JFileChooser.CANCEL_OPTION)
				aFile = null;

		// *********************************
		// * Abre arquivo cabealho (.hdr) *
		// *********************************
		//
		fileChooser = new JFileChooser(".");
		fileChooser.setDialogTitle("Abre arquivo cabealho (*.hdr)");
		fileChooser.addChoosableFileFilter(new TipoArquivo(".hdr"));
    status = fileChooser.showOpenDialog(null);
    if (status == JFileChooser.APPROVE_OPTION) {
			aFile = fileChooser.getSelectedFile();
			arquivoHeader = aFile.getName();
			tam = arquivoHeader.length();
			sufixo = arquivoHeader.substring((tam-3), tam);
			if (!sufixo.equals("hdr")) {
				String str = "Arquivo tem que ser do tipo .hdr!";
				JOptionPane.showMessageDialog(null, str);
				aFile = null;
			}
		} else 
			if (status == JFileChooser.CANCEL_OPTION)
				aFile = null;
  }
  
	// MTODOS PBLICOS

	// Retorna arquivo de treinamento
	//
	public String retornaArquivoTreinamento() {
		return arquivoTreina;
	}

	// Retorna arquivo cabealho
	//
	public String retornaArquivoHeader() {
		return arquivoHeader;
	}

  // Define topologia da rede a partir de dados pr-definidos
	// Mtodo retorna "true" se RNA foi criada, seno, retorna "false"
	//
  public boolean criaTopologia(int camadas, int entradas, int saida, 
															 int ocultos, int outFunc, int hiddenFunc,
															 long seed, float range) {
    boolean ret = false;
		
    nCamadas = camadas;						// Nmero de camadas da rede
    nEntradas = entradas;					// Nmero de neurnios de entrada
    nSaidas = saida;							// Nmero de neurnios de sada
    nOcultos = ocultos;						// Nmero de neurnios da camada oculta

		// Cria RNA do tipo retropropagao do erro
		//
		backPropNet = new Backpropagation(nCamadas, nEntradas, nSaidas);	

		// Inclui camada oculta. S usamos uma camada oculta
		//
		try{
       backPropNet.addHiddenLayer(nOcultos);
			 ret = true;
    } catch(LayerWasNotCreatedError erro){
       String msg = "Erro na adio da(s) camada(s) interna(s)";
			 String titulo = "Rede de Neurnios Artificiais";
			 JOptionPane.showMessageDialog(null, msg, titulo, 0);
       return false;
    }
		
		// Definio da funo de ativao dos neurnios das camadas
		// oculta e de sada
		//
    switch(hiddenFunc){
      case 0: // AsymmetricSigmoidFunction
          backPropNet.setHiddenFunction(new AsymmetricSigmoidFunction());
          break;
      case 1: // SymmetricSigmoidFunction 
          backPropNet.setHiddenFunction(new SymmetricSigmoidFunction());
          break; 
      case 2:// TangentHyperbolicFunction
          backPropNet.setHiddenFunction(new TangentHyperbolicFunction());
          break;
      case 3: // LinearFunction
          backPropNet.setHiddenFunction(new LinearFunction());
          break;
      case 4: // QuadraticFunction
          backPropNet.setHiddenFunction(new QuadraticFunction());
          break;
    }
			
		switch(outFunc){
      case 0: // AsymmetricSigmoidFunction
          backPropNet.setOutputFunction(new AsymmetricSigmoidFunction());
          break;
      case 1: // SymmetricSigmoidFunction 
          backPropNet.setOutputFunction(new SymmetricSigmoidFunction());
          break; 
      case 2:// TangentHyperbolicFunction
          backPropNet.setOutputFunction(new TangentHyperbolicFunction());
          break;
      case 3: // LinearFunction
          backPropNet.setOutputFunction(new LinearFunction());
          break;
      case 4: // QuadraticFunction
          backPropNet.setOutputFunction(new QuadraticFunction());
          break;
    }
		
		// Cria a topologia
		//
    backPropNet.createTopology(range, seed);		
			
    return ret;
  }

  // Atribui os pesos das sinapses
	//
  public boolean atribuiPesos(String arquivoPesos) {

		boolean ret = false;

    try {
      BackpropImporter bImp = new BackpropImporter();
      wCol = bImp.importFile(arquivoPesos);
      try {
         backPropNet.importWeights(wCol);
				 ret = true;
      } catch (ArrayIndexOutOfBoundsException ev) {
				 String msg = "WeightsCollection::Erro na importao dos pesos";
				 String titulo = "Rede de Neurnios Artificiais";
				 JOptionPane.showMessageDialog(null, msg, titulo, 0);
	      }    
    } catch (IOException ioe) {
		  String msg = "Erro na leitura do arquivo de pesos " + arquivoPesos;
			String titulo = "Rede de Neurnios Artificiais";
			JOptionPane.showMessageDialog(null, msg, titulo, 0);
    }
		return ret;
  }
  
	// Gerao da base de exemplos
	// Este mtodo carrega o arquivo contendo o conjunto de treinamento, gerando
	// o arquivo de exemplos. O arquivo de treinamento deve ter o formato:
	//
	// classe1 	classe2 ... 	classe n	sada desejada (no faz parte do arquivo!)
	//	0.10			0.15					0.12				0.2
	//	0.30			0.32					0.35				0.4
	//	etc
	//
	public boolean geraExemplos() {
		
		int entradas,		// colunas no arquivo de entrada com as classes
				saidas;			// colunas no arquivo de entrada com as sadas desejadas
		boolean ret = false;

		// Certifica que o usurio abriu o arquivo de treinamento (.rna)
		//
		if (arquivoTreina == null) {
		  String msg = "Favor abrir arquivo de exemplos!";
			String titulo = "Rede de Neurnios Artificiais";
			JOptionPane.showMessageDialog(null, msg, titulo, 0);
			return false;
		}

		entradas = backPropNet.getNInputs();
		saidas = backPropNet.getNOutputs();
		try { 
      eSetTreina = new ExamplesSet(entradas, saidas, arquivoTreina);
			ret = true;
		} catch (FileNotFoundException fnfe) {
		  String msg = "Erro na leitura do arquivo de treinamento " + arquivoTreina;
			String titulo = "RNA - Gera Exemplos";
			JOptionPane.showMessageDialog(null, msg, titulo, 0);
			System.out.println(fnfe.getMessage());
		} catch(IOException ioe) {
		  String msg = "Erro na leitura do arquivo de treinamento " + arquivoTreina;
			String titulo = "RNA - Gera Exemplos";
			JOptionPane.showMessageDialog(null, msg, titulo, 0);
			System.out.println(ioe.getMessage());
		} catch(NumberFormatException nfe) {
		  String msg = "Erro na leitura do arquivo de treinamento " + arquivoTreina;
			String titulo = "RNA - Gera Exemplos";
			JOptionPane.showMessageDialog(null, msg, titulo, 0);
			System.out.println(nfe.getMessage());
		}
		return ret;
	}

  // Treina rede. O treinamento utiliza o aprendizado com "exemplos" usando
	// somente uma base de exemplos para treino
	// Alm do arquivo de exemplos ou treinamento, usam-se as variveis:
	//	1. Taxa de aprendizado (lv)
	//	2. Momento (mom)
	//	3. pocas (epochs)
	//
  public boolean treinaRede(float thres, float lv, float mom, int epochs, 
														int peek) {

		float incLearning,		// incremento da taxa de aprendizado
					decLearning;		// decremento da taxa de aprendizado
		String arquivoRelatorio = null;
		
		// Nome do arquivo com relatrio de treinamento da rede
		//
		arquivoRelatorio = rnaDir + "\\relatAprendizadoRNA.txt";

		// Uso da taxa de aprendizado adaptativo
		//
    incLearning = 0.0f;
    decLearning = 1.0f;
		backPropNet.setAdaptiveLearningRate(incLearning, decLearning);
		
		// Configurao da margem de erro para contabilizao do erro
		//
    backPropNet.setCount(thres, Backpropagation.FOR_EXAMPLE);
		
		// Treinamento por exemplos
		//
		// Certifica que o usurio abriu o arquivo de treinamento (.rna)
		//
		if (arquivoTreina == null) {
		  String msg = "Favor abrir arquivo de treinamento!";
			String titulo = "RNA - Treina rede";
			JOptionPane.showMessageDialog(null, msg, titulo, 0);
			return false;
		}

		// Treina a rede com exemplos
		//
		System.out.println("\nTreinando a rede. Aguarde...\n");			
		backPropNet.learningExamples(eSetTreina, lv, mom, epochs);
		
		// Gera relatrio de aprendizado
		// Antes,  necessrio definir o intervalo de pocas, com a varivel peek
		//
		try {
			FileOutputStream rAprendizado = new FileOutputStream(arquivoRelatorio);
			PrintStream ps = new PrintStream(rAprendizado);			
			backPropNet.setPeek(peek);
			backPropNet.reportLearningExamples(eSetTreina, lv, mom, epochs, ps);
		} catch (IOException e) {
		  String msg = "Erro::arquivo com relatrio no foi gerado: " + 
										arquivoRelatorio;
			String titulo = "RNA - Treina rede";
			JOptionPane.showMessageDialog(null, msg, titulo, 0);
			System.out.println(e.getMessage());			
		}
		return true;
	}

	// Este mtodo l e abre os arquivos das imagens contidos no arquivo
	// cabealho (.hdr), armazenando-os em BufferedImage.
	//
	public boolean abreBandas() {
		int tam;
		FileReader headerFile = null;
		String line, nomeBanda[] = null;
		StringTokenizer st = null;
		File file;

		try {
			headerFile = new FileReader(arquivoHeader);
			BufferedReader br = new BufferedReader(headerFile);

			// L a primeira linha do arquivo .hdr e pega nmero de bandas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numBandas = Integer.parseInt(st.nextToken());
			nomeBanda = new String[numBandas];

			// Percorre todas as bandas
			//
			for (int nb = 0; nb < numBandas; nb++) {

				// L linha com nome da banda
				//
				line = br.readLine();
				st = new StringTokenizer(line);
				nomeBanda[nb] = st.nextToken();
				System.out.println("\nRNA::Banda: " + nomeBanda[nb]);
			}

			// L a linha do arquivo .hdr com nmero de temas
			//
			line = br.readLine();
			st = new StringTokenizer(line);
			numTemas = Integer.parseInt(st.nextToken());
			sTemas = new String[numTemas];

			// Armazena nomes dos temas, para gravao no arquivo .csv
			//
			for (int nt = 0; nt < numTemas; nt++) {

				// L linha com nome do tema
				//
				line = br.readLine();
				st = new StringTokenizer(line);
				sTemas[nt] = st.nextToken();
			}

			// Normaliza valores dos temas
			//
			valorNormalizado = new float[numTemas];
			for (int nt = 0; nt < numTemas; nt++) {
				valorNormalizado[nt] = (float)((nt+1) - 1)/(float)(numTemas - 1);
			}		
			headerFile.close();
    } catch (IOException e) {
				String msg = "abreBandas: problema com imagem: " + arquivoHeader;
				String titulo = "Rede de Neurnios Artificiais";
				JOptionPane.showMessageDialog(null, msg, titulo, 0);
				return false;
    }
		
		arquivoImagem = nomeBanda[0];

		// Armazena pixels em BufferedImage
		//
		banda = new BufferedImage[numBandas];
		for (int i = 0; i < numBandas; i++) {
	    file = new File(nomeBanda[i]);
			try {
				banda[i] = ImageIO.read(file);
		  } catch(Exception e) {
				String msg = "abreBandas: problema lendo: " + file;
				String titulo = "Rede de Neurnios Artificiais";
				JOptionPane.showMessageDialog(null, msg, titulo, 0);
				return false;
			}
		}

		return true;
	}
  
	// Classifica imagem
	//
	public void classificaImagem() {
		int w, h, tipo, tema, contaClasse[], nc;
		float pixels[], pixelOut[], pixelsNorm[], numerador, denominador;
		double valor, f, valor1;
		BufferedImage imagemClasses = null;
		WritableRaster outWR;
		Raster[] srcR;
		String arquivo = null;

		// Variveis usadas para gravar arquivo .csv
		//
		String[] cores = {"Vermelho", "Verde", "Azul", "Rosa", 
											"Turquesa", "Amarelo", "Violeta"};
		String ln, pV = ";", grupo, cor, pixel, porcento;
		FileWriter writer;
		BufferedWriter buf_writer; 
		String arquivoCSV = null;
		int percent;

		w = banda[0].getWidth();
		h = banda[0].getHeight();
		pixelOut = new float[1];
		pixels = new float[numBandas];
		pixelsNorm = new float[numBandas];
		srcR = new Raster[numBandas];
		contaClasse = new int[numTemas];

		// Prepara imagem para gravar classificao.
		//
		tipo = BufferedImage.TYPE_3BYTE_BGR;
		imagemClasses = new BufferedImage(w, h, tipo); 
		outWR = imagemClasses.getRaster(); 

		for (int nb = 0; nb < numBandas; nb++)
			srcR[nb] = banda[nb].getRaster();

		// Encontra valores extremos para posterior normalizao dos valores
		// dos pixels
		//
		encontraExtremos(srcR, w, h);

		// Percorre todas as imagens para calcular valores de classificao
		//
		for (int lin = 0; lin < h; lin++) {
			for (int col = 0; col < w; col++) {

				// Acessa pixels das bandas.
				//
				for (int nb = 0; nb < numBandas; nb++) {
					pixels[nb] = (float)srcR[nb].getSample(col, lin, 0);

					// Normaliza valores dos pixels
					//
					numerador = pixels[nb] - (float)min_bandaN[nb];
					denominador = (float)max_bandaN[nb] - (float)min_bandaN[nb];
					pixelsNorm[nb] = (float)numerador/(float)denominador;
				}

				// Ativa a rede e encontra valor de classificao dos pixels passados
				// para ela
				//
				pixelOut = classificaPixels(pixelsNorm);

				// Encontra valor do tema
				// Os temas foram normalizados, para atender requisito do algoritmo
				// da API ANNeF
				// Primeiro: multiplica valor de sada da RNA por nmero de temas menos
				// um. Esse valor serve para referenciar o valor de sada dentro da
				// escala dos valores normalizados dos temas.
				//
				valor1 = (double)pixelOut[0]*(double)(numTemas - 1);

				// Segundo: extrai valor inteiro do ponto flutuante, para alocar o
				// pixel em um tema
				//
				tema = (int) valor1;

				// Terceiro: separa valor decimal do ponto flutuante
				//
				f = valor1 - (double) tema;

				// Quarto: se o valor decimal for maior que 0,5, ento o pixel deve
				// ser alocado na classe imediatamente superior
				//
				if (f >= 0.5) {
					tema++;
				}

				// Grava pixel classificado como classe 't'.
				//
				gravaClasse(col, lin, tema, outWR);

				// Conta pixels da classe 't'.
				//
				contaClasse[tema]++;
			}	//col
		}	//lin

		// Grava imagem classificada.
		//
		try {
			String search = ".";
			int i = arquivoImagem.indexOf(search);
			arquivo = arquivoImagem.substring(0, i);
			arquivoCSV = arquivo;
			arquivo += "_RNA.png";
			System.out.println("\nRNA::Imagem classificada -> '" + arquivo + "'");
			ImageIO.write(imagemClasses, "png", new File(arquivo));
		} catch (IOException e) {
			String msg = "gravaImagem: problema gravando arquivo: " + arquivo;
			String titulo = "Rede de Neurnios Artificiais";
			JOptionPane.showMessageDialog(null, msg, titulo, 0);
		}

		// Grava arquivo .csv
		//
		arquivoCSV += "_" + String.valueOf(numTemas) + "_RNA" + ".csv";
		try {
				writer = new FileWriter(arquivoCSV);
				buf_writer = new BufferedWriter(writer);
				for (int t = 0; t < numTemas; t++) {
					System.out.print("\nClasse '" + sTemas[t] + "' = " + contaClasse[t]
														+ " pixels");
					percent = (contaClasse[t]*100)/(w*h);
					System.out.println(" (" + percent + "%)");
					ln = "";
					grupo = "Classe " + String.valueOf(t + 1);
					cor = cores[t];
					pixel = String.valueOf(contaClasse[t]);
					porcento = String.valueOf(percent);
					ln = grupo + pV + sTemas[t] + pV + cor + pV + pixel + pV + porcento;
					grava_linha(ln, buf_writer);
				}
				buf_writer.close();
				System.out.println("\nRNA::Criando arquivo -> " + arquivoCSV + "'\n");
		} catch (IOException e) {
				String msg = "Erro na gravao do arquivo " + arquivoCSV;
				String titulo = "Rede de Neurnios Artificiais";
				JOptionPane.showMessageDialog(null, msg, titulo, 0);
		}
	}
  
	// MTODOS PRIVADOS

  // Classifica os valores no vetor in em uma das classes treinada na rede
	// Esta funo precisa ser chamada para a ativao da rede...!
	//
  private float[] classificaPixels(float[] in) {
    float[] out;

    out = backPropNet.active(in);
    return out;
  }

	private void gravaClasse(int col, int lin, int tema,
		WritableRaster outWR) {
		switch (tema) {
			case 0:
				outWR.setSample(col, lin, 0, 255);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 0);
				break;
			case 1:
				outWR.setSample(col, lin, 0, 0);
				outWR.setSample(col, lin, 1, 255);
				outWR.setSample(col, lin, 2, 0);
				break;
			case 2:
				outWR.setSample(col, lin, 0, 0);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 255);
				break;
			case 3:
				outWR.setSample(col, lin, 0, 255);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 255);
				break;
			case 4:
				outWR.setSample(col, lin, 0, 0);
				outWR.setSample(col, lin, 1, 255);
				outWR.setSample(col, lin, 2, 255);
				break;
			case 5:
				outWR.setSample(col, lin, 0, 255);
				outWR.setSample(col, lin, 1, 255);
				outWR.setSample(col, lin, 2, 0);
				break;
			case 6:
				outWR.setSample(col, lin, 0, 128);
				outWR.setSample(col, lin, 1, 0);
				outWR.setSample(col, lin, 2, 255);
				break;
		}
	}

	// Os extremos so usados no processo de normalizao dos valores
	// dos pixels
	//
	private void encontraExtremos(Raster[] srcR, int w, int h) {
		int nc;

		min_bandaN = new int[numBandas];
		max_bandaN = new int[numBandas];

		for (int nb = 0; nb < numBandas; nb++) {
			min_bandaN[nb] = 255;
			max_bandaN[nb] = 0;
		}

		for (int lin = 0; lin < h; lin++) {
			for (int col = 0; col < w; col++) {
				for (int nb = 0; nb < numBandas; nb++) {
					nc = srcR[nb].getSample(col, lin, 0);
					min_bandaN[nb] = Math.min(nc, min_bandaN[nb]);
					max_bandaN[nb] = Math.max(nc, max_bandaN[nb]);
				}
			}
		}
	}

	private void grava_linha(String texto, BufferedWriter buf_writer) {
		try	{
			buf_writer.write(texto);
			buf_writer.newLine();		
		} catch (IOException e) {
			String msg = "Erro na gravao da linha " + texto;
			String titulo = "Rede de Neurnios Artificiais";
			JOptionPane.showMessageDialog(null, msg, titulo, 0);
		} 
	}
}
